/*
 * icom.c
 *
 * Copyright (C) 2001 Michael Anderson, IBM Corporation
 *
 * Serial device driver.
 *
 * Based on code from serial.c
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 *
 * NOTE:  Only for users with internal modem cards (eg. iSeries 2771, 2772)
 *        find the decimal number associated with your country below
 *        and set country code operations using module parameter at install
 *        time, eg.  for Argentina the command would be (be sure number is in "")
 *                      insmod icom.o iCom_country_code="52"
 *
 * Module paramter values for
 *   iCom_country_code:
                          52     * AR Argentina *
                          52     * AW Aruba *
                           1     * AU Australia *
                          52     * AT Austria *
                          52     * BH Bahrain *
                          52     * BE Belgium *
                          52     * BR Brazil *
                          52     * BN Brunei Darussalam *
                          52     * CA Canada *
                          52     * KY Cayman Islands *
                          52     * CL Chile *
                          52     * CN China *
                          52     * CO Colombia *
                          52     * CR Costa Rica *
                          52     * HR Croatia *
                          52     * CY Cyprus *
                          37     * CZ Czech Republic *
                          52     * DK Denmark *
                          52     * EC Ecuador *
                          52     * EG Egypt *
                          52     * FI Finland *
                          52     * FR France *
                          52     * DE Germany *
                          52     * GR Greece *
                          52     * GT Guatemala *
                          48     * HK China (Hong Kong S.A.R.) *
                          48     * HU Hungary *
                          52     * IS Iceland *
                          48     * IN India *
                          48     * ID Indonesia *
                          52     * IE Ireland *
                          48     * IL Israel *
                          52     * IT Italy *
                          52     * JM Jamaica *
                          16     * JP Japan *
                          52     * KR Korea, Republic of *
                          52     * LU Luxembourg *
                          52     * MO China (Macau S.A.R.) *
                          48     * MY Malaysia *
                          52     * MX Mexico *
                          52     * MA Morocco *
                          52     * NL Netherlands *
                          52     * AN Netherlands Antilles *
                           9     * NZ New Zealand *
                          52     * NO Norway *
                          52     * PK Pakistan *
                          52     * PA Panama *
                          52     * PE Peru *
                          48     * PH Philippines *
                          48     * PL Poland *
                          52     * PT Portugal *
                          52     * QA Qatar *
                          52     * RO Romania *
                          52     * RU Russia *
                          52     * SA Saudi Arabia *
                          48     * SG Singapore *
                          52     * SK Slovakia *
                          48     * SI Slovenia *
                          53     * ZA South Africa *
                          52     * ES Spain *
                          52     * LK Sri Lanka *
                          52     * SE Sweden *
                          52     * CH Switzerland *
                          52     * TW Taiwan *
                          52     * TH Thailand *
                          52     * TT Trinidad and Tobago *
                          52     * TR Turkey *
                          52     * UA Ukraine *
                          52     * AE United Arab Emirates *
                          52     * GB United Kingdom *
                          52     * US United States of America*
                          52     * UY Uruguay *
                          52     * VE Venezuela *
                          48     * VN Vietnam *
*/
#define SERIAL_DO_RESTART
#ifdef	MODVERSIONS
#include <linux/modversions.h>
#endif
#include <linux/module.h>

MODULE_AUTHOR ("Michael Anderson <mjanders@us.ibm.com>");
MODULE_DESCRIPTION ("IBM iSeries Serial IOA driver");
MODULE_SUPPORTED_DEVICE("IBM iSeries 2745, 2771, 2772, 2742, 2793 and 2805 Communications adapters");
MODULE_LICENSE("GPL");

#include <linux/config.h>
#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/interrupt.h>
#include <linux/tty.h>
#include <linux/termios.h>
#include <linux/fs.h>
#include <linux/tty_flip.h>
#include <linux/serial.h>
#include <linux/serial_reg.h>
#include <linux/major.h>
#include <linux/string.h>
#include <linux/fcntl.h>
#include <linux/ptrace.h>
#include <linux/ioport.h>
#include <linux/mm.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/pci.h>
#include <linux/vmalloc.h>
#include <linux/smp.h>
#include <linux/smp_lock.h>
#include <linux/spinlock.h>

#include <asm/system.h>
#include <asm/segment.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/uaccess.h>
#include <asm/bitops.h>
#include <asm/serial.h>

#ifdef CONFIG_PPC_ISERIES

#ifndef CONFIG_PPC64
#include <asm/iSeries/iSeries_VpdInfo.h>
#endif

#include <asm/iSeries/iSeries_pci.h>
#include <asm/iSeries/iSeries_dma.h>
#endif

/* adapter code loads */
#include "icom.h"

#ifdef MODULE
#define CONFIG_ICOM_MODEM_CC ""
#endif

static char *iCom_country_code = CONFIG_ICOM_MODEM_CC;
MODULE_PARM(iCom_country_code, "s");
MODULE_PARM_DESC(iCom_country_code, "Modem country code configuration");

#define ICOM_TRACE /* enable port trace capabalities */

#define ICOM_DRIVER_NAME "icom"
#define ICOM_VERSION_STR "1.1.0"
#define VENDOR_ID        0x1014
#define DEVICE_ID        0x0031
#define DEVICE_ID2       0x0219
#define MAX_ADAPTERS     40
#define NR_PORTS	 (active_adapters * 4)
#define MAX_PORTS        (MAX_ADAPTERS * 4)
#define ASYNC_CLOSING    0x08000000 /* Serial port is closing */
#define ASYNC_HUP_NOTIFY 0x0001 /* Notify getty on hangups and closes 
				   on the callout port */

#ifdef MODULE
static const struct pci_device_id iCom_pci_table[] __initdata =
{
	{
		vendor: VENDOR_ID,
		device: DEVICE_ID,
		subvendor: 0xFFFF,
		subdevice: 0xFFFF,
	},
	{
		vendor: VENDOR_ID,
		device: DEVICE_ID2,
		subvendor: VENDOR_ID,
		subdevice: 0x021a,
	},
	{
		vendor: VENDOR_ID,
		device: DEVICE_ID2,
		subvendor: VENDOR_ID,
		subdevice: 0x0251,
	},
	{
		vendor: VENDOR_ID,
		device: DEVICE_ID2,
		subvendor: VENDOR_ID,
		subdevice: 0x0252,
	},
	{ }
};
MODULE_DEVICE_TABLE(pci, iCom_pci_table);
#endif

/*
 * adapter defines and structures
 */
#define ICOM_CONTROL_START_A         0x00000008
#define ICOM_CONTROL_STOP_A          0x00000004
#define ICOM_CONTROL_START_B         0x00000002
#define ICOM_CONTROL_STOP_B          0x00000001
#define ICOM_CONTROL_START_C         0x00000008
#define ICOM_CONTROL_STOP_C          0x00000004
#define ICOM_CONTROL_START_D         0x00000002
#define ICOM_CONTROL_STOP_D          0x00000001
#define ICOM_IRAM_OFFSET             0x1000
#define ICOM_DCE_IRAM_OFFSET         0x0A00
#define ICOM_CABLE_ID_VALID          0x01
#define ICOM_CABLE_ID_MASK           0xF0
#define ICOM_DISABLE                 0x80
#define CMD_XMIT_RCV_ENABLE          0xC0
#define CMD_XMIT_ENABLE              0x40
#define CMD_RCV_DISABLE              0x00
#define CMD_RCV_ENABLE               0x80
#define CMD_RESTART                  0x01
#define CMD_HOLD_XMIT                0x02
#define CMD_SND_BREAK                0x04
#define RS232_CABLE                  0x06
#define V24_CABLE                    0x0E
#define V35_CABLE                    0x0C
#define V36_CABLE                    0x02
#define NO_CABLE                     0x00
#define START_DOWNLOAD               0x80
#define ICOM_INT_MASK_PRC_A          0x00003FFF
#define ICOM_INT_MASK_PRC_B          0x3FFF0000
#define ICOM_INT_MASK_PRC_C          0x00003FFF
#define ICOM_INT_MASK_PRC_D          0x3FFF0000
#define INT_RCV_COMPLETED            0x1000
#define INT_XMIT_COMPLETED           0x2000
#define INT_IDLE_DETECT              0x0800
#define INT_RCV_DISABLED             0x0400
#define INT_XMIT_DISABLED            0x0200
#define INT_RCV_XMIT_SHUTDOWN        0x0100
#define INT_FATAL_ERROR              0x0080
#define INT_CABLE_PULL               0x0020
#define INT_SIGNAL_CHANGE            0x0010
#define HDLC_PPP_PURE_ASYNC          0x02
#define HDLC_FF_FILL                 0x00
#define HDLC_HDW_FLOW                0x01
#define START_XMIT                   0x80
#define ICOM_ACFG_DRIVE1             0x20
#define ICOM_ACFG_NO_PARITY          0x00
#define ICOM_ACFG_PARITY_ENAB        0x02
#define ICOM_ACFG_PARITY_ODD         0x01
#define ICOM_ACFG_8BPC               0x00
#define ICOM_ACFG_7BPC               0x04
#define ICOM_ACFG_6BPC               0x08
#define ICOM_ACFG_5BPC               0x0C
#define ICOM_ACFG_1STOP_BIT          0x00
#define ICOM_ACFG_2STOP_BIT          0x10
#define ICOM_DTR                     0x80
#define ICOM_RTS                     0x40
#define ICOM_RI                      0x08
#define ICOM_DSR                     0x80
#define ICOM_DCD                     0x20
#define ICOM_CTS                     0x40

#define BAUD_TABLE_LIMIT             20
static int icom_acfg_baud[] = {
  300,
  600,
  900,
  1200,
  1800,
  2400,
  3600,
  4800,
  7200,
  9600,
  14400,
  19200,
  28800,
  38400,
  57600,
  76800,
  115200,
  153600,
  230400,
  307200,
  460800};

static int                active_adapters;
static struct tty_driver  serial_driver;
static int                serial_refcount = 0;
static struct tty_struct *serial_table[MAX_PORTS];
static struct termios    *serial_termios[MAX_PORTS];
static struct termios    *serial_termios_locked[MAX_PORTS];

struct iCom_regs {
  u32                  control;        /* Adapter Control Register     */
  u32                  interrupt;      /* Adapter Interrupt Register   */
  u32                  int_mask;       /* Adapter Interrupt Mask Reg   */
  u32                  int_pri;        /* Adapter Interrupt Priority r */
  u32                  int_reg_b;      /* Adapter non-masked Interrupt */
  u32                  resvd01;
  u32                  resvd02;
  u32                  resvd03;
  u32                  control_2;      /* Adapter Control Register 2   */
  u32                  interrupt_2;    /* Adapter Interrupt Register 2 */
  u32                  int_mask_2;     /* Adapter Interrupt Mask 2     */
  u32                  int_pri_2;      /* Adapter Interrupt Prior 2    */
  u32                  int_reg_2b;     /* Adapter non-masked 2         */
};

struct func_dram {
  u32                 reserved[108];          /* 0-1B0   reserved by personality code */
  u32                 RcvStatusAddr;          /* 1B0-1B3 Status Address for Next rcv */
  u8                  RcvStnAddr;             /* 1B4     Receive Station Addr */
  u8                  IdleState;              /* 1B5     Idle State */
  u8                  IdleMonitor;            /* 1B6     Idle Monitor */
  u8                  FlagFillIdleTimer;      /* 1B7     Flag Fill Idle Timer */
  u32                 XmitStatusAddr;         /* 1B8-1BB Transmit Status Address */
  u8                  StartXmitCmd;           /* 1BC     Start Xmit Command */
  u8                  HDLCConfigReg;          /* 1BD     Reserved */
  u8                  CauseCode;              /* 1BE     Cause code for fatal error */
  u8                  xchar;                  /* 1BF     High priority send */
  u32                 reserved3;              /* 1C0-1C3 Reserved */
  u8                  PrevCmdReg;             /* 1C4     Reserved */
  u8                  CmdReg;                 /* 1C5     Command Register */
  u8                  async_config2;          /* 1C6     Async Config Byte 2*/
  u8                  async_config3;          /* 1C7     Async Config Byte 3*/
  u8                  dce_resvd[20];          /* 1C8-1DB DCE Rsvd           */
  u8                  dce_resvd21;            /* 1DC     DCE Rsvd (21st byte*/
  u8                  misc_flags;             /* 1DD     misc flags         */
#define V2_HARDWARE     0x40
#define ICOM_HDW_ACTIVE 0x01
  u8                  call_length;            /* 1DE     Phone #/CFI buff ln*/
  u8                  call_length2;           /* 1DF     Upper byte (unused)*/
  u32                 call_addr;              /* 1E0-1E3 Phn #/CFI buff addr*/
  u16                 timer_value;            /* 1E4-1E5 general timer value*/
  u8                  timer_command;          /* 1E6     general timer cmd  */
  u8                  dce_command;            /* 1E7     dce command reg    */
  u8                  dce_cmd_status;         /* 1E8     dce command stat   */
  u8                  x21_r1_ioff;            /* 1E9     dce ready counter  */
  u8                  x21_r0_ioff;            /* 1EA     dce not ready ctr  */
  u8                  x21_ralt_ioff;          /* 1EB     dce CNR counter    */
  u8                  x21_r1_ion;             /* 1EC     dce ready I on ctr */
  u8                  rsvd_ier;               /* 1ED     Rsvd for IER (if ne*/
  u8                  ier;                    /* 1EE     Interrupt Enable   */
  u8                  isr;                    /* 1EF     Input Signal Reg   */
  u8                  osr;                    /* 1F0     Output Signal Reg  */
  u8                  reset;                  /* 1F1     Reset/Reload Reg   */
  u8                  disable;                /* 1F2     Disable Reg        */
  u8                  sync;                   /* 1F3     Sync Reg           */
  u8                  error_stat;             /* 1F4     Error Status       */
  u8                  cable_id;               /* 1F5     Cable ID           */
  u8                  cs_length;              /* 1F6     CS Load Length     */
  u8                  mac_length;             /* 1F7     Mac Load Length    */
  u32                 cs_load_addr;           /* 1F8-1FB Call Load PCI Addr */
  u32                 mac_load_addr;          /* 1FC-1FF Mac Load PCI Addr  */
};

#define NUM_XBUFFS 1
#define NUM_RBUFFS 2
#define RCV_BUFF_SZ 0x0200
#define XMIT_BUFF_SZ 0x1000
struct statusArea
{
  /**********************************************/
  /* Transmit Status Area                       */
  /**********************************************/
  struct {
    u32                    leNext;         /* Next entry in Little Endian on Adapter */
    u32                    leNextASD;
    u32                    leBuffer;       /* Buffer for entry in LE for Adapter */
    u16                    leLengthASD;
    u16                    leOffsetASD;
    u16                    leLength;       /* Length of data in segment */
    u16                    flags;
#define SA_FLAGS_DONE           0x0080          /* Done with Segment */
#define SA_FLAGS_CONTINUED      0x8000          /* More Segments */
#define SA_FLAGS_IDLE           0x4000          /* Mark IDLE after frm */
#define SA_FLAGS_READY_TO_XMIT  0x0800
#define SA_FLAGS_STAT_MASK      0x007F
  } xmit[NUM_XBUFFS];
    
  /**********************************************/
  /* Receive Status Area                        */
  /**********************************************/
  struct {
    u32                    leNext;         /* Next entry in Little Endian on Adapter */
    u32                    leNextASD;
    u32                    leBuffer;       /* Buffer for entry in LE for Adapter */
    u16                    WorkingLength;  /* size of segment */
    u16                    reserv01;
    u16                    leLength;       /* Length of data in segment */
    u16                    flags;
#define SA_FL_RCV_DONE           0x0010          /* Data ready */
#define SA_FLAGS_OVERRUN         0x0040
#define SA_FLAGS_PARITY_ERROR    0x0080 
#define SA_FLAGS_FRAME_ERROR     0x0001
#define SA_FLAGS_FRAME_TRUNC     0x0002
#define SA_FLAGS_BREAK_DET       0x0004    /* set conditionally by device driver, not hardware */
#define SA_FLAGS_RCV_MASK        0xFFE6
  } rcv[NUM_RBUFFS];
};

struct iCom_port {
  u8			imbed_modem;
#define   ICOM_UNKNOWN       1
#define   ICOM_RVX           2
#define   ICOM_IMBED_MODEM   3
  unsigned char         cable_id;
  int                   open_active_count;
  struct tty_struct 	*tty;
  unsigned long	int     event;
  struct tq_struct	tqueue;
  int                   flags;
  int                   xmit_fifo_size;
  int                   baud_base;
  wait_queue_head_t	close_wait;
  wait_queue_head_t	open_wait;
  wait_queue_head_t	delta_msr_wait;
  int                   blocked_open;
  unsigned short        close_delay;
  unsigned short        closing_wait;
  unsigned long int     timeout;
  long			session; /* Session of opening process */
  long			pgrp; /* pgrp of opening process */
  unsigned char         read_status_mask;
  unsigned char         ignore_status_mask;
  struct async_icount	icount;	
  struct termios	normal_termios;
  struct termios	callout_termios;
  unsigned long int     int_reg;
  struct iCom_regs      *global_reg;
  struct func_dram      *dram;
  int                   adapter;
  int                   port;
  int                   minor_number;
  struct statusArea     *statStg;
  dma_addr_t            statStg_pci;
  u32                   *xmitRestart;
  dma_addr_t            xmitRestart_pci;
  unsigned char         *xmit_buf;
  dma_addr_t            xmit_buf_pci;
  unsigned char         *recv_buf;
  dma_addr_t            recv_buf_pci;
  int                   next_rcv;
  int                   put_length;
  int                   passed_diags;
  int                   status;
#define ICOM_PORT_ACTIVE  1 /* Port exists. */
#define ICOM_PORT_OFF     0 /* Port does not exist. */
  int                   load_in_progress;
  u32                   tpr;
#define NUM_ERROR_ENTRIES 16
  u32                   error_data[NUM_ERROR_ENTRIES];
  u32                   thread_status;
#define STATUS_INIT 0x99999999
#define STATUS_PASS 0 
  unsigned long         *trace_blk;
};

static struct iCom_adapter {
  unsigned long int  base_addr;
  unsigned char      slot;
  unsigned char      irq_number;
  struct pci_dev     *pci_dev;
  struct iCom_port   port_info[4];
  int                version;
#define ADAPTER_V1   0x0001
#define ADAPTER_V2   0x0002
  u32		     subsystem_id;
#define FOUR_PORT_MODEL				0x02521014
#define V2_TWO_PORTS_RVX			0x021A1014
#define V2_ONE_PORT_RVX_ONE_PORT_IMBED_MDM	0x02511014
  int                numb_ports;
  unsigned int       saved_bar;
  unsigned int       saved_command_reg;
  u32	             tpr;
  u32                error_data[NUM_ERROR_ENTRIES];
  u32                diag_int1;
  u32                diag_int2;
  u32                diag_int_pri1;
  u32                diag_int_pri2;
  u32                diag_int_reset1;
  u32                diag_int_reset2;
  u32                resources;
#define HAVE_MALLOC_1	        0x00000001
#define HAVE_INT_HANDLE		0x00000010
  u32                *malloc_addr_1;
} *iCom_adapter_info;

struct mthread
{
    void       *thread;
    u32		task;
    u8		port;
    int         status;
#define STATUS_INIT 0x99999999
#define STATUS_PASS 0 
    u32         tpr;
    u32         error_data[NUM_ERROR_ENTRIES];
};

#define	 IOA_FAILURE	0xFF

static spinlock_t iComlock;
/*
   Utility functions
*/
static void diag_return_resources(struct iCom_adapter *iCom_adapter_ptr);
static void return_port_memory(struct iCom_port *iCom_port_info);
static void iCom_wait_until_sent(struct tty_struct *tty, int timeout);
static void do_softint(void *);
static void iCom_start(struct tty_struct * tty);
static void iCom_flush_buffer(struct tty_struct * tty);
static void iCom_set_code(struct iCom_port *iCom_port_info);
static void mdm_send(struct iCom_port *iCom_port_info, char *mdm_cmnd,
                     int cmnd_length);
static int mdm_rcv(struct iCom_port *iCom_port_info, char *exp_str);
#ifdef ICOM_TRACE
static void TRACE(struct iCom_port *,u32 , u32);
#else
#define TRACE(x,y,z) /* nub out calls to TRACE function */
#endif

#ifdef CONFIG_PPC64
extern int register_ioctl32_conversion(unsigned int cmd,
				       int (*handler)(unsigned int, unsigned int, unsigned long, struct file *));
extern int unregister_ioctl32_conversion(unsigned int cmd);
#else
static inline int register_ioctl32_conversion(unsigned int cmd,
					      int (*handler)(unsigned int,
							     unsigned int, unsigned long, struct file *))
{
	return 0;
}
static inline int unregister_ioctl32_conversion(unsigned int cmd)
{
	return 0;
}
#endif

static u8 iCom_readb(void *address)
{
	/* Issue a 'write memory barrier' prior to the mmio to ensure ordering */
	/* This translates to an eieio instruction on ppc */
	wmb();
	return readb(address);
}

static void iCom_writeb(u8 value, void *address)
{
	/* Issue a 'memory barrier' prior to the mmio to ensure data is flushed
	 to memory. This translates to a sync instruction on ppc */
	mb();
	writeb(value, address);
}

static u16 iCom_readw(void *address)
{
	/* Issue a 'write memory barrier' prior to the mmio to ensure ordering */
	/* This translates to an eieio instruction on ppc */
	wmb();
	return readw(address);
}

static void iCom_writew(u16 value, void *address)
{
	/* Issue a 'memory barrier' prior to the mmio to ensure data is flushed
	 to memory. This translates to a sync instruction on ppc */
	mb();
	writew(value, address);
}

static u32 iCom_readl(void *address)
{
	/* Issue a 'write memory barrier' prior to the mmio to ensure ordering */
	/* This translates to an eieio instruction on ppc */
	wmb();
	return readl(address);
}

static void iCom_writel(u32 value, void *address)
{
	/* Issue a 'memory barrier' prior to the mmio to ensure data is flushed
	 to memory. This translates to a sync instruction on ppc */
	mb();
	writel(value, address);
}

static int get_port_memory(struct iCom_port *iCom_port_info)
{
  int index;
  int number_of_buffs;
  unsigned long int stgAddr;
  unsigned long int startStgAddr;
  unsigned long int offset;

  TRACE(iCom_port_info,TRACE_GET_PORT_MEM,0);

  iCom_port_info->xmit_buf = (unsigned char *)kmalloc(4096,GFP_KERNEL | GFP_DMA);
  iCom_port_info->xmit_buf_pci = pci_map_single(iCom_adapter_info[iCom_port_info->adapter].pci_dev,
						(void *)iCom_port_info->xmit_buf,
						4096,
						PCI_DMA_BIDIRECTIONAL);

  if (!iCom_port_info->xmit_buf) {
    printk("iCom:  ERROR, Can not allocate Transmit buffer\n");
    return -ENOMEM;
  }
  TRACE(iCom_port_info,TRACE_GET_PORT_MEM,(unsigned long)iCom_port_info->xmit_buf);

  iCom_port_info->recv_buf = (unsigned char *)kmalloc(4096,GFP_KERNEL | GFP_DMA);
  iCom_port_info->recv_buf_pci = pci_map_single(iCom_adapter_info[iCom_port_info->adapter].pci_dev,
						(void *)iCom_port_info->recv_buf,
						4096,
						PCI_DMA_BIDIRECTIONAL);

  if (!iCom_port_info->recv_buf) {
    printk("iCom:  ERROR, Can not allocate Receive buffer\n");
    return_port_memory(iCom_port_info);
    return -ENOMEM;
  }
  TRACE(iCom_port_info,TRACE_GET_PORT_MEM,(unsigned long)iCom_port_info->recv_buf);

  iCom_port_info->statStg = (struct statusArea *)kmalloc(4096,GFP_KERNEL | GFP_DMA);
  iCom_port_info->statStg_pci = pci_map_single(iCom_adapter_info[iCom_port_info->adapter].pci_dev,
					       (void *)iCom_port_info->statStg,
					       4096,
					       PCI_DMA_BIDIRECTIONAL);

  if (!iCom_port_info->statStg) {
    printk("iCom:  ERROR, Can not allocate Status buffer\n");
    return_port_memory(iCom_port_info);
    return -ENOMEM;
  }
  TRACE(iCom_port_info,TRACE_GET_PORT_MEM,(unsigned long)iCom_port_info->statStg);

  iCom_port_info->xmitRestart = (u32 *)kmalloc(128,GFP_KERNEL | GFP_DMA);/*only 4 bytes needed */
  iCom_port_info->xmitRestart_pci = pci_map_single(iCom_adapter_info[iCom_port_info->adapter].pci_dev,
						   iCom_port_info->xmitRestart,
						   4,
						   PCI_DMA_BIDIRECTIONAL);

  if (!iCom_port_info->xmitRestart) {
    printk("iCom:  ERROR, Can not allocate xmit Restart buffer\n");
    return_port_memory(iCom_port_info);
    return -ENOMEM;
  }

  memset(iCom_port_info->statStg, 0,4096);

  /* FODs */
  number_of_buffs = NUM_XBUFFS;
  stgAddr = (unsigned long int)iCom_port_info->statStg;
  startStgAddr = stgAddr;
  for (index = 0; index < number_of_buffs; index++)
  {
    TRACE(iCom_port_info,TRACE_FOD_ADDR,stgAddr);
    stgAddr = stgAddr + sizeof(iCom_port_info->statStg->xmit[0]);
    if (index < (number_of_buffs - 1))
    {
      iCom_port_info->statStg->xmit[index].flags = 0;
      iCom_port_info->statStg->xmit[index].leNext = 0;
      iCom_port_info->statStg->xmit[index].leNextASD = 0;
      iCom_port_info->statStg->xmit[index].leLengthASD = (unsigned short int)cpu_to_le16(XMIT_BUFF_SZ);
      iCom_port_info->statStg->xmit[index].leOffsetASD = 0;
      TRACE(iCom_port_info,TRACE_FOD_ADDR,stgAddr);
      TRACE(iCom_port_info,TRACE_FOD_XBUFF,(unsigned long)iCom_port_info->xmit_buf);
      iCom_port_info->statStg->xmit[index].leBuffer = cpu_to_le32(iCom_port_info->xmit_buf_pci);
    }
    else if (index == (number_of_buffs - 1))
    {
      iCom_port_info->statStg->xmit[index].flags = 0;
      iCom_port_info->statStg->xmit[index].leNext = 0;
      iCom_port_info->statStg->xmit[index].leNextASD = 0;
      iCom_port_info->statStg->xmit[index].leLengthASD = (unsigned short int)cpu_to_le16(XMIT_BUFF_SZ);
      iCom_port_info->statStg->xmit[index].leOffsetASD = 0;
      TRACE(iCom_port_info,TRACE_FOD_XBUFF,(unsigned long)iCom_port_info->xmit_buf);
      iCom_port_info->statStg->xmit[index].leBuffer = cpu_to_le32(iCom_port_info->xmit_buf_pci);
    }
    else
    {
      iCom_port_info->statStg->xmit[index].flags = 0;
      iCom_port_info->statStg->xmit[index].leNext = 0;
      iCom_port_info->statStg->xmit[index].leNextASD = 0;
      iCom_port_info->statStg->xmit[index].leLengthASD = 0;
      iCom_port_info->statStg->xmit[index].leOffsetASD = 0;
      iCom_port_info->statStg->xmit[index].leBuffer = 0;
    }
  }
  /* FIDs */
  startStgAddr = stgAddr;

  /* fill in every entry, even if no buffer */
  number_of_buffs = NUM_RBUFFS;
  for (index = 0; index < number_of_buffs; index++)
  {
    TRACE(iCom_port_info,TRACE_FID_ADDR,stgAddr);
    stgAddr = stgAddr + sizeof(iCom_port_info->statStg->rcv[0]);
    iCom_port_info->statStg->rcv[index].leLength = 0;
    iCom_port_info->statStg->rcv[index].WorkingLength = (unsigned short int)cpu_to_le16(RCV_BUFF_SZ); 
    if (index < (number_of_buffs - 1))
    {
      offset = stgAddr - (unsigned long)iCom_port_info->statStg;
      iCom_port_info->statStg->rcv[index].leNext = (unsigned long)cpu_to_le32(iCom_port_info->statStg_pci + offset);
      TRACE(iCom_port_info,TRACE_FID_RBUFF,(unsigned long)iCom_port_info->recv_buf);
      iCom_port_info->statStg->rcv[index].leBuffer = cpu_to_le32(iCom_port_info->recv_buf_pci);
    }
    else if (index == (number_of_buffs - 1))
    {
      offset = startStgAddr - (unsigned long)iCom_port_info->statStg;
      iCom_port_info->statStg->rcv[index].leNext = (unsigned long)cpu_to_le32(iCom_port_info->statStg_pci + offset);
      TRACE(iCom_port_info,TRACE_FID_RBUFF,(unsigned long)iCom_port_info->recv_buf + 2048);
      iCom_port_info->statStg->rcv[index].leBuffer = cpu_to_le32(iCom_port_info->recv_buf_pci + 2048);
    }
    else
    {
      iCom_port_info->statStg->rcv[index].leNext = 0;
      iCom_port_info->statStg->rcv[index].leBuffer = 0;
    }
  }

  return 0;
}

static void return_port_memory(struct iCom_port *iCom_port_info)
{
  TRACE(iCom_port_info, TRACE_RET_PORT_MEM,0);
  if (iCom_port_info->recv_buf) {
    pci_unmap_single(iCom_adapter_info[iCom_port_info->adapter].pci_dev,
		     iCom_port_info->recv_buf_pci,
		     4096,
		     PCI_DMA_BIDIRECTIONAL);
    kfree((void *)iCom_port_info->recv_buf);
    iCom_port_info->recv_buf = 0;
  }
  if (iCom_port_info->xmit_buf) {
    pci_unmap_single(iCom_adapter_info[iCom_port_info->adapter].pci_dev,
		     iCom_port_info->xmit_buf_pci,
		     4096,
		     PCI_DMA_BIDIRECTIONAL);
    kfree((void *)iCom_port_info->xmit_buf);
    iCom_port_info->xmit_buf = 0;
  }
  if (iCom_port_info->statStg) {
    pci_unmap_single(iCom_adapter_info[iCom_port_info->adapter].pci_dev,
		     iCom_port_info->statStg_pci,
		     4096,
		     PCI_DMA_BIDIRECTIONAL);
    kfree((void *)iCom_port_info->statStg);
    iCom_port_info->statStg = 0;
  }

  if (iCom_port_info->xmitRestart) {
    pci_unmap_single(iCom_adapter_info[iCom_port_info->adapter].pci_dev,
		     iCom_port_info->xmitRestart_pci,
		     4,
		     PCI_DMA_BIDIRECTIONAL);
    kfree(iCom_port_info->xmitRestart);
    iCom_port_info->xmitRestart = 0;
  }
  TRACE(iCom_port_info,TRACE_RET_MEM,0);
}

static void stop_processor(struct iCom_port *iCom_port_info)
{
  unsigned long      temp;

  switch (iCom_port_info->port) {
  case 0:
    temp = iCom_readl(&iCom_port_info->global_reg->control);
    temp = (temp & ~ICOM_CONTROL_START_A) | ICOM_CONTROL_STOP_A;
    iCom_writel(temp,&iCom_port_info->global_reg->control);
    TRACE(iCom_port_info,TRACE_STOP_PROC_A,0);
    break;
  case 1:
    temp = iCom_readl(&iCom_port_info->global_reg->control);
    temp = (temp & ~ICOM_CONTROL_START_B) | ICOM_CONTROL_STOP_B;
    iCom_writel(temp,&iCom_port_info->global_reg->control);
    TRACE(iCom_port_info,TRACE_STOP_PROC_B,0);
    break;
  case 2:
    temp = iCom_readl(&iCom_port_info->global_reg->control_2);
    temp = (temp & ~ICOM_CONTROL_START_C) | ICOM_CONTROL_STOP_C;
    iCom_writel(temp,&iCom_port_info->global_reg->control_2);
    TRACE(iCom_port_info,TRACE_STOP_PROC_C,0);
    break;
  case 3:
    temp = iCom_readl(&iCom_port_info->global_reg->control_2);
    temp = (temp & ~ICOM_CONTROL_START_D) | ICOM_CONTROL_STOP_D;
    iCom_writel(temp,&iCom_port_info->global_reg->control_2);
    TRACE(iCom_port_info,TRACE_STOP_PROC_D,0);
    break;
  default:
    printk("iCom:  ERROR:  invalid port assignment\n");
  }
}

static void start_processor(struct iCom_port *iCom_port_info)
{
  unsigned long      temp;

  switch (iCom_port_info->port) {
  case 0:
    temp = iCom_readl(&iCom_port_info->global_reg->control);
    temp = (temp & ~ICOM_CONTROL_STOP_A) | ICOM_CONTROL_START_A;
    iCom_writel(temp,&iCom_port_info->global_reg->control);
    TRACE(iCom_port_info,TRACE_START_PROC_A,0);
    break;
  case 1:
    temp = iCom_readl(&iCom_port_info->global_reg->control);
    temp = (temp & ~ICOM_CONTROL_STOP_B) | ICOM_CONTROL_START_B;
    iCom_writel(temp,&iCom_port_info->global_reg->control);
    TRACE(iCom_port_info,TRACE_START_PROC_B,0);
    break;
  case 2:
    temp = iCom_readl(&iCom_port_info->global_reg->control_2);
    temp = (temp & ~ICOM_CONTROL_STOP_C) | ICOM_CONTROL_START_C;
    iCom_writel(temp,&iCom_port_info->global_reg->control_2);
    TRACE(iCom_port_info,TRACE_START_PROC_C,0);
    break;
  case 3:
    temp = iCom_readl(&iCom_port_info->global_reg->control_2);
    temp = (temp & ~ICOM_CONTROL_STOP_D) | ICOM_CONTROL_START_D;
    iCom_writel(temp,&iCom_port_info->global_reg->control_2);
    TRACE(iCom_port_info,TRACE_START_PROC_D,0);
    break;
  default:
    printk("iCom:  ERROR: invalid port assignment\n");
  }
}

/*
 * irq = no lock
 */
static void loadCode (struct iCom_port *iCom_port_info)
{
  char               *iram_ptr;
  int                index;
  int                status = 0;
  char               *dram_ptr = (char *)iCom_port_info->dram;
  unsigned long int  temp;
  unsigned char      *new_page;
  unsigned char      cable_id;

  TRACE(iCom_port_info,TRACE_GET_MEM,0); /* this really gets memory for trace */
  TRACE(iCom_port_info,TRACE_LOAD_MEM,(u32)dram_ptr);

  /* Clear out any pending interrupts */
  iCom_writew(0x3FFF,(void *)iCom_port_info->int_reg);

  TRACE(iCom_port_info,TRACE_CLEAR_INTERRUPTS,0);

  /* Stop processor */
  stop_processor(iCom_port_info);

  /* Zero out DRAM */
  for (index = 0; index < 512; index++)
  {
    iCom_writeb(0x00,&dram_ptr[index]);
  }

  /* Load Call Setup into Adapter */
  iram_ptr = (char *)iCom_port_info->dram + ICOM_IRAM_OFFSET;
  for (index = 0; index < sizeof(callSetup); index++)
  {
    iCom_writeb(callSetup[index],&iram_ptr[index]);
  }

  /* Load Resident DCE portion of Adapter */
  iram_ptr = (char *) iCom_port_info->dram + ICOM_IRAM_OFFSET +
    ICOM_DCE_IRAM_OFFSET;

  /* Load the RV dce code */
  for (index = 0; index < sizeof(resRVdce); index++) {
    iCom_writeb(resRVdce[index],&iram_ptr[index]);
  }

  /* Set Hardware level */
  if ((iCom_adapter_info[iCom_port_info->adapter].version | ADAPTER_V2) == ADAPTER_V2) {
    iCom_writeb(V2_HARDWARE,&(iCom_port_info->dram->misc_flags));
  }

  /* Start the processor in Adapter */
  start_processor(iCom_port_info);

  iCom_writeb((HDLC_PPP_PURE_ASYNC | HDLC_FF_FILL),&(iCom_port_info->dram->HDLCConfigReg));
  iCom_writeb(0x04,&(iCom_port_info->dram->FlagFillIdleTimer)); /* 0.5 seconds */
  iCom_writeb(0x00,&(iCom_port_info->dram->CmdReg));
  iCom_writeb(0x10,&(iCom_port_info->dram->async_config3));
  iCom_writeb((ICOM_ACFG_DRIVE1 | ICOM_ACFG_NO_PARITY | ICOM_ACFG_8BPC | ICOM_ACFG_1STOP_BIT),&(iCom_port_info->dram->async_config2));

  /*Set up data in iCom DRAM to indicate where personality
   *code is located and its length.
   */
  new_page = (unsigned char *)kmalloc(4096,GFP_KERNEL | GFP_DMA);
  for (index = 0; index < sizeof(funcLoad); index++) {
    new_page[index] = funcLoad[index];
  }
  temp = pci_map_single(iCom_adapter_info[iCom_port_info->adapter].pci_dev,new_page,4096,PCI_DMA_BIDIRECTIONAL);

  iCom_writeb((char)(sizeof(funcLoad)/16),&iCom_port_info->dram->mac_length);
  iCom_writel(temp,&iCom_port_info->dram->mac_load_addr);

  /*Setting the syncReg to 0x80 causes adapter to start downloading
    the personality code into adapter instruction RAM.
    Once code is loaded, it will begin executing and, based on
    information provided above, will start DMAing data from
    shared memory to adapter DRAM.
  */
  iCom_writeb(START_DOWNLOAD,&iCom_port_info->dram->sync);

  /* Wait max 1 Sec for data download and processor to start */
  for (index = 0; index < 100; index++ ) {
      current->state = TASK_INTERRUPTIBLE;
      schedule_timeout(HZ/100);
      if (iCom_readb(&iCom_port_info->dram->misc_flags) & ICOM_HDW_ACTIVE) break;
  }

  if (index == 100) status = -1;

  pci_unmap_single(iCom_adapter_info[iCom_port_info->adapter].pci_dev,temp,4096,PCI_DMA_BIDIRECTIONAL);
  kfree(new_page);
  
  /*
   * check Cable ID
   */
  cable_id = iCom_readb(&iCom_port_info->dram->cable_id);
  if (cable_id & ICOM_CABLE_ID_VALID)
  {
      /* Get cable ID into the lower 4 bits (standard form) */
      cable_id = (cable_id & ICOM_CABLE_ID_MASK) >> 4;
      iCom_port_info->cable_id = cable_id;
  }
  else {
      iCom_port_info->cable_id = NO_CABLE;
  }

  if (status != 0)
  {
    /* Clear out any pending interrupts */
    iCom_writew(0x3FFF,(void *)iCom_port_info->int_reg);

    /* Turn off port */
    iCom_writeb(ICOM_DISABLE,&(iCom_port_info->dram->disable));

    /* Stop processor */
    stop_processor(iCom_port_info);

    /* Fail the port */
    if (cable_id != NO_CABLE) {
	printk("iCom:  Error, minor number %d port not opertional\n", iCom_port_info->minor_number);
	iCom_port_info->passed_diags = 0;
    }
  }
}

/*
 * This routine is called to set the port to match
 * the specified baud rate for a serial port.
 * irq = locked
 */
static void change_speed(struct iCom_port *iCom_port_info,
			 struct termios *old_termios, unsigned long flags)
{
  int	         baud;
  unsigned       cflag;
  int	         bits;
  char           new_config2;
  char           new_config3;
  char           tmp_byte;
  int            index;
  int            rcv_buff,xmit_buff;
  unsigned long int offset;

  TRACE(iCom_port_info,TRACE_CHANGE_SPEED | TRACE_TIME,jiffies);

  if (!iCom_port_info->tty || !iCom_port_info->tty->termios)
    return;
  cflag = iCom_port_info->tty->termios->c_cflag;

  new_config2 = ICOM_ACFG_DRIVE1;

  /* byte size and parity */
  switch (cflag & CSIZE) {
    case CS5: /* 5 bits/char */
      new_config2 |= ICOM_ACFG_5BPC;
      bits = 7;
      break;
    case CS6: /* 6 bits/char */
      new_config2 |= ICOM_ACFG_6BPC;
      bits = 8;
      break;
    case CS7: /* 7 bits/char */
      new_config2 |= ICOM_ACFG_7BPC;
      bits = 9;
      break;
    case CS8: /* 8 bits/char */
      new_config2 |= ICOM_ACFG_8BPC;
      bits = 10;
      break;
    default:  bits = 10;  break;
  }
  if (cflag & CSTOPB) {
    /* 2 stop bits */
    new_config2 |= ICOM_ACFG_2STOP_BIT;
    bits++;
  }
  if (cflag & PARENB) {
    /* parity bit enabled */
    new_config2 |= ICOM_ACFG_PARITY_ENAB;
    TRACE(iCom_port_info, TRACE_PARENB,0);
    bits++;
  }
  if (cflag & PARODD) {
    /* odd parity */
    new_config2 |= ICOM_ACFG_PARITY_ODD;
    TRACE(iCom_port_info, TRACE_PARODD,0);
  }

  /* Determine divisor based on baud rate */
  baud = tty_get_baud_rate(iCom_port_info->tty);
  if (!baud)
    baud = 9600;	/* B0 transition handled in rs_set_termios */

  for (index = 0; index < BAUD_TABLE_LIMIT; index++) {
    if (icom_acfg_baud[index] == baud) {
      new_config3 = index;
      break;
    }
  }

  iCom_port_info->timeout = XMIT_BUFF_SZ*HZ*bits/baud;
  iCom_port_info->timeout += HZ/50;		/* Add .02 seconds of slop */

  /* CTS flow control flag and modem status interrupts */
  if (cflag & CRTSCTS) {
    iCom_port_info->flags |= ASYNC_CTS_FLOW;
    tmp_byte = iCom_readb(&(iCom_port_info->dram->HDLCConfigReg));
    tmp_byte |= HDLC_HDW_FLOW;
    iCom_writeb(tmp_byte, &(iCom_port_info->dram->HDLCConfigReg));
  } else {
    iCom_port_info->flags &= ~ASYNC_CTS_FLOW;
    tmp_byte = iCom_readb(&(iCom_port_info->dram->HDLCConfigReg));
    tmp_byte &= ~HDLC_HDW_FLOW;
    iCom_writeb(tmp_byte, &(iCom_port_info->dram->HDLCConfigReg));
  }
  if (cflag & CLOCAL)
    iCom_port_info->flags &= ~ASYNC_CHECK_CD;
  else {
    iCom_port_info->flags |= ASYNC_CHECK_CD;
  }

  /*
   * Set up parity check flag
   */
  iCom_port_info->read_status_mask = SA_FLAGS_OVERRUN | SA_FL_RCV_DONE;
  if (I_INPCK(iCom_port_info->tty))
    iCom_port_info->read_status_mask |= SA_FLAGS_FRAME_ERROR | SA_FLAGS_PARITY_ERROR;

  if (I_BRKINT(iCom_port_info->tty) || I_PARMRK(iCom_port_info->tty))
    iCom_port_info->read_status_mask |= SA_FLAGS_BREAK_DET;

  /*
   * Characters to ignore
   */
  iCom_port_info->ignore_status_mask = 0;
  if (I_IGNPAR(iCom_port_info->tty))
    iCom_port_info->ignore_status_mask |= SA_FLAGS_PARITY_ERROR | SA_FLAGS_FRAME_ERROR;
  if (I_IGNBRK(iCom_port_info->tty)) {
    iCom_port_info->ignore_status_mask |= SA_FLAGS_BREAK_DET;
    /*
     * If we're ignore parity and break indicators, ignore 
     * overruns too.  (For real raw support).
     */
    if (I_IGNPAR(iCom_port_info->tty))
      iCom_port_info->ignore_status_mask |= SA_FLAGS_OVERRUN;
  }

  /*
   * !!! ignore all characters if CREAD is not set
   */
  if ((cflag & CREAD) == 0)
    iCom_port_info->ignore_status_mask |= SA_FL_RCV_DONE;

  /* Turn off Receiver to prepare for reset */
  iCom_writeb(CMD_RCV_DISABLE,&iCom_port_info->dram->CmdReg);

  spin_unlock_irqrestore(&iComlock,flags);
  for (index = 0; index < 10; index++) {
    /* Wait 0.1 Sec for receive operations to complete*/
    current->state = TASK_INTERRUPTIBLE;
    schedule_timeout(HZ/10);

    if (iCom_readb(&iCom_port_info->dram->PrevCmdReg) == 0x00) {
      break;
    }
  }
  spin_lock_irqsave(&iComlock, flags);

  /* clear all current buffers of data */
  for (rcv_buff = 0; rcv_buff < NUM_RBUFFS; rcv_buff++) {
    iCom_port_info->statStg->rcv[rcv_buff].flags = 0;
    iCom_port_info->statStg->rcv[rcv_buff].leLength = 0;
    iCom_port_info->statStg->rcv[rcv_buff].WorkingLength = (unsigned short int)cpu_to_le16(RCV_BUFF_SZ);
  }

  for (xmit_buff = 0; xmit_buff < NUM_XBUFFS;  xmit_buff++) {
    iCom_port_info->statStg->xmit[xmit_buff].flags = 0;
  }

  /* activate changes and start xmit and receiver here */
  /* Enable the receiver */
  iCom_writeb(new_config3,&(iCom_port_info->dram->async_config3));
  iCom_writeb(new_config2,&(iCom_port_info->dram->async_config2));
  tmp_byte = iCom_readb(&(iCom_port_info->dram->HDLCConfigReg));
  tmp_byte |= HDLC_PPP_PURE_ASYNC | HDLC_FF_FILL;
  iCom_writeb(tmp_byte,&(iCom_port_info->dram->HDLCConfigReg));
  iCom_writeb(0x04, &(iCom_port_info->dram->FlagFillIdleTimer)); /* 0.5 seconds */
  iCom_writeb(0xFF, &(iCom_port_info->dram->ier)); /* enable modem signal interrupts */

  /* reset processor */
  iCom_writeb(CMD_RESTART,&iCom_port_info->dram->CmdReg);
  spin_unlock_irqrestore(&iComlock, flags);
  for (index = 0; index < 10; index++) {
    /* Wait for reset operation */
    current->state = TASK_INTERRUPTIBLE;
    schedule_timeout(HZ/10);

    if (iCom_readb(&iCom_port_info->dram->CmdReg) == 0x00) {
      break;
    }
  }
  spin_lock_irqsave(&iComlock, flags);

  /* Enable Transmitter and Reciever */
  offset = (unsigned long int)&iCom_port_info->statStg->rcv[0] - (unsigned long int)iCom_port_info->statStg;
  iCom_writel(iCom_port_info->statStg_pci + offset,&iCom_port_info->dram->RcvStatusAddr);
  iCom_port_info->next_rcv = 0;
  iCom_port_info->put_length = 0;
  *iCom_port_info->xmitRestart = 0;
  iCom_writel(iCom_port_info->xmitRestart_pci,&iCom_port_info->dram->XmitStatusAddr);
  TRACE(iCom_port_info,TRACE_XR_ENAB,0);
  iCom_writeb(CMD_XMIT_RCV_ENABLE,&iCom_port_info->dram->CmdReg);
}

static int block_til_ready(struct tty_struct *tty, struct file * filp,
                           struct iCom_port *iCom_port_info, long int flags)
{
  DECLARE_WAITQUEUE(wait, current);
  int		retval;
  int		do_clocal = 0, extra_count = 0;

  /*
   * If the device is in the middle of being closed, then block
   * until it's done, and then try again.
   */
  if (tty_hung_up_p(filp) ||
      (iCom_port_info->flags & ASYNC_CLOSING)) {
      if (iCom_port_info->flags & ASYNC_CLOSING) {
	  spin_unlock_irqrestore(&iComlock,flags);
	  interruptible_sleep_on(&iCom_port_info->close_wait);
	  spin_lock_irqsave(&iComlock,flags);
      }    
#ifdef SERIAL_DO_RESTART
    return ((iCom_port_info->flags & ASYNC_HUP_NOTIFY) ?
            -EAGAIN : -ERESTARTSYS);
#else
    return -EAGAIN;
#endif
  }

  /*
   * If this is a callout device, then just make sure the normal
   * device isn't being used.
   */
  if (tty->driver.subtype == SERIAL_TYPE_CALLOUT) {
    if (iCom_port_info->flags & ASYNC_NORMAL_ACTIVE)
      return -EBUSY;
    if ((iCom_port_info->flags & ASYNC_CALLOUT_ACTIVE) &&
        (iCom_port_info->flags & ASYNC_SESSION_LOCKOUT) &&
        (iCom_port_info->session != current->session))
      return -EBUSY;
    if ((iCom_port_info->flags & ASYNC_CALLOUT_ACTIVE) &&
        (iCom_port_info->flags & ASYNC_PGRP_LOCKOUT) &&
        (iCom_port_info->pgrp != current->pgrp))
      return -EBUSY;
    iCom_port_info->flags |= ASYNC_CALLOUT_ACTIVE;
    return 0;
  }

  /*
   * If non-blocking mode is set, or the port is not enabled,
   * then make the check up front and then exit.
   */
  if ((filp->f_flags & O_NONBLOCK) ||
      (tty->flags & (1 << TTY_IO_ERROR))) {
    if (iCom_port_info->flags & ASYNC_CALLOUT_ACTIVE)
      return -EBUSY;
    iCom_port_info->flags |= ASYNC_NORMAL_ACTIVE;
    return 0;
  }

  if (iCom_port_info->flags & ASYNC_CALLOUT_ACTIVE) {
    if (iCom_port_info->normal_termios.c_cflag & CLOCAL)
      do_clocal = 1;
  } else {
    if (tty->termios->c_cflag & CLOCAL)
      do_clocal = 1;
  }

  /*
   * Block waiting for the carrier detect and the line to become
   * free (i.e., not in use by the callout).  While we are in
   * this loop, open_active_count is dropped by one, so that
   * rs_close() knows when to free things.  We restore it upon
   * exit, either normal or abnormal.
   */
  retval = 0;
  add_wait_queue(&iCom_port_info->open_wait, &wait);

  if (!tty_hung_up_p(filp)) {
    extra_count = 1;
    iCom_port_info->open_active_count--;
  }
  iCom_port_info->blocked_open++;
  while (1) {
    if (!(iCom_port_info->flags & ASYNC_CALLOUT_ACTIVE) &&
        (tty->termios->c_cflag & CBAUD)) {
      /* raise DTR and RTS */
      TRACE(iCom_port_info,TRACE_RAISE_DTR_RTS,0);
      iCom_writeb(0xC0,&iCom_port_info->dram->osr);
    }
    current->state = TASK_INTERRUPTIBLE;
    if (tty_hung_up_p(filp) ||
        !(iCom_port_info->flags & ASYNC_INITIALIZED)) {
#ifdef SERIAL_DO_RESTART
      if (iCom_port_info->flags & ASYNC_HUP_NOTIFY)
        retval = -EAGAIN;
      else
        retval = -ERESTARTSYS;	
#else
      retval = -EAGAIN;
#endif
      break;
    }

    if (!(iCom_port_info->flags & ASYNC_CALLOUT_ACTIVE) &&
        !(iCom_port_info->flags & ASYNC_CLOSING) &&
        (do_clocal || (iCom_readb(&iCom_port_info->dram->isr) & ICOM_DCD))) /* Carrier Detect */
      break;
    if (signal_pending(current)) {
      retval = -ERESTARTSYS;
      break;
    }
    spin_unlock_irqrestore(&iComlock,flags);
    printk("iCom:  WAIT for CD\n");
    schedule();
    spin_lock_irqsave(&iComlock,flags);
  }
  current->state = TASK_RUNNING;
  remove_wait_queue(&iCom_port_info->open_wait, &wait);
  if (extra_count)
    iCom_port_info->open_active_count++;
  iCom_port_info->blocked_open--;

  if (retval)
    return retval;
  iCom_port_info->flags |= ASYNC_NORMAL_ACTIVE;

  return 0;
}

/*
  irq = locked
*/
static int startup(struct iCom_port *iCom_port_info, unsigned long flags)
{
  int	            retval=0;
  unsigned long int temp;
  unsigned char     cable_id, raw_cable_id;

  TRACE(iCom_port_info,TRACE_STARTUP,0);

  if (iCom_port_info->flags & ASYNC_INITIALIZED) {
    goto errout;
  }

  if (iCom_port_info->dram == 0x00000000) {
      /* should NEVER be zero */
      return -ENODEV;
  }

  /*
   * check Cable ID
   */
  raw_cable_id = iCom_readb(&iCom_port_info->dram->cable_id);
  TRACE(iCom_port_info,TRACE_CABLE_ID,raw_cable_id);

  /* Get cable ID into the lower 4 bits (standard form) */
  cable_id = (raw_cable_id & ICOM_CABLE_ID_MASK) >> 4;

  /* Check Cable ID is RS232 */
  if (!(raw_cable_id & ICOM_CABLE_ID_VALID) || (cable_id != RS232_CABLE))
  {
      /* reload adapter code, pick up any potential changes in cable id */
      if (iCom_port_info->load_in_progress) {
	  printk("iCom:  Unusable Port, minor number (%d) currently being initialized by another task\n", iCom_port_info->minor_number);
	  return -ENODEV;
      }
      iCom_port_info->load_in_progress = 1;
      spin_unlock_irqrestore(&iComlock,flags);
      loadCode(iCom_port_info);
      spin_lock_irqsave(&iComlock, flags);
      iCom_port_info->load_in_progress = 0;

      /* still no sign of RS232, error out */
      if (iCom_port_info->cable_id != RS232_CABLE) {
	  printk("iCom:  Unusable Port, minor number (%d) incorrect cable attached, only RS232 cables permitted\n",iCom_port_info->minor_number);
	  return -ENODEV;
      }
  }

  /*
   * set appropriate modem signals
   */
  if (iCom_port_info->tty->termios->c_cflag & CBAUD) {
    /* raise DTR and RTS */
    TRACE(iCom_port_info,TRACE_RAISE_DTR_RTS,0);
    iCom_writeb(0xC0,&iCom_port_info->dram->osr);
  }

  /*
   * Finally, clear and  enable interrupts
   */
  switch (iCom_port_info->port) {
  case 0:
    /* Clear out any pending interrupts */
    iCom_writew(0x00FF,(void *)iCom_port_info->int_reg);

    /* Enable interrupts for first port */
    TRACE(iCom_port_info,TRACE_ENABLE_INTERRUPTS_PA,0);
    temp = iCom_readl(&iCom_port_info->global_reg->int_mask);
    iCom_writel((temp & ~ICOM_INT_MASK_PRC_A),&iCom_port_info->global_reg->int_mask);
    break;
  case 1:
    /* Clear out any pending interrupts */
    iCom_writew(0x3F00,(void *)iCom_port_info->int_reg);

    /* Enable interrupts for second port */
    TRACE(iCom_port_info,TRACE_ENABLE_INTERRUPTS_PB,0);
    temp = iCom_readl(&iCom_port_info->global_reg->int_mask);
    iCom_writel((temp & ~ICOM_INT_MASK_PRC_B),&iCom_port_info->global_reg->int_mask);
    break;
  case 2:
    /* Clear out any pending interrupts */
    iCom_writew(0x00FF,(void *)iCom_port_info->int_reg);

    /* Enable interrupts for first port */
    TRACE(iCom_port_info,TRACE_ENABLE_INTERRUPTS_PC,0);
    temp = iCom_readl(&iCom_port_info->global_reg->int_mask_2);
    iCom_writel((temp & ~ICOM_INT_MASK_PRC_C),&iCom_port_info->global_reg->int_mask_2);
    break;
  case 3:
    /* Clear out any pending interrupts */
    iCom_writew(0x3F00,(void *)iCom_port_info->int_reg);

    /* Enable interrupts for second port */
    TRACE(iCom_port_info,TRACE_ENABLE_INTERRUPTS_PD,0);
    temp = iCom_readl(&iCom_port_info->global_reg->int_mask_2);
    iCom_writel((temp & ~ICOM_INT_MASK_PRC_D),&iCom_port_info->global_reg->int_mask_2);
    break;
  default:
    printk("iCom:  ERROR:  Invalid port defined\n");
  }

  if (iCom_port_info->tty)
    clear_bit(TTY_IO_ERROR, &iCom_port_info->tty->flags);

  /*
   * Set up the tty->alt_speed kludge
   */
  if (iCom_port_info->tty) {
    if ((iCom_port_info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_HI)
      iCom_port_info->tty->alt_speed = 57600;
    if ((iCom_port_info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_VHI)
      iCom_port_info->tty->alt_speed = 115200;
    if ((iCom_port_info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_SHI)
      iCom_port_info->tty->alt_speed = 230400;
    if ((iCom_port_info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_WARP)
      iCom_port_info->tty->alt_speed = 460800;
  }

  /*
   * and set the speed of the serial port
   */
  change_speed(iCom_port_info, 0, flags);

  iCom_port_info->flags |= ASYNC_INITIALIZED;
  return 0;

  errout:
  return retval;
}

/*
 * This routine will shutdown a serial port; interrupts are disabled, and
 * DTR is dropped if the hangup on close termio flag is on.
 *
 * irq = locked
 */
static void shutdown(struct iCom_port * iCom_port_info)
{
  unsigned long int temp;
  unsigned char     cmdReg;

  TRACE(iCom_port_info,TRACE_SHUTDOWN | TRACE_TIME,jiffies);

  if (!(iCom_port_info->flags & ASYNC_INITIALIZED))
    return;

  /*
   * clear delta_msr_wait queue to avoid mem leaks: we may free the irq
   * here so the queue might never be waken up
   */
  wake_up_interruptible(&iCom_port_info->delta_msr_wait);

  /*
   * disable all interrupts
   */
  switch (iCom_port_info->port) {
  case 0:
    TRACE(iCom_port_info,TRACE_DIS_INTERRUPTS_PA,0);
    temp = iCom_readl(&iCom_port_info->global_reg->int_mask);
    iCom_writel((temp | ICOM_INT_MASK_PRC_A),&iCom_port_info->global_reg->int_mask);
    break;
  case 1:
    TRACE(iCom_port_info,TRACE_DIS_INTERRUPTS_PB,0);
    temp = iCom_readl(&iCom_port_info->global_reg->int_mask);
    iCom_writel((temp | ICOM_INT_MASK_PRC_B),&iCom_port_info->global_reg->int_mask);
    break;
  case 2:
    TRACE(iCom_port_info,TRACE_DIS_INTERRUPTS_PC,0);
    temp = iCom_readl(&iCom_port_info->global_reg->int_mask_2);
    iCom_writel((temp | ICOM_INT_MASK_PRC_C),&iCom_port_info->global_reg->int_mask_2);
    break;
  case 3:
    TRACE(iCom_port_info,TRACE_DIS_INTERRUPTS_PD,0);
    temp = iCom_readl(&iCom_port_info->global_reg->int_mask_2);
    iCom_writel((temp | ICOM_INT_MASK_PRC_D),&iCom_port_info->global_reg->int_mask_2);
    break;
  default:
    printk("iCom:  ERROR:  Invalid port assignment\n");
  }

  /*
   * disable break condition
   */
  cmdReg = iCom_readb(&iCom_port_info->dram->CmdReg);
  if ((cmdReg | CMD_SND_BREAK) == CMD_SND_BREAK) {
    iCom_writeb(cmdReg & ~CMD_SND_BREAK,&iCom_port_info->dram->CmdReg);
  }

  if (!iCom_port_info->tty || (iCom_port_info->tty->termios->c_cflag & HUPCL)) {
    /* drop DTR and RTS */
    TRACE(iCom_port_info,TRACE_DROP_DTR_RTS,0);
    iCom_writeb(0x00,&iCom_port_info->dram->osr);
  }
  
  if (iCom_port_info->tty)
    set_bit(TTY_IO_ERROR, &iCom_port_info->tty->flags);

  iCom_port_info->flags &= ~ASYNC_INITIALIZED;
}

/*
   Primary interface routines to iCom Driver
*/
static int iCom_open(struct tty_struct * tty, struct file * filp)
{
  int               line;
  int               adapter_entry;
  int               port_entry;
  struct iCom_port *iCom_port_info;
  int               retval;
  unsigned long     flags;

  /*
      Minor Number
      _ _ _ _ b (lower nibble)
      ___ ___
       |   |
       |   - port number (lowest 2 bits is port identifier)
       - adapter number (remaining higher order bits identify adapter #)
  */

  MOD_INC_USE_COUNT;
  line = MINOR(tty->device) - tty->driver.minor_start;
  if ((line < 0) || (line >= NR_PORTS)) {
      printk("iCom:  Invalid Minor number (%d) on open\n",MINOR(tty->device));
      MOD_DEC_USE_COUNT;
      return -ENODEV;
  }

  adapter_entry = (line & 0xFFFC) >> 2; /* shift adapter # into position */
  port_entry = line & 0x0003; /* mask of port number */

  if ((port_entry == 1) &&
      (iCom_adapter_info[adapter_entry].version == ADAPTER_V2) &&
      (iCom_adapter_info[adapter_entry].subsystem_id != FOUR_PORT_MODEL)) {
      port_entry = 2;
  }
  iCom_port_info = &iCom_adapter_info[adapter_entry].port_info[port_entry];

  if (iCom_port_info->status != ICOM_PORT_ACTIVE) {
      printk("iCom:  Unusable Port, minor number (%d) invalid minor number on open\n",MINOR(tty->device));
      MOD_DEC_USE_COUNT;
      return -ENODEV;
  }

  if (!iCom_port_info->passed_diags) {
      printk("iCom:  Unusable Port, minor number (%d) failed diagnostic checks, see previously logged messages\n",MINOR(tty->device));
      MOD_DEC_USE_COUNT;
      return -ENODEV;
  }

  spin_lock_irqsave(&iComlock,flags);
  TRACE(iCom_port_info,TRACE_DEVICE_NUMB,tty->device);
  tty->driver_data = iCom_port_info;
  iCom_port_info->tty = tty;
  iCom_port_info->open_active_count++;

  /*
   * If the port is the middle of closing, bail out now
   */
  if (tty_hung_up_p(filp) ||
      (iCom_port_info->flags & ASYNC_CLOSING)) {

      MOD_DEC_USE_COUNT;
      iCom_port_info->open_active_count--;
      spin_unlock_irqrestore(&iComlock,flags);
      if (iCom_port_info->flags & ASYNC_CLOSING)
	  interruptible_sleep_on(&iCom_port_info->close_wait);
#ifdef SERIAL_DO_RESTART
      return ((iCom_port_info->flags & ASYNC_HUP_NOTIFY) ?
            -EAGAIN : -ERESTARTSYS);
#else
      return -EAGAIN;
#endif
  }

  /*
   * Start up serial port
   */
  retval = startup(iCom_port_info, flags);
  if (retval) {
    /* reset open variables */
    iCom_port_info->open_active_count--;
    TRACE(iCom_port_info,TRACE_STARTUP_ERROR,0);
    spin_unlock_irqrestore(&iComlock,flags);
    return retval;
  }

  retval = block_til_ready(tty, filp, iCom_port_info, flags);
  if (retval) {
    MOD_DEC_USE_COUNT;
    iCom_port_info->open_active_count--;
    spin_unlock_irqrestore(&iComlock,flags);
    return retval;
  }

  if ((iCom_port_info->open_active_count == 1) &&
      (iCom_port_info->flags & ASYNC_SPLIT_TERMIOS)) {
    if (tty->driver.subtype == SERIAL_TYPE_NORMAL)
      *tty->termios = iCom_port_info->normal_termios;
    else 
      *tty->termios = iCom_port_info->callout_termios;
    change_speed(iCom_port_info, 0, flags);
  }

  iCom_port_info->session = current->session;
  iCom_port_info->pgrp = current->pgrp;

  spin_unlock_irqrestore(&iComlock,flags);
  return 0;
}

/*
 * ------------------------------------------------------------
 * iCom_close()
 * 
 * This routine is called when the serial port gets closed.  First, we
 * wait for the last remaining data to be sent.  Then, we unlink its
 * async structure from the interrupt chain if necessary.
 * ------------------------------------------------------------
 */
static void iCom_close(struct tty_struct * tty, struct file * filp)
{
  struct iCom_port *iCom_port_info;
  unsigned long     flags;
  unsigned char     cmdReg;


  if (!tty) {
    printk("iCom:  iCom_close - no tty\n");
    return;
  }

  iCom_port_info = (struct iCom_port *)tty->driver_data;
  if (!iCom_port_info) {
    printk("iCom:  iCom_close - no tty->driver_data\n");
    return;
  }
  
  TRACE(iCom_port_info,TRACE_CLOSE,0);
  spin_lock_irqsave(&iComlock,flags);

  if (tty_hung_up_p(filp)) {
    TRACE(iCom_port_info,TRACE_CLOSE_HANGUP,0);
    MOD_DEC_USE_COUNT;
    spin_unlock_irqrestore(&iComlock,flags);
    return;
  }

  if ((tty->count == 1) && (iCom_port_info->open_active_count != 1)) {
    /*
     * Uh, oh.  tty->count is 1, which means that the tty
     * structure will be freed.  open_active_count should always
     * be one in these conditions.  If it's greater than
     * one, we've got real problems, since it means the
     * serial port won't be shutdown.
     */
    iCom_port_info->open_active_count = 1;
  }

  if (--iCom_port_info->open_active_count < 0) {
    iCom_port_info->open_active_count = 0;
  }

  if (iCom_port_info->open_active_count) {
    TRACE(iCom_port_info,TRACE_OPEN_ACTIVE,0);
    MOD_DEC_USE_COUNT;
    spin_unlock_irqrestore(&iComlock,flags);
    return;
  }
  iCom_port_info->flags |= ASYNC_CLOSING;

  /*
   * Save the termios structure, since this port may have
   * separate termios for callout and dialin.
   */
  if (iCom_port_info->flags & ASYNC_NORMAL_ACTIVE)
    iCom_port_info->normal_termios = *tty->termios;
  if (iCom_port_info->flags & ASYNC_CALLOUT_ACTIVE)
    iCom_port_info->callout_termios = *tty->termios;

  /*
   * Now we wait for the transmit buffer to clear; and we notify 
   * the line discipline to only process XON/XOFF characters.
   */
  tty->closing = 1;
  if (iCom_port_info->closing_wait != ASYNC_CLOSING_WAIT_NONE) {
    spin_unlock_irqrestore(&iComlock,flags);
    tty_wait_until_sent(tty, iCom_port_info->closing_wait);
    spin_lock_irqsave(&iComlock,flags);
  }

  /*
   * At this point we stop accepting input.  To do this, we
   * disable the receive line status interrupts, and tell the
   * interrupt driver to stop checking the data ready bit in the
   * line status register.
   */
  if (iCom_port_info->flags & ASYNC_INITIALIZED) {
    cmdReg = iCom_readb(&iCom_port_info->dram->CmdReg);
    iCom_writeb(cmdReg & (unsigned char)~CMD_RCV_ENABLE,&iCom_port_info->dram->CmdReg);

   /*
     * Before we drop DTR, make sure the UART transmitter
     * has completely drained; this is especially
     * important if there is a transmit FIFO!
     */
    spin_unlock_irqrestore(&iComlock,flags);
    iCom_wait_until_sent(tty, iCom_port_info->timeout);
    spin_lock_irqsave(&iComlock,flags);
  }

  shutdown(iCom_port_info);

  spin_unlock_irqrestore(&iComlock,flags);
  if (tty->driver.flush_buffer)
    tty->driver.flush_buffer(tty);
  if (tty->ldisc.flush_buffer)
    tty->ldisc.flush_buffer(tty);
  spin_lock_irqsave(&iComlock,flags);
  tty->closing = 0;
  iCom_port_info->event = 0;
  iCom_port_info->tty = 0;

  if (iCom_port_info->blocked_open) {
    if (iCom_port_info->close_delay) {
	current->state = TASK_INTERRUPTIBLE;
	spin_unlock_irqrestore(&iComlock,flags);
	schedule_timeout(iCom_port_info->close_delay);
	spin_lock_irqsave(&iComlock,flags);
    }
    wake_up_interruptible(&iCom_port_info->open_wait);
  }
  iCom_port_info->flags &= ~(ASYNC_NORMAL_ACTIVE|ASYNC_CALLOUT_ACTIVE|
                   ASYNC_CLOSING);

  wake_up_interruptible(&iCom_port_info->close_wait);

  MOD_DEC_USE_COUNT;
  spin_unlock_irqrestore(&iComlock,flags);
}

static int iCom_write(struct tty_struct * tty, int from_user,
			const unsigned char * buf, int count)
{
  struct iCom_port  *iCom_port_info = (struct iCom_port *)tty->driver_data;
  unsigned long     data_count = count;
  unsigned char     *data;
  unsigned char     cmdReg;
  unsigned long int offset;
  unsigned long int flags;


  if (!tty) {
    printk("iCom:  iCom_write - no tty\n");
    return 0;
  }

  spin_lock_irqsave(&iComlock,flags);

  iCom_port_info = (struct iCom_port *)tty->driver_data;
  TRACE(iCom_port_info,TRACE_WRITE | TRACE_TIME,jiffies);

  if (cpu_to_le16(iCom_port_info->statStg->xmit[0].flags) & SA_FLAGS_READY_TO_XMIT) {
      TRACE(iCom_port_info,TRACE_WRITE_FULL,0);
      spin_unlock_irqrestore(&iComlock,flags);
      return 0;
  }

  if (data_count > XMIT_BUFF_SZ)
    data_count = XMIT_BUFF_SZ;

  if (from_user) {
    data_count -= copy_from_user(iCom_port_info->xmit_buf, buf, data_count);
    if (!data_count) {
      TRACE(iCom_port_info,TRACE_WRITE_NODATA,0);
      spin_unlock_irqrestore(&iComlock,flags);
      return -EFAULT;
    }
  } else {
    memcpy(iCom_port_info->xmit_buf, buf, data_count);
  }

  data = iCom_port_info->xmit_buf;

  if (data_count) {
      iCom_port_info->statStg->xmit[0].flags = (unsigned short int)cpu_to_le16(SA_FLAGS_READY_TO_XMIT);
      iCom_port_info->statStg->xmit[0].leLength = (unsigned short int)cpu_to_le16(data_count);
      offset = (unsigned long int)&iCom_port_info->statStg->xmit[0] - (unsigned long int)iCom_port_info->statStg;
      *iCom_port_info->xmitRestart = cpu_to_le32(iCom_port_info->statStg_pci + offset);
      cmdReg = iCom_readb(&iCom_port_info->dram->CmdReg);
      iCom_writeb(cmdReg | CMD_XMIT_RCV_ENABLE,&iCom_port_info->dram->CmdReg);
      iCom_writeb(START_XMIT,&iCom_port_info->dram->StartXmitCmd);
      TRACE(iCom_port_info,TRACE_WRITE_START,data_count);
  }

  spin_unlock_irqrestore(&iComlock,flags);

  return data_count;
}

static void iCom_put_char(struct tty_struct * tty, unsigned char ch)
{
  /* iCom_put_char adds the character to the current buffer, the
   * data is not actually sent until iCom_flush_chars is called.
   * Per definition iCom_flush_chars MUST be called after
   * iCom_put_char
   */

  unsigned char     *data;
  struct iCom_port  *iCom_port_info = (struct iCom_port *)tty->driver_data;
  unsigned long int flags;

  spin_lock_irqsave(&iComlock,flags);
  TRACE(iCom_port_info,TRACE_PUT_CHAR, ch);

  if (cpu_to_le16(iCom_port_info->statStg->xmit[0].flags) & SA_FLAGS_READY_TO_XMIT) {
      TRACE(iCom_port_info,TRACE_PUT_FULL,0);
      spin_unlock_irqrestore(&iComlock,flags);
      return;
  }

  data = iCom_port_info->xmit_buf;
  data[iCom_port_info->put_length] = ch;

  if (!tty->stopped && !tty->hw_stopped) {
    iCom_port_info->put_length++;
  }

  spin_unlock_irqrestore(&iComlock,flags);
}

static void iCom_flush_chars(struct tty_struct * tty)
{
  struct iCom_port  *iCom_port_info = (struct iCom_port *)tty->driver_data;
  unsigned char     cmdReg;
  unsigned long int offset;
  unsigned long int flags;

  spin_lock_irqsave(&iComlock,flags);
  TRACE(iCom_port_info,TRACE_FLUSH_CHAR | TRACE_TIME,jiffies);
  if (iCom_port_info->put_length) {
      TRACE(iCom_port_info,TRACE_START_FLUSH,iCom_port_info->put_length);
      iCom_port_info->statStg->xmit[0].flags = (unsigned short int)cpu_to_le16(SA_FLAGS_READY_TO_XMIT);
      iCom_port_info->statStg->xmit[0].leLength = (unsigned short int)cpu_to_le16(iCom_port_info->put_length);
      offset = (unsigned long int)&iCom_port_info->statStg->xmit[0] - (unsigned long int)iCom_port_info->statStg;
      *iCom_port_info->xmitRestart = cpu_to_le32(iCom_port_info->statStg_pci + offset);
      cmdReg = iCom_readb(&iCom_port_info->dram->CmdReg);
      iCom_writeb(cmdReg | CMD_XMIT_RCV_ENABLE,&iCom_port_info->dram->CmdReg);
      iCom_writeb(START_XMIT,&iCom_port_info->dram->StartXmitCmd);
  }
  iCom_port_info->put_length = 0;
  spin_unlock_irqrestore(&iComlock,flags);
}

static int iCom_write_room(struct tty_struct * tty)
{
  int bytes_avail;
  struct iCom_port *iCom_port_info = tty->driver_data;
  
  if (cpu_to_le16(iCom_port_info->statStg->xmit[0].flags) & SA_FLAGS_READY_TO_XMIT)
      bytes_avail = 0;
  else
      bytes_avail = XMIT_BUFF_SZ;

  TRACE(iCom_port_info,TRACE_WRITE_ROOM,bytes_avail);
  return bytes_avail;
}

static int iCom_chars_in_buffer(struct tty_struct * tty)
{
  unsigned long int dram;
  struct iCom_port *iCom_port_info = (struct iCom_port *)tty->driver_data;
  int number_remaining = 0;

  TRACE(iCom_port_info,TRACE_CHARS_IN_BUFF,0);
  if (cpu_to_le16(iCom_port_info->statStg->xmit[0].flags) & SA_FLAGS_READY_TO_XMIT) {
      dram = (unsigned long int)iCom_port_info->dram;
      number_remaining = iCom_readw((void *)(dram + 0x168));
      TRACE(iCom_port_info,TRACE_CHARS_REMAIN,number_remaining);
  }
  return number_remaining;
}

static int get_modem_info(struct iCom_port * iCom_port_info, unsigned int *value)
{
  unsigned char status,control;
  unsigned int result;

  TRACE(iCom_port_info,TRACE_GET_MODEM,0);

  status = iCom_readb(&iCom_port_info->dram->isr);
  control = iCom_readb(&iCom_port_info->dram->osr);

  result =  ((control & 0x40) ? TIOCM_RTS : 0)
    | ((control & ICOM_DTR) ? TIOCM_DTR : 0)
    | ((status  & ICOM_DCD) ? TIOCM_CAR : 0)
    | ((status  & ICOM_RI ) ? TIOCM_RNG : 0)
    | ((status  & ICOM_DSR) ? TIOCM_DSR : 0)
    | ((status  & ICOM_CTS) ? TIOCM_CTS : 0);
  return put_user(result,value);
}

static int set_modem_info(struct iCom_port * iCom_port_info, unsigned int cmd,
			  unsigned int *value)
{
  int error;
  unsigned int arg;
  unsigned char local_osr;

  TRACE(iCom_port_info,TRACE_SET_MODEM,0);
  local_osr = iCom_readb(&iCom_port_info->dram->osr);

  error = get_user(arg, value);
  if (error)
    return error;
  switch (cmd) {
    case TIOCMBIS: 
      if (arg & TIOCM_RTS) {
	TRACE(iCom_port_info,TRACE_RAISE_RTS,0);
        local_osr |= ICOM_RTS;
      }
      if (arg & TIOCM_DTR) {
	TRACE(iCom_port_info,TRACE_RAISE_DTR,0);
        local_osr |= ICOM_DTR;
      }
      break;
    case TIOCMBIC:
      if (arg & TIOCM_RTS) {
        TRACE(iCom_port_info,TRACE_LOWER_RTS,0);
        local_osr &= ~ICOM_RTS;
      }
      if (arg & TIOCM_DTR) {
	TRACE(iCom_port_info,TRACE_LOWER_DTR,0);
        local_osr &= ~ICOM_DTR;
      }
      break;
    case TIOCMSET:
      local_osr = ((local_osr & ~(ICOM_RTS | ICOM_DTR))
                   | ((arg & TIOCM_RTS) ? ICOM_RTS : 0)
                   | ((arg & TIOCM_DTR) ? ICOM_DTR : 0));
      break;
    default:
      return -EINVAL;
  }

  iCom_writeb(local_osr,&iCom_port_info->dram->osr);
  return 0;
}

static int get_serial_info(struct iCom_port * iCom_port_info,
			   struct serial_struct * retinfo)
{
  struct serial_struct tmp;

  TRACE(iCom_port_info,TRACE_GET_SERIAL,0);

  if (!retinfo)
    return -EFAULT;
  memset(&tmp, 0, sizeof(tmp));
  tmp.type = 0x00; /* device specific, PORT_UNKNOWN */
  tmp.line = iCom_port_info->adapter; /* adapter number */
  tmp.port = iCom_port_info->port; /* port number on adapter */
  tmp.irq = iCom_adapter_info[iCom_port_info->adapter].irq_number;
  tmp.flags = iCom_port_info->flags;
  tmp.xmit_fifo_size = XMIT_BUFF_SZ;
  tmp.baud_base = 0x00; /* device specific */
  tmp.close_delay = iCom_port_info->close_delay;
  tmp.closing_wait = iCom_port_info->closing_wait;
  tmp.custom_divisor = 0x00;  /* device specific */
  tmp.hub6 = 0x00; /* device specific */
  if (copy_to_user(retinfo,&tmp,sizeof(*retinfo)))
    return -EFAULT;
  return 0;
}

static int set_serial_info(struct iCom_port * iCom_port_info,
                           struct serial_struct * new_info)
{
  struct serial_struct new_serial;
  int                  old_flags;
  int 		       retval = 0;
  unsigned long        flags;

  TRACE(iCom_port_info,TRACE_SET_SERIAL,0);

  if (copy_from_user(&new_serial,new_info,sizeof(new_serial)))
    return -EFAULT;

  old_flags = iCom_port_info->flags;
  /* new_serial.irq --- irq of adapter will not change, PCI only */
  /* new_serial.xmit_fifo_size -- can not change on this device */
  /* new_serial.baud_base -- ??? */
  /* new_serial.custom_divisor -- device specific */
  /* new_serial.hub6 -- device specific */
  /* new_serial.type -- device specific */
  /* new_serial.port -- address of port will not change, PCI only */

  if (!capable(CAP_SYS_ADMIN)) {
    if ((new_serial.baud_base != iCom_port_info->baud_base) ||
        (new_serial.close_delay != iCom_port_info->close_delay) ||
        ((new_serial.flags & ~ASYNC_USR_MASK) !=
         (iCom_port_info->flags & ~ASYNC_USR_MASK)))
      return -EPERM;
    iCom_port_info->flags = ((iCom_port_info->flags & ~ASYNC_USR_MASK) |
                             (new_serial.flags & ASYNC_USR_MASK));
    goto check_and_exit;
  }

  if (new_serial.baud_base < 9600) {
    return -EINVAL;
  }

  /*
   * OK, past this point, all the error checking has been done.
   * At this point, we start making changes.....
   */
  iCom_port_info->baud_base = new_serial.baud_base;
  iCom_port_info->flags = ((iCom_port_info->flags & ~ASYNC_FLAGS) |
                           (new_serial.flags & ASYNC_FLAGS));
  iCom_port_info->close_delay = new_serial.close_delay * HZ/100;
  iCom_port_info->closing_wait = new_serial.closing_wait * HZ/100;
  iCom_port_info->tty->low_latency = (iCom_port_info->flags & ASYNC_LOW_LATENCY) ? 1 : 0;

  check_and_exit:
  spin_lock_irqsave(&iComlock,flags);
  if (iCom_port_info->flags & ASYNC_INITIALIZED) {
    if (((iCom_port_info->flags & ASYNC_SPD_MASK) !=
         (old_flags & ASYNC_SPD_MASK))) {
      if ((iCom_port_info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_HI)
        iCom_port_info->tty->alt_speed = 57600;
      if ((iCom_port_info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_VHI)
        iCom_port_info->tty->alt_speed = 115200;
      if ((iCom_port_info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_SHI)
        iCom_port_info->tty->alt_speed = 230400;
      if ((iCom_port_info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_WARP)
        iCom_port_info->tty->alt_speed = 460800;
      change_speed(iCom_port_info, 0, flags);
    }
  } else
    retval = startup(iCom_port_info, flags);

  spin_unlock_irqrestore(&iComlock,flags);

  return retval;
}

/*
 * get_lsr_info - get line status register info
 *
 * Purpose: Let user call ioctl() to get info when the UART physically
 * 	    is emptied.  On bus types like RS485, the transmitter must
 * 	    release the bus after transmitting. This must be done when
 * 	    the transmit shift register is empty, not be done when the
 * 	    transmit holding register is empty.  This functionality
 * 	    allows an RS485 driver to be written in user space. 
 */
static int get_lsr_info(struct iCom_port * info, unsigned int *value)
{
  unsigned char status;
  unsigned int result;

  TRACE(info,TRACE_SET_LSR,0);

  status = cpu_to_le16(info->statStg->xmit[0].flags);
  result = ((status & SA_FLAGS_DONE) ? TIOCSER_TEMT : 0);
  return put_user(result,value);
}

static int iCom_ioctl(struct tty_struct * tty, struct file * filp,
			unsigned int cmd, unsigned long arg) 
{
  int error;
  struct iCom_port * iCom_port_info = (struct iCom_port *)tty->driver_data;
  struct async_icount cprev, cnow;	/* kernel counter temps */
  struct serial_icounter_struct *p_cuser;	/* user space */
  unsigned long flags;

  TRACE(iCom_port_info,TRACE_IOCTL | TRACE_TIME,jiffies);
  if ((cmd != TIOCGSERIAL) && (cmd != TIOCSSERIAL) &&
      (cmd != TIOCSERCONFIG) && (cmd != TIOCSERGSTRUCT) &&
      (cmd != TIOCMIWAIT) && (cmd != TIOCGICOUNT)) {
    if (tty->flags & (1 << TTY_IO_ERROR))
      return -EIO;
  }

  switch (cmd) {
    case 0x4300:
      if (copy_to_user((void *)arg,iCom_port_info->trace_blk,TRACE_BLK_SZ))
	  return -EFAULT;
      return 0;
    case TIOCMGET:
      return get_modem_info(iCom_port_info, (unsigned int *) arg);
    case TIOCMBIS:
    case TIOCMBIC:
    case TIOCMSET:
      return set_modem_info(iCom_port_info, cmd, (unsigned int *) arg);
    case TIOCGSERIAL:
      return get_serial_info(iCom_port_info,
                             (struct serial_struct *) arg);
    case TIOCSSERIAL:
      return set_serial_info(iCom_port_info,
                             (struct serial_struct *) arg);

    case TIOCSERGETLSR: /* Get line status register */
      return get_lsr_info(iCom_port_info, (unsigned int *) arg);

      /*
       * Wait for any of the 4 modem inputs (DCD,RI,DSR,CTS) to change
       * - mask passed in arg for lines of interest
       *   (use |'ed TIOCM_RNG/DSR/CD/CTS for masking)
       * Caller should use TIOCGICOUNT to see which one it was
       */
    case TIOCMIWAIT:
      spin_lock_irqsave(&iComlock,flags);
      /* note the counters on entry */
      cprev = iCom_port_info->icount;
      spin_unlock_irqrestore(&iComlock,flags);
      while (1) {
        interruptible_sleep_on(&iCom_port_info->delta_msr_wait);
        /* see if a signal did it */
        if (signal_pending(current))
          return -ERESTARTSYS;
        spin_lock_irqsave(&iComlock,flags);
        cnow = iCom_port_info->icount; /* atomic copy */
        spin_unlock_irqrestore(&iComlock,flags);
        if (cnow.rng == cprev.rng && cnow.dsr == cprev.dsr && 
            cnow.dcd == cprev.dcd && cnow.cts == cprev.cts)
          return -EIO; /* no change => error */
        if ( ((arg & TIOCM_RNG) && (cnow.rng != cprev.rng)) ||
             ((arg & TIOCM_DSR) && (cnow.dsr != cprev.dsr)) ||
             ((arg & TIOCM_CD)  && (cnow.dcd != cprev.dcd)) ||
             ((arg & TIOCM_CTS) && (cnow.cts != cprev.cts)) ) {
          return 0;
        }
        cprev = cnow;
      }
      /* NOTREACHED */

      /* 
       * Get counter of input serial line interrupts (DCD,RI,DSR,CTS)
       * Return: write counters to the user passed counter struct
       * NB: both 1->0 and 0->1 transitions are counted except for
       *     RI where only 0->1 is counted.
       */
    case TIOCGICOUNT:
      spin_lock_irqsave(&iComlock,flags);
      cnow = iCom_port_info->icount;
      spin_unlock_irqrestore(&iComlock,flags);
      p_cuser = (struct serial_icounter_struct *) arg;
      error = put_user(cnow.cts, &p_cuser->cts);
      if (error) return error;
      error = put_user(cnow.dsr, &p_cuser->dsr);
      if (error) return error;
      error = put_user(cnow.rng, &p_cuser->rng);
      if (error) return error;
      error = put_user(cnow.dcd, &p_cuser->dcd);
      if (error) return error;
      error = put_user(cnow.rx, &p_cuser->rx);
      if (error) return error;
      error = put_user(cnow.tx, &p_cuser->tx);
      if (error) return error;
      error = put_user(cnow.frame, &p_cuser->frame);
      if (error) return error;
      error = put_user(cnow.overrun, &p_cuser->overrun);
      if (error) return error;
      error = put_user(cnow.parity, &p_cuser->parity);
      if (error) return error;
      error = put_user(cnow.brk, &p_cuser->brk);
      if (error) return error;
      error = put_user(cnow.buf_overrun, &p_cuser->buf_overrun);
      if (error) return error;			
      return 0;

    case TIOCSERGWILD:
    case TIOCSERSWILD:
      /* "setserial -W" is called in Debian boot */
      printk ("TIOCSER?WILD ioctl obsolete, ignored.\n");
      return 0;

    default:
      TRACE(iCom_port_info,TRACE_IOCTL_IGNORE,cmd);
      return -ENOIOCTLCMD;
  }
  return 0;
}

static void iCom_send_xchar(struct tty_struct * tty, char ch)
{
  struct iCom_port *iCom_port_info = (struct iCom_port *)tty->driver_data;
  unsigned char    xdata;
  int              index;
  unsigned long    flags;

  spin_lock_irqsave(&iComlock,flags);
  TRACE(iCom_port_info,TRACE_SEND_XCHAR,ch);
  /* attempt sending char for a period of .1 second */
  for (index = 0; index < 10; index++ ) {
      xdata = iCom_readb(&iCom_port_info->dram->xchar);
      if (xdata == 0x00) {
	  TRACE(iCom_port_info,TRACE_QUICK_WRITE,0);
	  iCom_writeb(ch,&iCom_port_info->dram->xchar);
	  break;
      }
      current->state = TASK_INTERRUPTIBLE;
      spin_unlock_irqrestore(&iComlock,flags);
      schedule_timeout(HZ/100);
      spin_lock_irqsave(&iComlock,flags);
  }
  spin_unlock_irqrestore(&iComlock,flags);
}

static void iCom_throttle(struct tty_struct * tty)
{
  struct iCom_port *iCom_port_info = (struct iCom_port *)tty->driver_data;
  unsigned char    osr;

  TRACE(iCom_port_info,TRACE_THROTTLE,0);
  if (I_IXOFF(tty))
    iCom_send_xchar(tty, STOP_CHAR(tty));

  if (tty->termios->c_cflag & CRTSCTS) {
    osr = iCom_readb(&iCom_port_info->dram->osr);
    iCom_writeb(osr & ~ICOM_RTS,&iCom_port_info->dram->osr);
  }
}

static void iCom_unthrottle(struct tty_struct * tty)
{
  struct iCom_port *iCom_port_info = (struct iCom_port *)tty->driver_data;
  unsigned char    osr;

  TRACE(iCom_port_info,TRACE_UNTHROTTLE,0);
  if (I_IXOFF(tty)) {
    iCom_send_xchar(tty, START_CHAR(tty));
  }
  if (tty->termios->c_cflag & CRTSCTS) {
    osr = iCom_readb(&iCom_port_info->dram->osr);
    iCom_writeb(osr | ICOM_RTS,&iCom_port_info->dram->osr);
  }
}

static void iCom_set_termios(struct tty_struct * tty, struct termios * old_termios)
{
  struct iCom_port *iCom_port_info = (struct iCom_port *)tty->driver_data;
  unsigned int     cflag = tty->termios->c_cflag;
  unsigned char    osr;
  unsigned long    flags;
#define RELEVANT_IFLAG(iflag) (iflag & (IGNBRK|BRKINT|IGNPAR|PARMRK|INPCK))

  spin_lock_irqsave(&iComlock,flags);
  TRACE(iCom_port_info,TRACE_SET_TERMIOS,0);
  if ((cflag == old_termios->c_cflag)
      && (RELEVANT_IFLAG(tty->termios->c_iflag) 
	  == RELEVANT_IFLAG(old_termios->c_iflag))) {
      spin_unlock_irqrestore(&iComlock,flags);
      return;
  }

  change_speed(iCom_port_info, old_termios, flags);

  /* Handle transition to B0 status */
  if ((old_termios->c_cflag & CBAUD) &&
      !(cflag & CBAUD)) {
    osr = iCom_readb(&iCom_port_info->dram->osr);
    TRACE(iCom_port_info,TRACE_DROP_DTR_RTS,0);
    iCom_writeb(osr & ~(ICOM_DTR|ICOM_RTS),&iCom_port_info->dram->osr);
  }

  /* Handle transition away from B0 status */
  if (!(old_termios->c_cflag & CBAUD) &&
      (cflag & CBAUD)) {
    osr = iCom_readb(&iCom_port_info->dram->osr);
    TRACE(iCom_port_info,TRACE_RAISE_DTR,0);
    osr |= ICOM_DTR;
    if (!(tty->termios->c_cflag & CRTSCTS) || 
        !test_bit(TTY_THROTTLED, &tty->flags)) {
      TRACE(iCom_port_info,TRACE_RAISE_RTS,0);
      osr |= ICOM_RTS;
    }
    iCom_writeb(osr,&iCom_port_info->dram->osr);
  }

  spin_unlock_irqrestore(&iComlock,flags);

  /* Handle turning off CRTSCTS */
  if ((old_termios->c_cflag & CRTSCTS) &&
      !(tty->termios->c_cflag & CRTSCTS)) {
    tty->hw_stopped = 0;
    iCom_start(tty);
  }

#if 0
  /*
   * No need to wake up processes in open wait, since they
   * sample the CLOCAL flag once, and don't recheck it.
   * XXX  It's not clear whether the current behavior is correct
   * or not.  Hence, this may change.....
   */
  if (!(old_termios->c_cflag & CLOCAL) &&
      (tty->termios->c_cflag & CLOCAL))
    wake_up_interruptible(&iCom_port_info->open_wait);
#endif
}

static void iCom_stop(struct tty_struct * tty)
{
  struct iCom_port *iCom_port_info = (struct iCom_port *)tty->driver_data;
  unsigned char    cmdReg;
  unsigned long    flags;

  spin_lock_irqsave(&iComlock,flags);
  TRACE(iCom_port_info,TRACE_STOP,0);
  cmdReg = iCom_readb(&iCom_port_info->dram->CmdReg);
  iCom_writeb(cmdReg | CMD_HOLD_XMIT,&iCom_port_info->dram->CmdReg);
  spin_unlock_irqrestore(&iComlock,flags);
}

static void iCom_start(struct tty_struct * tty)
{
  struct iCom_port *iCom_port_info = (struct iCom_port *)tty->driver_data;
  unsigned char    cmdReg;
  unsigned long    flags;

  spin_lock_irqsave(&iComlock,flags);
  TRACE(iCom_port_info,TRACE_START,0);
  cmdReg = iCom_readb(&iCom_port_info->dram->CmdReg);
  iCom_writeb(cmdReg & ~CMD_HOLD_XMIT,&iCom_port_info->dram->CmdReg);
  spin_unlock_irqrestore(&iComlock,flags);
}

static void iCom_hangup(struct tty_struct * tty)
{
  struct iCom_port *iCom_port_info = (struct iCom_port *)tty->driver_data;
  unsigned long    flags;

  TRACE(iCom_port_info,TRACE_HANGUP,0);
  iCom_flush_buffer(tty);
  spin_lock_irqsave(&iComlock,flags);
  shutdown(iCom_port_info);
  iCom_port_info->open_active_count = 0;
  iCom_port_info->flags &= ~(ASYNC_NORMAL_ACTIVE|ASYNC_CALLOUT_ACTIVE);
  iCom_port_info->tty = 0;
  wake_up_interruptible(&iCom_port_info->open_wait);
  spin_unlock_irqrestore(&iComlock,flags);
}

static void iCom_break(struct tty_struct *tty, int break_state)
{
  struct iCom_port *iCom_port_info = (struct iCom_port *)tty->driver_data;
  unsigned char    cmdReg;
  unsigned long    flags;

  spin_lock_irqsave(&iComlock,flags);
  TRACE(iCom_port_info,TRACE_BREAK,0);
  cmdReg = iCom_readb(&iCom_port_info->dram->CmdReg);
  if (break_state == -1) {
    iCom_writeb(cmdReg | CMD_SND_BREAK,&iCom_port_info->dram->CmdReg);
  }
  else{
    iCom_writeb(cmdReg & ~CMD_SND_BREAK,&iCom_port_info->dram->CmdReg);
  }
  spin_unlock_irqrestore(&iComlock,flags);
}

/*
 * iCom_wait_until_sent() --- wait until the transmitter is empty
 */
static void iCom_wait_until_sent(struct tty_struct *tty, int timeout)
{
  struct iCom_port *iCom_port_info = (struct iCom_port *)tty->driver_data;
  unsigned long orig_jiffies, char_time;
  int status;

  TRACE(iCom_port_info,TRACE_WAIT_UNTIL_SENT,0);

  orig_jiffies = jiffies;
  /*
   * Set the check interval to be 1/5 of the estimated time to
   * send a single character, and make it at least 1.  The check
   * interval should also be less than the timeout.
   * 
   * Note: we have to use pretty tight timings here to satisfy
   * the NIST-PCTS.
   */
  char_time = (iCom_port_info->timeout - HZ/50) / iCom_port_info->xmit_fifo_size;
  char_time = char_time / 5;
  if (char_time == 0)
      char_time = 1;
  if (timeout) {
      if (timeout < char_time)
	char_time = timeout;
  }
  /*
   * If the transmitter hasn't cleared in twice the approximate
   * amount of time to send the entire FIFO, it probably won't
   * ever clear.  This assumes the UART isn't doing flow
   * control, which is currently the case.  Hence, if it ever
   * takes longer than iCom_port_info->timeout, this is probably due to a
   * UART bug of some kind.  So, we clamp the timeout parameter at
   * 2*iCom_port_info->timeout.
   */
  if (!timeout || timeout > 2*iCom_port_info->timeout)
      timeout = 2*iCom_port_info->timeout;

  status = cpu_to_le16(iCom_port_info->statStg->xmit[0].flags);
  while (status & SA_FLAGS_DONE ) {  /*data still transmitting*/

      current->state = TASK_INTERRUPTIBLE;
      schedule_timeout(char_time);
      if (signal_pending(current))
	  break;
      if (timeout && time_after(jiffies, orig_jiffies + timeout))
	  break;
      status = cpu_to_le16(iCom_port_info->statStg->xmit[0].flags);
  }
  current->state = TASK_RUNNING;
}

/*
 * /proc fs routines....
 */
static inline int line_info(char *buf, int free_space, struct iCom_port *iCom_port_info)
{
  char	line[72], control, status;
  int	ret, baud_index, len;
  int   port;

  if ((iCom_port_info->port == 2) &&
      (iCom_adapter_info[iCom_port_info->adapter].subsystem_id != FOUR_PORT_MODEL))
      port = 1;
  else
      port = iCom_port_info->port;

  memset(line, 0,sizeof(line));
  len = sprintf(line,"   port:%d      maj:%d min:%d cable:",
                port,
                243,
                iCom_port_info->minor_number);
  if (!iCom_port_info->passed_diags) {
    len += sprintf(line+len," PortFailed\n");
  }
  else if (iCom_port_info->imbed_modem == ICOM_IMBED_MODEM) {
    len += sprintf(line+len," InternalModem\n");
  }
  else if (iCom_port_info->cable_id == RS232_CABLE) {
    len += sprintf(line+len," RS232\n");
  }
  else {
    len += sprintf(line+len," ----\n");
  }

  if (len > free_space) return 0;
  ret = sprintf(buf, "%s", line);

  memset(line, 0,sizeof(line));
  baud_index = iCom_readb(&iCom_port_info->dram->async_config3);
  len = sprintf(line, "     baud:%d",icom_acfg_baud[baud_index]);

  len += sprintf(line+len, " tx:%d rx:%d",
                 iCom_port_info->icount.tx, iCom_port_info->icount.rx);

  if (iCom_port_info->icount.frame)
    len += sprintf(line+len, " fe:%d", iCom_port_info->icount.frame);

  if (iCom_port_info->icount.parity)
    len += sprintf(line+len, " pe:%d", iCom_port_info->icount.parity);

  if (iCom_port_info->icount.brk)
    len += sprintf(line+len, " brk:%d", iCom_port_info->icount.brk);	

  if (iCom_port_info->icount.overrun)
    len += sprintf(line+len, " oe:%d", iCom_port_info->icount.overrun);

  if ((ret + len) > free_space) return 0;
  ret += sprintf(buf+ret, "%s", line);

  /*
   * Last thing is the RS-232 status lines
   */
  memset(line, 0,sizeof(line));
  status = iCom_readb(&iCom_port_info->dram->isr);
  control = iCom_readb(&iCom_port_info->dram->osr);

  line[0] = 0;
  line[1] = 0;
  len = 0;
  if (control & ICOM_RTS)
    len += sprintf(line+len, "|RTS");
  if (status & ICOM_CTS)
    len += sprintf(line+len, "|CTS");
  if (control & ICOM_DTR)
    len += sprintf(line+len, "|DTR");
  if (status & ICOM_DSR)
    len += sprintf(line+len, "|DSR");
  if (status & ICOM_DCD)
    len += sprintf(line+len, "|CD");
  if (status & ICOM_RI)
    len += sprintf(line+len, "|RI");

  if ((ret + len) > free_space) return 0;
  ret += sprintf(buf+ret, "   %s\n", line+1);

  return ret;
}

int iCom_read_proc(char *page, char **start, off_t off, int count,
		   int *eof, void *data)
{
  int i, j, len = 0, l;
  off_t	begin = 0;
  char line[72];
  struct LocationDataStruct *p_location_data = NULL;

  len += sprintf(page, "%s: %s\n",ICOM_DRIVER_NAME, ICOM_VERSION_STR);
  for (i = 0; i < active_adapters; i++) {
    memset(line, 0,sizeof(line));
#ifdef CONFIG_PPC_ISERIES
    p_location_data = iSeries_GetLocationData(iCom_adapter_info[i].pci_dev);

    if (p_location_data != NULL)
    {
      l = sprintf(line,"Frame ID: %d", p_location_data->FrameId);
      l += sprintf(line+l," Card Position: %s", p_location_data->CardLocation);
      kfree(p_location_data);
    }
    else
      l = sprintf(line, "Adapter %d", i);
#else
    l = sprintf(line, "Adapter %d", i);
#endif
    switch (iCom_adapter_info[i].subsystem_id) {
      case FOUR_PORT_MODEL:
        l += sprintf(line+l," CCIN: 2805\n");
        break;
      case V2_TWO_PORTS_RVX:
        l += sprintf(line+l," CCIN: 2742\n");
        break;
      case V2_ONE_PORT_RVX_ONE_PORT_IMBED_MDM:
        l += sprintf(line+l," CCIN: 2793\n");
        break;
      default:
        l += sprintf(line+l, "\n");
    }

    if ((l + len) < count) {
      len += sprintf(page + len,"%s",line);
      if (len+begin <= off) {
        begin += len;
        len = 0;
      }
    }
    else
      goto done;

    for (j= 0; j < 4; j++) {
	if (iCom_adapter_info[i].port_info[j].status == ICOM_PORT_ACTIVE) {
	    l = line_info(page + len, count - len, &iCom_adapter_info[i].port_info[j]);
	    /* l == 0 means line_info could not fit new data into remaining space */
	    if (l == 0)
		goto done;
	    len += l;
	    if (len+begin <= off) {
		begin += len;
		len = 0;
	    }
	}
    }
  }
  *eof = 1;
  done:
    if (off >= len+begin)
      return 0;
  /*
     if returning multiple pages, the time delta between sending pages
     may experience changes in the data size represented in the pages
     already sent.  The logic as is may duplicate a line but should
     never miss a line
  */
  *start = page;
  return len;
}

static void iCom_flush_buffer(struct tty_struct * tty)
{
  struct iCom_port *iCom_port_info = (struct iCom_port *)tty->driver_data;
  unsigned char    cmdReg;
  unsigned long    flags;

  spin_lock_irqsave(&iComlock,flags);
  TRACE(iCom_port_info,TRACE_FLUSH_BUFFER,0);
  /*
   * with no CMD_XMIT_ENABLE is same as disabling xmitter.  This should
   * result in an interrupt if currently transmitting
   */
  cmdReg = iCom_readb(&iCom_port_info->dram->CmdReg);
  iCom_writeb(cmdReg & ~CMD_XMIT_ENABLE,&iCom_port_info->dram->CmdReg);  
  spin_unlock_irqrestore(&iComlock,flags);
}

/*
 * This routine is used by the interrupt handler to schedule
 * processing in the software interrupt portion of the driver.
 */
static inline void rs_sched_event(struct iCom_port *info,
				  int event)
{
  info->event |= 1 << event;
  queue_task(&info->tqueue, &tq_immediate);
  mark_bh(IMMEDIATE_BH);
}

static inline void check_modem_status(struct iCom_port *iCom_port_info)
{
  static char old_status = 0;
  char delta_status;
  unsigned char status;

  /*modem input register */
  status = iCom_readb(&iCom_port_info->dram->isr);
  TRACE(iCom_port_info,TRACE_CHECK_MODEM,status);
  delta_status = status ^ old_status;
  if (delta_status) {
    if (delta_status & ICOM_RI)
      iCom_port_info->icount.rng++;
    if (delta_status & ICOM_DSR)
      iCom_port_info->icount.dsr++;
    if (delta_status & ICOM_DCD)
      iCom_port_info->icount.dcd++;
    if (delta_status & ICOM_CTS)
      iCom_port_info->icount.cts++;
  
    wake_up_interruptible(&iCom_port_info->delta_msr_wait);
    old_status = status;
  }

  if ((iCom_port_info->flags & ASYNC_CHECK_CD) && (status & ICOM_DCD)) {
    if (status & ICOM_DCD) /* Carrier Detect up */
      wake_up_interruptible(&iCom_port_info->open_wait);
    else if (!((iCom_port_info->flags & ASYNC_CALLOUT_ACTIVE) &&
               (iCom_port_info->flags & ASYNC_CALLOUT_NOHUP))) {
      if (iCom_port_info->tty)
        tty_hangup(iCom_port_info->tty);
    }
  }

  if (iCom_port_info->flags & ASYNC_CTS_FLOW) {
    if (iCom_port_info->tty->hw_stopped) {
      if (status & ICOM_CTS) {  /* CTS up */
        iCom_port_info->tty->hw_stopped = 0;
	TRACE(iCom_port_info,TRACE_CTS_UP,0);
        rs_sched_event(iCom_port_info, 0);
        return;
      }
    } else {
      if (!(status & ICOM_CTS)) { /* CTS down */
        iCom_port_info->tty->hw_stopped = 1;
	TRACE(iCom_port_info,TRACE_CTS_DOWN,0);
      }
    }
  }
}

static void process_interrupt(u16 port_int_reg, struct iCom_port *iCom_port_info)
{
  short int           count, rcv_buff;
  struct tty_struct   *tty = iCom_port_info->tty;
  unsigned char       *data;
  unsigned short int  status;
  struct async_icount *icount;
  unsigned long int   offset;


  TRACE(iCom_port_info,TRACE_INTERRUPT | TRACE_TIME,jiffies);

  if (port_int_reg & (INT_XMIT_COMPLETED | INT_XMIT_DISABLED)) {
    if (port_int_reg & (INT_XMIT_COMPLETED))
	TRACE(iCom_port_info,TRACE_XMIT_COMPLETE,0);
    else
	TRACE(iCom_port_info,TRACE_XMIT_DISABLED,0);

    /* clear buffer in use bit */
    iCom_port_info->statStg->xmit[0].flags &= cpu_to_le16(~SA_FLAGS_READY_TO_XMIT);
    iCom_port_info->icount.tx += (unsigned short int)cpu_to_le16(iCom_port_info->statStg->xmit[0].leLength);

    /* activate write queue */
    rs_sched_event(iCom_port_info, 0);
  }

  if (port_int_reg & INT_RCV_COMPLETED) {

    TRACE(iCom_port_info,TRACE_RCV_COMPLETE,0);
    rcv_buff = iCom_port_info->next_rcv;

    status = cpu_to_le16(iCom_port_info->statStg->rcv[rcv_buff].flags);
    while (status & SA_FL_RCV_DONE) {

      TRACE(iCom_port_info,TRACE_FID_STATUS,status);

      count = cpu_to_le16(iCom_port_info->statStg->rcv[rcv_buff].leLength);
	
      TRACE(iCom_port_info,TRACE_RCV_COUNT,count);
      if (count > (TTY_FLIPBUF_SIZE - tty->flip.count))
        count = TTY_FLIPBUF_SIZE - tty->flip.count;

      TRACE(iCom_port_info,TRACE_REAL_COUNT,count);

      offset = cpu_to_le32(iCom_port_info->statStg->rcv[rcv_buff].leBuffer) - iCom_port_info->recv_buf_pci;

      memcpy(tty->flip.char_buf_ptr,(unsigned char *)((unsigned long int)iCom_port_info->recv_buf + offset),count);

      data = (unsigned char *)tty->flip.char_buf_ptr;

      if (count > 0) {
	tty->flip.count += count - 1;
	tty->flip.char_buf_ptr += count - 1;
	
	memset(tty->flip.flag_buf_ptr, 0, count);
	tty->flip.flag_buf_ptr += count - 1;
      }

      icount = &iCom_port_info->icount;
      icount->rx += count;

      /* Break detect logic */
      if ((status & SA_FLAGS_FRAME_ERROR) && (tty->flip.char_buf_ptr[0] == 0x00)) {
	status &= ~SA_FLAGS_FRAME_ERROR;
	status |= SA_FLAGS_BREAK_DET;
        TRACE(iCom_port_info,TRACE_BREAK_DET,0);
      }

      if (status & (SA_FLAGS_BREAK_DET | SA_FLAGS_PARITY_ERROR |
                    SA_FLAGS_FRAME_ERROR | SA_FLAGS_OVERRUN)) {

        if (status & SA_FLAGS_BREAK_DET)
          icount->brk++;
        if (status & SA_FLAGS_PARITY_ERROR)
          icount->parity++;
        if (status & SA_FLAGS_FRAME_ERROR)
          icount->frame++;
        if (status & SA_FLAGS_OVERRUN)
          icount->overrun++;

        /*
         * Now check to see if character should be
         * ignored, and mask off conditions which
         * should be ignored.
         */ 
        if (status & iCom_port_info->ignore_status_mask) {
	  TRACE(iCom_port_info,TRACE_IGNORE_CHAR,0);
          goto ignore_char;
        }

        status &= iCom_port_info->read_status_mask;

        if (status & SA_FLAGS_BREAK_DET) {
          *tty->flip.flag_buf_ptr = TTY_BREAK;
          if (iCom_port_info->flags & ASYNC_SAK)
            do_SAK(tty);
        } else if (status & SA_FLAGS_PARITY_ERROR) {
	  TRACE(iCom_port_info,TRACE_PARITY_ERROR,0);
          *tty->flip.flag_buf_ptr = TTY_PARITY;
	}
        else if (status & SA_FLAGS_FRAME_ERROR)
          *tty->flip.flag_buf_ptr = TTY_FRAME;
        if (status & SA_FLAGS_OVERRUN) {
          /*
           * Overrun is special, since it's
           * reported immediately, and doesn't
           * affect the current character
           */
          if (tty->flip.count < TTY_FLIPBUF_SIZE) {
            tty->flip.count++;
            tty->flip.flag_buf_ptr++;
            tty->flip.char_buf_ptr++;
            *tty->flip.flag_buf_ptr = TTY_OVERRUN;
          }
        }
      }

      tty->flip.flag_buf_ptr++;
      tty->flip.char_buf_ptr++;
      tty->flip.count++;
      ignore_char:
      iCom_port_info->statStg->rcv[rcv_buff].flags = 0;
      iCom_port_info->statStg->rcv[rcv_buff].leLength = 0;
      iCom_port_info->statStg->rcv[rcv_buff].WorkingLength = (unsigned short int)cpu_to_le16(RCV_BUFF_SZ);

      rcv_buff++;
      if (rcv_buff == NUM_RBUFFS) rcv_buff = 0;

      status = cpu_to_le16(iCom_port_info->statStg->rcv[rcv_buff].flags);
    }
    iCom_port_info->next_rcv = rcv_buff;
    tty_flip_buffer_push(tty);
  }
}

static void iCom_interrupt(int irq, void * dev_id, struct pt_regs * regs)
{
  unsigned long int   int_reg;
  u32                 adapter_interrupts;
  u16                 port_int_reg;
  struct iCom_adapter *iCom_adapter_ptr;
  struct iCom_port    *iCom_port_info;
  unsigned long       flags;

  spin_lock_irqsave(&iComlock,flags);

  /* find iCom_port_info for this interrupt */
  iCom_adapter_ptr = (struct iCom_adapter *)dev_id;

  if ((iCom_adapter_ptr->version | ADAPTER_V2) == ADAPTER_V2) {
    int_reg = iCom_adapter_ptr->base_addr + 0x8024;

    adapter_interrupts = iCom_readl((void *)int_reg);
    
    if (adapter_interrupts & 0x00003FFF) {
      /* port 2 interrupt,  NOTE:  for all ADAPTER_V2, port 2 will be active */
      iCom_port_info = &iCom_adapter_ptr->port_info[2];
      port_int_reg = (u16)adapter_interrupts;
      process_interrupt(port_int_reg, iCom_port_info);
      check_modem_status(iCom_port_info);
    }
    if (adapter_interrupts & 0x3FFF0000) {
      /* port 3 interrupt */
      iCom_port_info = &iCom_adapter_ptr->port_info[3];
      if (iCom_port_info->status == ICOM_PORT_ACTIVE) {
	  port_int_reg = (u16)(adapter_interrupts >> 16);
	  process_interrupt(port_int_reg, iCom_port_info);
	  check_modem_status(iCom_port_info);
      }
    }
    
    /* Clear out any pending interrupts */
    iCom_writel(adapter_interrupts,(void *)int_reg);
    
    int_reg = iCom_adapter_ptr->base_addr + 0x8004;
  }
  else {
    int_reg = iCom_adapter_ptr->base_addr + 0x4004;
  }

  adapter_interrupts = iCom_readl((void *)int_reg);

  if (adapter_interrupts & 0x00003FFF) {
    /* port 0 interrupt, NOTE:  for all adapters, port 0 will be active */
    iCom_port_info = &iCom_adapter_ptr->port_info[0];
    port_int_reg = (u16)adapter_interrupts;
    process_interrupt(port_int_reg, iCom_port_info);
    check_modem_status(iCom_port_info);
  }
  if (adapter_interrupts & 0x3FFF0000) {
    /* port 1 interrupt */
    iCom_port_info = &iCom_adapter_ptr->port_info[1];
    if (iCom_port_info->status == ICOM_PORT_ACTIVE) {
	port_int_reg = (u16)(adapter_interrupts >> 16);
	process_interrupt(port_int_reg, iCom_port_info);
	check_modem_status(iCom_port_info);
    }
  }
  
  /* Clear out any pending interrupts */
  iCom_writel(adapter_interrupts,(void *)int_reg);
  spin_unlock_irqrestore(&iComlock,flags);
}

/*
 * -------------------------------------------------------------------
 * Here ends the serial interrupt routines.
 * -------------------------------------------------------------------
 */

/*
 * This routine is used to handle the "bottom half" processing for the
 * serial driver, known also the "software interrupt" processing.
 * This processing is done at the kernel interrupt level, after the
 * iCom_interrupt() has returned, BUT WITH INTERRUPTS TURNED ON.  This
 * is where time-consuming activities which can not be done in the
 * interrupt driver proper are done; the interrupt driver schedules
 * them using rs_sched_event(), and they get done here.
 */
static void do_softint(void *private_)
{
  struct iCom_port	*info = (struct iCom_port *) private_;
  struct tty_struct	*tty;

  tty = info->tty;
  if (!tty)
    return;

  if (test_and_clear_bit(0, &info->event)) {
    if ((tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) && tty->ldisc.write_wakeup)
      (tty->ldisc.write_wakeup)(tty);
    wake_up_interruptible(&tty->write_wait);
    TRACE(info,TRACE_WAKEUP,0);
  }
}

/*******************************************/
/* Diagnostic Check for previous Error Log */
/*******************************************/
static u8 __init NoErrorYet(struct iCom_adapter *iCom_adapter_ptr,
		u8	port)  /* IOA_FAILURE indicates an Adapter failure */
{
    if(port == IOA_FAILURE)
    {
	if(iCom_adapter_ptr->error_data[0] == 0)
	    return -1; /* return non zero (TRUE) */
    }
    else
    {
	if(iCom_adapter_ptr->port_info[port].tpr == 0)
	   return -1; /* return non zero (TRUE) */
    }

/* must be an error so return 0 (FALSE) */
return 0;
}

/************************************/
/* Diagnostic Error Logging Routine */
/************************************/
void __init diag_error(struct iCom_adapter *iCom_adapter_ptr,
		u8	port,  /* IOA_FAILURE indicates an Adapter failure */
		char*	text, /* error message */
		u32	tpr,
		u32*	error_data_ptr,
		u32	num_error_entries)

{
    u32	i,j;
    u32 error_data[NUM_ERROR_ENTRIES];
    u32 *temp32ptr;
    int version;
    char port_letter = '0';
    struct LocationDataStruct *p_location_data = NULL;

    
    if((iCom_adapter_ptr->version | ADAPTER_V2) != ADAPTER_V2)
	version = 1;
    else
	version = 2;

    if(port == 0)
	port_letter = 'A';
    else if(port == 1)
	port_letter = 'B';
    else if(port == 2)
	port_letter = 'C';
    else if(port == 3)
	port_letter = 'D';

    if(NoErrorYet(iCom_adapter_ptr, port))
    {
	if(port == IOA_FAILURE)
	{
	    iCom_adapter_ptr->tpr = tpr;
	    temp32ptr = &iCom_adapter_ptr->error_data[0];

	    /* set all ports to failed */
	    for(j=0;j<NUM_ERROR_ENTRIES;j++)
	    {
		iCom_adapter_ptr->port_info[j].passed_diags = 0; /* failed */
	    }

	    /* find the MIN of num_error_entries and NUM_ERROR_ENTRIES */
	    if(num_error_entries < NUM_ERROR_ENTRIES)
		j=num_error_entries;
	    else
		j=NUM_ERROR_ENTRIES;
	    for(i=0;i<j;i++)
	    {
		temp32ptr[i] = error_data_ptr[i];
	    }
	}
	else
	{
	    iCom_adapter_ptr->port_info[port].tpr = tpr;
	    temp32ptr = &iCom_adapter_ptr->port_info[port].error_data[0];
	    iCom_adapter_ptr->port_info[port].passed_diags = 0;/* failed */

	    /* find the MIN of num_error_entries and NUM_ERROR_ENTRIES */
	    if(num_error_entries < NUM_ERROR_ENTRIES)
		j=num_error_entries;
	    else
		j=NUM_ERROR_ENTRIES;
	    for(i=0;i<j;i++)
	    {
		temp32ptr[i] = error_data_ptr[i];
	    }
	}

	/* copy error data into local array */
	for(i=0;i<j;i++)
	    error_data[i] = error_data_ptr[i];

	/* set the rest of the error data array to zero */
	for(;i<NUM_ERROR_ENTRIES;i++)
	    error_data[i] = 0;

	printk("iCom: ERROR #################################################\n");
#ifdef CONFIG_PPC_ISERIES
	p_location_data = iSeries_GetLocationData(iCom_adapter_ptr->pci_dev);

	if (p_location_data != NULL)
	{
	    printk("iCom: Frame ID: %d Card Position: %s.\n",p_location_data->FrameId, p_location_data->CardLocation);
	    kfree(p_location_data);
	}
#endif
	if(port == IOA_FAILURE)
	    printk("iCom: Version %d adapter failed.\n",version);
	else
	    printk("iCom: Version %d adapter port %c failed.\n",version, port_letter);
	printk("iCom: ");
	printk(text); /* print text */
	/* print sixteen words of error data */
	printk("iCom: %.8x %.8x %.8x %.8x\n", tpr, error_data[0], error_data[1], error_data[2]);
	printk("iCom: %.8x %.8x %.8x %.8x\n", error_data[3], error_data[4], error_data[5], error_data[6]);
	printk("iCom: %.8x %.8x %.8x %.8x\n", error_data[7], error_data[8], error_data[9], error_data[10]);
	printk("iCom: %.8x %.8x %.8x %.8x\n", error_data[11], error_data[12], error_data[13], error_data[14]);

	printk("iCom: ERROR #################################################\n");
    }
    return;
}

/**************************************************/
/* Diagnostic Interrupt register checking routine */
/**************************************************/
static u32 __init diag_check_ints(struct iCom_adapter *iCom_adapter_ptr,
		    u32 exp_int1,		/* expected int one value */
		    u32 exp_int2,		/* expected int two value */
		    char *msg,			/* message to put in error log */
		    u32 tpr)			/* TPR to put in error log */
{
  u32		error_data[NUM_ERROR_ENTRIES];
  u32		*error_ptr = &error_data[0];

#define PORT_A_INT_MASK 0x0000FFFF
#define PORT_B_INT_MASK 0xFFFF0000


    u32	status = 0;	/* pass/fail */

    if ((iCom_adapter_ptr->version | ADAPTER_V2) != ADAPTER_V2){
    /* check port A */
	if((iCom_adapter_ptr->diag_int_reset1 & PORT_A_INT_MASK )
	   != (exp_int1 & PORT_A_INT_MASK))
	{
	    error_data[0] = exp_int1;
	    error_data[1] = iCom_adapter_ptr->diag_int_reset1;

	    diag_error(iCom_adapter_ptr,
		       0,  /* IOA_FAILURE indicates an Adapter failure */
		       msg, /* error message */
		       tpr,		/* tpr */
		       error_ptr,		/* error data pointer */
		       2);	/* number of error words. */
	    status = -1;
	}
    /* check port B */
	if((iCom_adapter_ptr->diag_int_reset1 & PORT_B_INT_MASK )
	   != (exp_int1 & PORT_B_INT_MASK))
	{
	    error_data[0] = exp_int1;
	    error_data[1] = iCom_adapter_ptr->diag_int_reset1;

	    diag_error(iCom_adapter_ptr,
		       1,  /* IOA_FAILURE indicates an Adapter failure */
		       msg, /* error message */
		       tpr,		/* tpr */
		       error_ptr,		/* error data pointer */
		       2);	/* number of error words. */
	    status = -1;
	}
    }
    else { /* this is a four port adapter */
	if((iCom_adapter_ptr->diag_int_reset1 & PORT_A_INT_MASK )
	   != (exp_int1 & PORT_A_INT_MASK))
	{
	    error_data[0] = exp_int1;
	    error_data[1] = iCom_adapter_ptr->diag_int_reset1;

	    diag_error(iCom_adapter_ptr,
		       0,  /* IOA_FAILURE indicates an Adapter failure */
		       msg, /* error message */
		       tpr,		/* tpr */
		       error_ptr,		/* error data pointer */
		       2);	/* number of error words. */
	    status = -1;
	}
    /* check port B */
	if((iCom_adapter_ptr->diag_int_reset1 & PORT_B_INT_MASK )
	   != (exp_int1 & PORT_B_INT_MASK))
	{
	    error_data[0] = exp_int1;
	    error_data[1] = iCom_adapter_ptr->diag_int_reset1;

	    diag_error(iCom_adapter_ptr,
		       1,  /* IOA_FAILURE indicates an Adapter failure */
		       msg, /* error message */
		       tpr,		/* tpr */
		       error_ptr,		/* error data pointer */
		       2);	/* number of error words. */
	    status = -1;
	}
        /* check port C */
	if((iCom_adapter_ptr->diag_int_reset2 & PORT_A_INT_MASK )
	   != (exp_int2 & PORT_A_INT_MASK))
	{
	    error_data[0] = exp_int2;
	    error_data[1] = iCom_adapter_ptr->diag_int_reset2;

	    diag_error(iCom_adapter_ptr,
		       2,  /* IOA_FAILURE indicates an Adapter failure */
		       msg, /* error message */
		       tpr,		/* tpr */
		       error_ptr,		/* error data pointer */
		       2);	/* number of error words. */
	    status = -1;
	}
    /* check port D */
	if((iCom_adapter_ptr->diag_int_reset2 & PORT_B_INT_MASK )
	   != (exp_int2 & PORT_B_INT_MASK))
	{
	    error_data[0] = exp_int2;
	    error_data[1] = iCom_adapter_ptr->diag_int_reset2;

	    diag_error(iCom_adapter_ptr,
		       3,  /* IOA_FAILURE indicates an Adapter failure */
		       msg, /* error message */
		       tpr,		/* tpr */
		       error_ptr,		/* error data pointer */
		       2);	/* number of error words. */
	    status = -1;
	}
    }
    return status;
}

/*************************************/
/* Diagnostic Initialization Routine */
/*************************************/
static u32 __init diag_init(struct iCom_adapter *iCom_adapter_ptr)
{
u8	i,j;
u32	rc;

/* init rc to zero */
rc = 0;

iCom_adapter_ptr->tpr = DIAG_INIT_ERROR_DATA;
/* set the TPR and error data to zero for the adapter */
for(j=0;j<NUM_ERROR_ENTRIES;j++)
{
    iCom_adapter_ptr->error_data[j] = 0;
}

/* set the TPR and error data to zero for all ports */
for(i=0;i<iCom_adapter_ptr->numb_ports;i++)
{
    iCom_adapter_ptr->port_info[i].tpr = 0;
    iCom_adapter_ptr->port_info[i].passed_diags = 0; /* false */
    for(j=0;j<NUM_ERROR_ENTRIES;j++)
    {
	iCom_adapter_ptr->port_info[i].error_data[j] = 0;
    }
}

/* Set the global register value */
iCom_adapter_ptr->tpr = DIAG_INIT_SET_GLOBAL_REGS;
if (iCom_adapter_ptr->version == ADAPTER_V1) {
    iCom_adapter_ptr->port_info[0].global_reg = (struct iCom_regs *)((char *)iCom_adapter_ptr->base_addr + 0x4000);
    iCom_adapter_ptr->port_info[1].global_reg = (struct iCom_regs *)((char *)iCom_adapter_ptr->base_addr + 0x4000);
}
else {
    iCom_adapter_ptr->port_info[0].global_reg = (struct iCom_regs *)((char *)iCom_adapter_ptr->base_addr + 0x8000);
    iCom_adapter_ptr->port_info[1].global_reg = (struct iCom_regs *)((char *)iCom_adapter_ptr->base_addr + 0x8000);
    iCom_adapter_ptr->port_info[2].global_reg = (struct iCom_regs *)((char *)iCom_adapter_ptr->base_addr + 0x8000);
    iCom_adapter_ptr->port_info[3].global_reg = (struct iCom_regs *)((char *)iCom_adapter_ptr->base_addr + 0x8000);
}

iCom_adapter_ptr->tpr = DIAG_INIT_END;
return rc;
}


/********************************/
/* Diagnostic Interrupt handler */
/********************************/
void __init diag_int_handler(int irq, void * dev_id, struct pt_regs * regs)
{
  u8                  temp_8;
  u32                 temp_32;
  struct iCom_adapter *iCom_adapter_ptr;
  /* find iCom_port_info for this interrupt */
  iCom_adapter_ptr = (struct iCom_adapter *)dev_id;


  /* if BAR0 is 0, then this must be a BIST */
  /* int so reset the config regs           */
  pci_read_config_dword(iCom_adapter_ptr->pci_dev, PCI_COMMAND, &temp_32);

  /* if memory space is not enabled, this must be the BIST interrupt */
  /* note that version 2 adapters will already be set up so it will  */
  /* skip this piece of code.                                        */
  if(!(temp_32 & 0x00000002))
  {
      /* restore the config registers */
      pci_write_config_dword(iCom_adapter_ptr->pci_dev,PCI_COMMAND, iCom_adapter_ptr->saved_command_reg);
      pci_write_config_dword(iCom_adapter_ptr->pci_dev,0x44, 0x8300830A);
      pci_write_config_dword(iCom_adapter_ptr->pci_dev, PCI_BASE_ADDRESS_0, iCom_adapter_ptr->saved_bar);
  }

  iCom_adapter_ptr->tpr = DIAG_INTHANDLE_START;
  if ((iCom_adapter_ptr->version | ADAPTER_V2) == ADAPTER_V2) {
    iCom_adapter_ptr->diag_int1 = iCom_readl((void*)(iCom_adapter_ptr->base_addr+0x8004));
    iCom_adapter_ptr->diag_int2 = iCom_readl((void*)(iCom_adapter_ptr->base_addr+0x8024));
    iCom_adapter_ptr->diag_int_pri1 = iCom_readl((void*)(iCom_adapter_ptr->base_addr+0x800C));
    iCom_adapter_ptr->diag_int_pri2 = iCom_readl((void*)(iCom_adapter_ptr->base_addr+0x802C));
  }
  else{
    iCom_adapter_ptr->diag_int1 = iCom_readl((void*)(iCom_adapter_ptr->base_addr+0x4004));
    iCom_adapter_ptr->diag_int2 = 0; /* does not exist */
    iCom_adapter_ptr->diag_int_pri1 = iCom_readl((void*)(iCom_adapter_ptr->base_addr+0x400C));
    iCom_adapter_ptr->diag_int_pri2 = 0x80; /* no int */
  }

  /* if pri1 has an int set, clear that int */
  iCom_adapter_ptr->tpr = DIAG_INTHANDLE_CHECK_FOR_INTS;
  if(iCom_adapter_ptr->diag_int_pri1 != 0x80) { /* 0x80 means "no int" */
    temp_8 = iCom_adapter_ptr->diag_int_pri1 >> 2; /* determine how many bits to shift left */
    temp_32 = 1;
    temp_32 <<= temp_8; /* shift a 1 that many bits */
    iCom_adapter_ptr->diag_int_reset1 |= temp_32; /* record the int being reset */
    if ((iCom_adapter_ptr->version | ADAPTER_V2) == ADAPTER_V2)
    {
	iCom_writel(temp_32,(void*)iCom_adapter_ptr->base_addr+0x8004); /* reset int */
    }
    else
      iCom_writel(temp_32,(void*)iCom_adapter_ptr->base_addr+0x4004); /* reset int */
  }
  else /* clear the int from pri2 */{
    temp_8 = iCom_adapter_ptr->diag_int_pri2 >> 2; /* determine how many bits to shift left */
    temp_32 = 1;
    temp_32 <<= temp_8; /* shift a 1 that many bits */
    iCom_adapter_ptr->diag_int_reset2 |= temp_32; /* record the int being reset */
    iCom_writel(temp_32,(void*)iCom_adapter_ptr->base_addr+0x8024); /* reset int */
  }
  iCom_adapter_ptr->tpr = DIAG_INTHANDLE_END;
  return;

}


/************************/
/* Diagnostic BIST Test */
/************************/
static u32 __init diag_bist(struct iCom_adapter *iCom_adapter_ptr)
{

  u32           rc;
  u32           temp_32;
  u8            i;
  u32		error_data[NUM_ERROR_ENTRIES];
  u32		*error_ptr = &error_data[0];
  int		done;

  iCom_adapter_ptr->tpr = DIAG_BIST_CLEAR_INTS;
  /* make sure the all ints are cleared before we start */
  if ((iCom_adapter_ptr->version | ADAPTER_V2) == ADAPTER_V2) {
    iCom_writel(0xFFFFFFFF,(void*)(iCom_adapter_ptr->base_addr+0x8004)); 
    iCom_writel(0xFFFFFFFF,(void*)(iCom_adapter_ptr->base_addr+0x8024));
  }
  else{
    iCom_writel(0xFFFFFFFF,(void*)(iCom_adapter_ptr->base_addr+0x4004)); 
  }

  /* version 2 adapters must not cause a real int because */
  /* accesses to any register immediately after BIST will */
  /* fail. We must first mask the ints, cause BIST to     */
  /* happen, then unmask the int to get to the interrupt  */
  /* handler. Note that version 1 adapters can not mask   */
  /* off the BIST int.                                    */
  if ((iCom_adapter_ptr->version | ADAPTER_V2) == ADAPTER_V2) {
      temp_32 = iCom_readl((void*)(iCom_adapter_ptr->base_addr+0x8008));
      temp_32 |= 0x80000000;
      iCom_writel(temp_32,(void*)(iCom_adapter_ptr->base_addr+0x8008));
  }

  /**** save registers that will have to be restored ****/
  
  /* get the control reg */
  iCom_adapter_ptr->tpr = DIAG_BIST_SAVE_CMD_REG;
  if((rc = pci_read_config_dword(iCom_adapter_ptr->pci_dev, PCI_COMMAND, &iCom_adapter_ptr->saved_command_reg)))
  {
      error_data[0] = rc;

      diag_error(iCom_adapter_ptr,
		 IOA_FAILURE,  /* IOA_FAILURE indicates an Adapter failure */
		 "iCom: Bist Test. Failed to read command reg.\n", /* error message */
		 DIAG_BIST_READ_CMD_REG_FAIL,		/* tpr */
		 error_ptr,		/* error data pointer */
		 1);	/* number of error words. */
      return 0x0000000f; /* fail the whole adapter */
  }

  /* get the memory bar reg */
  iCom_adapter_ptr->tpr = DIAG_BIST_SAVE_BAR0_REG;
  if((rc = pci_read_config_dword(iCom_adapter_ptr->pci_dev, PCI_BASE_ADDRESS_0, &iCom_adapter_ptr->saved_bar)))
  {
      error_data[0] = rc;

      diag_error(iCom_adapter_ptr,
		 IOA_FAILURE,  /* IOA_FAILURE indicates an Adapter failure */
		 "iCom: Bist Test. Failed to read BAR 0.\n", /* error message */
		 DIAG_BIST_READ_BAR0_FAIL,		/* tpr */
		 error_ptr,		/* error data pointer */
		 1);	/* number of error words. */
      return 0x0000000f; /* fail the whole adapter */
  }

  /**** hook in the interrupt handler ****/

  iCom_adapter_ptr->diag_int1 = 0; /* init */
  
  /* save off irq and request irq line */
  iCom_adapter_ptr->tpr = DIAG_BIST_HOOK_INT_LINE;
  if ((rc = request_irq(iCom_adapter_ptr->pci_dev->irq, diag_int_handler, SA_INTERRUPT |
                  SA_SHIRQ, ICOM_DRIVER_NAME, (void *)iCom_adapter_ptr))) {
      error_data[0] = rc;

      diag_error(iCom_adapter_ptr,
		 IOA_FAILURE,  /* IOA_FAILURE indicates an Adapter failure */
		 "iCom: Bist Test. Failed to hook int handler.\n", /* error message */
		 DIAG_BIST_HOOK_INT_HANDLER_FAIL,		/* tpr */
		 error_ptr,		/* error data pointer */
		 1);	/* number of error words. */
      return 0x0000000f; /* fail the whole adapter */
  }
  else
      iCom_adapter_ptr->resources |= HAVE_INT_HANDLE;

  /**** start bist ****/
  iCom_adapter_ptr->tpr = DIAG_BIST_START_BIST;
  if((rc = pci_write_config_dword(iCom_adapter_ptr->pci_dev, 0x0C, 0x40000000))) {
    printk("iCom: Diagnostic failed. CD700230, %8x.\n", rc);
    error_data[0] = rc;

    diag_error(iCom_adapter_ptr,
	       IOA_FAILURE,  /* IOA_FAILURE indicates an Adapter failure */
	       "iCom: Bist Test. Failed to write BIST reg.\n", /* error message */
	       DIAG_BIST_START_BIST_FAIL, /* tpr */
	       error_ptr,		/* error data pointer */
	       1);	/* number of error words. */
    return 0x0000000f;
  }

  /* set done to zero before the check for version 2 below */
  done = 0;

  /* restore the config registers on version 2 adapters */
  if ((iCom_adapter_ptr->version | ADAPTER_V2) == ADAPTER_V2) {

    /**** wait 2 seconds ****/
    current->state = TASK_INTERRUPTIBLE;
    schedule_timeout(HZ*2); /* HZ = 1 second */

    /* restore the config registers */
    pci_write_config_dword(iCom_adapter_ptr->pci_dev,PCI_COMMAND, iCom_adapter_ptr->saved_command_reg);
    pci_write_config_dword(iCom_adapter_ptr->pci_dev,0x44, 0x42004200);
    pci_write_config_dword(iCom_adapter_ptr->pci_dev,0x48, 0x42004200);
    pci_write_config_dword(iCom_adapter_ptr->pci_dev, PCI_BASE_ADDRESS_0, iCom_adapter_ptr->saved_bar);

    /* now clear the mask bit to force the interrupt to happen */
    temp_32 = iCom_readl((void*)(iCom_adapter_ptr->base_addr+0x8008));
    temp_32 = temp_32 & 0x7FFFFFFF;
    iCom_writel(temp_32,(void*)(iCom_adapter_ptr->base_addr+0x8008));
    current->state = TASK_INTERRUPTIBLE;
    schedule_timeout(HZ/10); /* HZ = 1 second */
    done = 1;
  }

  /**** wait for 6 seconds or int ****/
  iCom_adapter_ptr->tpr = DIAG_BIST_WAIT_FOR_INT;
  for(i=0;((i<60) && (done==0));i++)
  {
    if(iCom_adapter_ptr->diag_int1 == 0)
    {
      /* wait 100ms */
      current->state = TASK_INTERRUPTIBLE;
      schedule_timeout(HZ/10); /* HZ = 1 second */
    }
    else
    {
	done = 1;
    }
  }

  /* if not passed */
  iCom_adapter_ptr->tpr = DIAG_BIST_CHECK_INT_REGS;
  if(iCom_adapter_ptr->diag_int1 != 0x80000000) {
      error_data[0] = iCom_adapter_ptr->diag_int1;
      error_data[1] = i;

      diag_error(iCom_adapter_ptr,
		 IOA_FAILURE,  /* IOA_FAILURE indicates an Adapter failure */
		 "iCom: Bist Test. Bist int did not occur.\n", /* error message */
		 DIAG_BIST_NO_BIST_INT_FAIL,		/* tpr */
		 error_ptr,		/* error data pointer */
		 2);	/* number of error words. */
      return 0x0000000f; /* fail the whole adapter */
  }
  
  /**** see if bist completed successfully ****/
  iCom_adapter_ptr->tpr = DIAG_CHECK_READ_BIST_REG;
  if((rc = pci_read_config_dword(iCom_adapter_ptr->pci_dev, 0x0C, &temp_32)))
  {
      error_data[0] = rc;

      diag_error(iCom_adapter_ptr,
		 IOA_FAILURE,  /* IOA_FAILURE indicates an Adapter failure */
		 "iCom: Bist Test. Unable to read config offset 0xC.\n", /* error message */
		 DIAG_BIST_READ_CONFIG_OFFSET_0C_FAIL,		/* tpr */
		 error_ptr,		/* error data pointer */
		 1);	/* number of error words. */
      return 0x0000000f; /* fail the whole adapter */
  }

  /**** check tha the MISR and PRPG are not set ****/
  iCom_adapter_ptr->tpr = DIAG_BIST_CHECK_MISR_AND_PRPG;
  if(temp_32 & 0x03000000) {
      error_data[0] = temp_32;

      diag_error(iCom_adapter_ptr,
		 IOA_FAILURE,  /* IOA_FAILURE indicates an Adapter failure */
		 "iCom: Bist Test. Incorrect MISR or PRPG.\n", /* error message */
		 DIAG_BIST_INCORRECT_MISR_OR_PRPG,		/* tpr */
		 error_ptr,		/* error data pointer */
		 1);	/* number of error words. */
      return 0x0000000f; /* fail the whole adapter */
  }

  /**** return int handler ****/
  iCom_adapter_ptr->tpr = DIAG_BIST_RETURN_INT_HANDLE;
  free_irq(iCom_adapter_ptr->pci_dev->irq, (void *)iCom_adapter_ptr);
  iCom_adapter_ptr->resources &= ~HAVE_INT_HANDLE;

  return 0; /* pass */
}

/***************/
/* memory test */
/***************/
static u32 __init mem_test(struct iCom_adapter *iCom_adapter_ptr,
	     u32 port,
	     unsigned long addr,
	     u32 len)
{
  u8            data;
  u8		actual_data;
  u32           i;
  u32		error_data[NUM_ERROR_ENTRIES];
  u32		*error_ptr = &error_data[0];

  /**** addr tag test ****/

  /* set the memory */
  iCom_adapter_ptr->tpr = DIAG_MEM_WRITE_ADDR_TAG;
  data = 0;
  for(i=0;i<len;i++)
  {
    iCom_writeb(data, (void*)(addr + i));
    ++data;
    if(data == 254)
      data = 0;
  }

  /* check the memory */
  iCom_adapter_ptr->tpr = DIAG_MEM_CHECK_ADDR_TAG;
  data = 0;
  for(i=0;i<len;i++)
  {
      actual_data = iCom_readb((void*)(addr + i));

      if(actual_data != data)
      {
	  error_data[0] = (u32)data;
	  error_data[1] = (u32)actual_data;
	  error_data[2] = (u32)addr;
	  error_data[3] = (u32)addr+i;

	  diag_error(iCom_adapter_ptr,
		     port,  /* IOA_FAILURE indicates an Adapter failure */
		     "iCom: Memory Test. Address Tag Test Failed.\n", /* error message */
		     DIAG_MEM_ADDR_TAG_TEST_FAILED,		/* tpr */
		     error_ptr,		/* error data pointer */
		     4);	/* number of error words. */
	  return -1; /* fail the whole adapter */
      }
      ++data;
      if(data == 254)
	  data = 0;
  }

  /**** write A's ****/
  
  iCom_adapter_ptr->tpr = DIAG_MEM_WRITE_AA;
  for(i=0;i<len;i++)
  {
    iCom_writeb(0xAA, (void*)(addr + i));
  }
  
  iCom_adapter_ptr->tpr = DIAG_MEM_CHECK_AA;
  for(i=0;i<len;i++)
  {
    actual_data = iCom_readb((void*)(addr + i));
    if(actual_data != 0xAA)
    {
	error_data[0] = 0x000000AA;
	error_data[1] = (u32)actual_data;
	error_data[2] = (u32)addr;
	error_data[3] = (u32)addr+i;

	diag_error(iCom_adapter_ptr,
		   port,  /* IOA_FAILURE indicates an Adapter failure */
		   "iCom: Memory Test. 0xAA Test Failed.\n", /* error message */
		   DIAG_MEM_AA_TEST_FAIL,		/* tpr */
		   error_ptr,		/* error data pointer */
		   4);	/* number of error words. */
	return -1;
    }
    iCom_writeb(0x55, (void*)(addr + i));
  }

  /**** check for 55's ****/
  
  iCom_adapter_ptr->tpr = DIAG_MEM_CHECK_55;
  for(i=0;i<len;i++)
  {
    actual_data = iCom_readb((void*)(addr + i));
    if(actual_data != 0x55)
    {
	error_data[0] = 0x00000055;
	error_data[1] = (u32)actual_data;
	error_data[2] = (u32)addr;
	error_data[3] = (u32)addr+i;

	diag_error(iCom_adapter_ptr,
		   port,  /* IOA_FAILURE indicates an Adapter failure */
		   "iCom: Memory Test. 0x55 Test Failed.\n", /* error message */
		   DIAG_MEM_55_TEST_FAIL,		/* tpr */
		   error_ptr,		/* error data pointer */
		   4);	/* number of error words. */
	return -1;
    }
    iCom_writeb(0, (void*)(addr + i));
  }

  /**** check for 00's ****/
  
  iCom_adapter_ptr->tpr = DIAG_MEM_CHECK_00;
  for(i=0;i<len;i++)
  {
    actual_data = iCom_readb((void*)(addr + i));
    if(actual_data != 0)
    {
	error_data[0] = 0x00000000;
	error_data[1] = (u32)actual_data;
	error_data[2] = (u32)addr;
	error_data[3] = (u32)addr+i;

	diag_error(iCom_adapter_ptr,
		   port,  /* IOA_FAILURE indicates an Adapter failure */
		   "iCom: Memory Test. 0x00 Test Failed.\n", /* error message */
		   DIAG_MEM_00_TEST_FAIL,		/* tpr */
		   error_ptr,		/* error data pointer */
		   4);	/* number of error words. */
	return -1;
    }
  }
return 0;
}


/**************************/
/* Diagnostic Memory test */
/**************************/
static u32 __init diag_memory(struct iCom_adapter *iCom_adapter_ptr)
{
  u8            i;
  u32           port;
  u32           rc;
  unsigned long mem1_addr, mem2_addr;
  u32           mem1_len, mem2_len;
  struct iCom_port    *iCom_port_info;
  u32		error_data[NUM_ERROR_ENTRIES];
  u32		*error_ptr = &error_data[0];


  /* for the number of ports on adapter */
  iCom_adapter_ptr->tpr = DIAG_MEM_FOR_EVERY_PORT;
  for(i=0;i<iCom_adapter_ptr->numb_ports;i++)
  {

  /* Stop processor to make sure the instruction ram can be tested */
      iCom_adapter_ptr->tpr = DIAG_MEM_STOP_PROC;
      iCom_port_info = &iCom_adapter_ptr->port_info[i];
      stop_processor(iCom_port_info);

      switch (i)
      {
	  case 0: /* this port exists on all adpaters */
	      port = 1; /* bit for port 0 */
	      mem1_addr = iCom_adapter_ptr->base_addr;
	      mem1_len = 0x200;
	      mem2_addr = iCom_adapter_ptr->base_addr + 0x1000;
	      if(iCom_adapter_ptr->version == ADAPTER_V1)
		  mem2_len = 0xc00;
	      else
		  mem2_len = 0x1000;
	      break;

	  case 1:
	      if(iCom_adapter_ptr->port_info[1].status == ICOM_PORT_ACTIVE) {
		  port = 2; /* bit for port 1 */
		  mem1_addr = iCom_adapter_ptr->base_addr + 0x2000;
		  mem1_len = 0x200;
		  mem2_addr = iCom_adapter_ptr->base_addr + 0x3000;
		  if(iCom_adapter_ptr->version == ADAPTER_V1)
		      mem2_len = 0xc00;
		  else
		      mem2_len = 0x1000;
	      }
	      else
		  continue;
	      break;

	  case 2: /* if four ports exist, this one will be valid */
	      port = 4; /* bit for port 2 */
	      mem1_addr = iCom_adapter_ptr->base_addr + 0x4000;
	      mem1_len = 0x200;
	      mem2_addr = iCom_adapter_ptr->base_addr + 0x5000;
	      mem2_len = 0x1000;
	      break;

	  case 3:
	      if(iCom_adapter_ptr->port_info[3].status == ICOM_PORT_ACTIVE) {
		  port = 8; /* bit for port 3 */
		  mem1_addr = iCom_adapter_ptr->base_addr + 0x6000;
		  mem1_len = 0x200;
		  mem2_addr = iCom_adapter_ptr->base_addr + 0x7000;
		  mem2_len = 0x1000;
	      }
	      else
		  continue;
	      break;
	  default:
	      {
		  error_data[0] = i;

		  diag_error(iCom_adapter_ptr,
			     IOA_FAILURE,  /* IOA_FAILURE indicates an Adapter failure */
			     "iCom: Memory Test. Invalid Port Select.\n", /* error message */
			     DIAG_MEM_INVALID_PORT_SELECT,		/* tpr */
			     error_ptr,		/* error data pointer */
			     1);	/* number of error words. */
		  return -1;
	      }

      } /* end of switch */

    /* test the memory */
      iCom_adapter_ptr->tpr = DIAG_MEM_CALL_MEM_TEST;
      if((rc = mem_test(iCom_adapter_ptr, port, mem1_addr, mem1_len)))
	  return port;
      if((rc = mem_test(iCom_adapter_ptr, port, mem2_addr, mem2_len)))
	  return port;

  } /* end of for loop */

  /* pass */
  return 0;

} /* end of function */

/*****************************/
/* Diagnostic Port Load Code */
/*****************************/
void __init diag_port_load(int port,
		    struct iCom_port *iCom_port_info,
		    long int base_addr,
		    unsigned char *code,
		    int size)
{
    char               *iram_ptr;
    int                index;
    char              *dram_ptr;


    /* point to the DRAM */
    dram_ptr = (char*)(base_addr + 0x2000*port);


  /* Stop processor */
    iCom_port_info->tpr = DIAG_LOAD_STOP_PROC;
    stop_processor(iCom_port_info);

    /* Zero out DRAM */
    iCom_port_info->tpr = DIAG_LOAD_ZERO_DRAM;
    for (index = 0; index < 512; index++)
    {
	iCom_writeb(0x00,&dram_ptr[index]);
    }

    /* Load Code into Adapter */
    iCom_port_info->tpr = DIAG_LOAD_LOAD_CODE;
    iram_ptr = (char *)dram_ptr + ICOM_IRAM_OFFSET;
    for (index = 0; index < size; index++)
    {
	iCom_writeb(code[index],&iram_ptr[index]);
    }

    iCom_port_info->tpr = DIAG_LOAD_DONE;
    iCom_port_info->tpr = 0; /* must set this to zero when done */
                             /* for IfNoError to work.          */
    return ;
} 
 
/*****************************/
/* Diagnostic Interrupt Test */
/*****************************/
static u32 __init diag_int(struct iCom_adapter *iCom_adapter_ptr)
{

  u32           temp_32;
  u32           rc;
  u8            i;
  u32           exp_int1, exp_int2;
  u8            done;
  u32		error_data[NUM_ERROR_ENTRIES];
  u32		*error_ptr = &error_data[0];
  
  /* make sure the all ints are cleared before we start */
  iCom_adapter_ptr->tpr = DIAG_INT_CLEAR_INTS;
  if ((iCom_adapter_ptr->version | ADAPTER_V2) == ADAPTER_V2) {
    temp_32 = iCom_readl((void*)(iCom_adapter_ptr->base_addr+0x8004));
    iCom_writel(temp_32,(void*)(iCom_adapter_ptr->base_addr+0x8004)); 
    temp_32 = iCom_readl((void*)(iCom_adapter_ptr->base_addr+0x8024));
    iCom_writel(temp_32,(void*)(iCom_adapter_ptr->base_addr+0x8024)); 
  }
  else{
    temp_32 = iCom_readl((void*)(iCom_adapter_ptr->base_addr+0x4004));
    iCom_writel(temp_32,(void*)(iCom_adapter_ptr->base_addr+0x4004)); 
  }
  
  /* save off irq and request irq line */
  iCom_adapter_ptr->tpr = DIAG_INT_GET_INT_HANDLE;
  if ((rc = request_irq(iCom_adapter_ptr->pci_dev->irq, diag_int_handler, SA_INTERRUPT |
		       SA_SHIRQ, ICOM_DRIVER_NAME, (void *)iCom_adapter_ptr))) {
      error_data[0] = rc;

      diag_error(iCom_adapter_ptr,
		 IOA_FAILURE,  /* IOA_FAILURE indicates an Adapter failure */
		 "iCom: Interrupt Test. request_irq call failed.\n", /* error message */
		 DIAG_INT_GET_IRQ_LINE_FAIL,		/* tpr */
		 error_ptr,		/* error data pointer */
		 1);	/* number of error words. */
      return 0x0000000f;
  }
  else
      iCom_adapter_ptr->resources |= HAVE_INT_HANDLE;

  /* load all active procs */
  iCom_adapter_ptr->tpr = DIAG_INT_LOAD_PROCS;
  for(i=0;i<iCom_adapter_ptr->numb_ports;i++)
    if(iCom_adapter_ptr->port_info[i].status == ICOM_PORT_ACTIVE)
	diag_port_load(i,
		       &iCom_adapter_ptr->port_info[i],
                       iCom_adapter_ptr->base_addr,
		       topcidiag, 
		       sizeof(topcidiag));

  /* clear int regs */
  iCom_adapter_ptr->tpr = DIAG_INT_CLEAR_INT_VARS;
  iCom_adapter_ptr->diag_int_reset1 = 0;
  iCom_adapter_ptr->diag_int_reset2 = 0;

  /* unmask all interrupts */
  if ((iCom_adapter_ptr->version | ADAPTER_V2) != ADAPTER_V2)
      iCom_writel(0,(void*)(iCom_adapter_ptr->base_addr+0x4008));
  else
  {
      iCom_writel(0,(void*)(iCom_adapter_ptr->base_addr+0x8008));
      iCom_writel(0,(void*)(iCom_adapter_ptr->base_addr+0x8028));
  }
      
  

  /* Start all procs */
  iCom_adapter_ptr->tpr = DIAG_INT_START_PROCS;
  for(i=0;i<iCom_adapter_ptr->numb_ports;i++)
    if(iCom_adapter_ptr->port_info[i].status == ICOM_PORT_ACTIVE)
    {
	start_processor(&iCom_adapter_ptr->port_info[i]);
    }
  
  /* Start all tests */
  iCom_adapter_ptr->tpr = DIAG_INT_START_TESTS;
  for(i=0;i<iCom_adapter_ptr->numb_ports;i++)
    if(iCom_adapter_ptr->port_info[i].status == ICOM_PORT_ACTIVE)
    {
	/* 0x06 is the interrupt test */
	iCom_writeb((0x06),(void*)(iCom_adapter_ptr->base_addr+i*0x2000)); 
    }

  /* determine what ints to look for */
  iCom_adapter_ptr->tpr = DIAG_INT_DETERMINE_INTS_TO_CHECK_FOR;
  if ((iCom_adapter_ptr->version | ADAPTER_V2) != ADAPTER_V2) {
    exp_int1 = (u32)0x3FFF3FFF;
    exp_int2 = 0;
  }
  else {
    exp_int1 = 0x00003FFF;
    exp_int2 = 0x00003FFF;
    if(iCom_adapter_ptr->port_info[1].status == ICOM_PORT_ACTIVE)
      exp_int1 |= 0x3FFF0000;
    if(iCom_adapter_ptr->port_info[3].status == ICOM_PORT_ACTIVE)
      exp_int2 |= 0x3FFF0000;
  }



  /* wait for all ints to set or for a second to pass */
  done = 0;
  temp_32 = 0;
  iCom_adapter_ptr->tpr = DIAG_INT_WAIT_FOR_INTS;
  while((!done) && (temp_32 < 100))
  {
    if(((iCom_adapter_ptr->diag_int_reset1 & 0x3fff3fff) == exp_int1) &&
       ((iCom_adapter_ptr->diag_int_reset2 & 0x3fff3fff) == exp_int2))
      done = 1;

    /**** wait 1 millisecond ****/
    current->state = TASK_INTERRUPTIBLE;
    schedule_timeout(HZ/100); /* HZ = 1 second */
    temp_32++;
  }

  temp_32 = 0;

  iCom_adapter_ptr->tpr = DIAG_INT_CHECK_FOR_INTS;
  if(!done) /* failed */ 
  {
      /* fail the appropriate port/s */
      rc = diag_check_ints(iCom_adapter_ptr,
			   exp_int1,
			   exp_int2,
			   "Interrupt Test. Int did not occur.\n",
			   DIAG_INT_WRONG_INT_FAIL);
      return -1;
  }

  
  /* stop all procs */
  iCom_adapter_ptr->tpr = DIAG_INT_STOP_PROCS;
  for(i=0;i<iCom_adapter_ptr->numb_ports;i++)
    if(iCom_adapter_ptr->port_info[i].status == ICOM_PORT_ACTIVE)
      stop_processor(&iCom_adapter_ptr->port_info[i]);

  /**** return int handler ****/
  iCom_adapter_ptr->tpr = DIAG_INT_FREE_INT_HANDLE;
  free_irq(iCom_adapter_ptr->pci_dev->irq, (void *)iCom_adapter_ptr);
  iCom_adapter_ptr->resources &= ~HAVE_INT_HANDLE;

  iCom_adapter_ptr->tpr = DIAG_INT_END;
return 0; 
} 
 
/*******************************/
/* Diagnostic Instruction Test */
/*******************************/
static u32 __init diag_inst(struct iCom_adapter *iCom_adapter_ptr)
{

  u32           temp_32;
  u32           rc;
  u8            i;
  u32           exp_int1, exp_int2;
  u8            done;
  u32		error_data[NUM_ERROR_ENTRIES];
  u32		*error_ptr = &error_data[0];
  dma_addr_t	dma_addr_pci;
  int		addr1, addr2, addr3, addr4;
  
  /* make sure the all ints are cleared before we start */
  iCom_adapter_ptr->tpr = DIAG_INST_CLEAR_INTS;
  if ((iCom_adapter_ptr->version | ADAPTER_V2) == ADAPTER_V2) {
    temp_32 = iCom_readl((void*)(iCom_adapter_ptr->base_addr+0x8004));
    iCom_writel(temp_32,(void*)(iCom_adapter_ptr->base_addr+0x8004)); 
    temp_32 = iCom_readl((void*)(iCom_adapter_ptr->base_addr+0x8024));
    iCom_writel(temp_32,(void*)(iCom_adapter_ptr->base_addr+0x8024)); 
  }
  else{
    temp_32 = iCom_readl((void*)(iCom_adapter_ptr->base_addr+0x4004));
    iCom_writel(temp_32,(void*)(iCom_adapter_ptr->base_addr+0x4004)); 
  }
  
  /* save off irq and request irq line */
  iCom_adapter_ptr->tpr = DIAG_INST_GET_INT_HANDLE;
  if ((rc = request_irq(iCom_adapter_ptr->pci_dev->irq, diag_int_handler, SA_INTERRUPT |
		       SA_SHIRQ, ICOM_DRIVER_NAME, (void *)iCom_adapter_ptr))) {
      error_data[0] = rc;

      diag_error(iCom_adapter_ptr,
		 IOA_FAILURE,  /* IOA_FAILURE indicates an Adapter failure */
		 "iCom: Instruction Test. request_irq call failed.\n", /* error message */
		 DIAG_INST_GET_IRQ_LINE_FAIL,		/* tpr */
		 error_ptr,		/* error data pointer */
		 1);	/* number of error words. */
      return 0x0000000f;
  }
  else
      iCom_adapter_ptr->resources |= HAVE_INT_HANDLE;

  /* load all active procs */
  iCom_adapter_ptr->tpr = DIAG_INST_LOAD_PROCS;
  for(i=0;i<iCom_adapter_ptr->numb_ports;i++)
    if(iCom_adapter_ptr->port_info[i].status == ICOM_PORT_ACTIVE)
	diag_port_load(i,
		       &iCom_adapter_ptr->port_info[i],
                       iCom_adapter_ptr->base_addr,
		       toinstdiag, 
		       sizeof(toinstdiag));

  /* clear int regs */
  iCom_adapter_ptr->tpr = DIAG_INST_CLEAR_INT_VARS;
  iCom_adapter_ptr->diag_int_reset1 = 0;
  iCom_adapter_ptr->diag_int_reset2 = 0;

  /* get storage to DMA to and from. Need 8 bytes for each port */
  iCom_adapter_ptr->malloc_addr_1 = (u32 *)kmalloc(32,GFP_KERNEL | GFP_DMA);
  iCom_adapter_ptr->resources |= HAVE_MALLOC_1; /* indicate we have storage from malloc */
  dma_addr_pci = pci_map_single(iCom_adapter_ptr->pci_dev,
				iCom_adapter_ptr->malloc_addr_1,	/* addr to convert */
				32,		/* size */
				PCI_DMA_BIDIRECTIONAL);
  addr1 = (int)dma_addr_pci;
  addr2 = addr1+8;
  addr3 = addr1+16;
  addr4 = addr1+24;

  iCom_writel(addr1,(void*)(iCom_adapter_ptr->base_addr+8)); 
  iCom_writel(addr2,(void*)(iCom_adapter_ptr->base_addr+8+0x2000)); 
  if ((iCom_adapter_ptr->version | ADAPTER_V2) == ADAPTER_V2){ 
      iCom_writel(addr3,(void*)(iCom_adapter_ptr->base_addr+8+0x4000)); 
      iCom_writel(addr4,(void*)(iCom_adapter_ptr->base_addr+8+0x6000));
  }

  /* start all procs */
  iCom_adapter_ptr->tpr = DIAG_INST_START_PROCS;
  for(i=0;i<iCom_adapter_ptr->numb_ports;i++)
    if(iCom_adapter_ptr->port_info[i].status == ICOM_PORT_ACTIVE)
      start_processor(&iCom_adapter_ptr->port_info[i]);
  
  /* Start all tests */
  iCom_adapter_ptr->tpr = DIAG_INST_START_TESTS;
  for(i=0;i<iCom_adapter_ptr->numb_ports;i++)
    if(iCom_adapter_ptr->port_info[i].status == ICOM_PORT_ACTIVE)
	iCom_writeb((0x01),(void*)(iCom_adapter_ptr->base_addr+i*0x2000)); 

  /* determine what ints to look for */
  iCom_adapter_ptr->tpr = DIAG_INST_DETERMINE_INTS_TO_CHECK_FOR;
  if ((iCom_adapter_ptr->version | ADAPTER_V2) != ADAPTER_V2) {
    exp_int1 = (u32)0x01000100;
    exp_int2 = 0;
  }
  else {
      if(iCom_adapter_ptr->subsystem_id == FOUR_PORT_MODEL)
      {
	  exp_int1 = 0x01000100;
	  exp_int2 = 0x01000100;
      }
      else /* must be a two port model */
      {
	  exp_int1 = 0x00000100;
	  exp_int2 = 0x00000100;
      }
  }
  
  /* wait for all ints to set or for a second to pass */
  done = 0;
  temp_32 = 0;
  iCom_adapter_ptr->tpr = DIAG_INST_WAIT_FOR_INTS;
  while((!done) && (temp_32 < 1000))
  {
    if((iCom_adapter_ptr->diag_int_reset1 == exp_int1) &&
       (iCom_adapter_ptr->diag_int_reset2 == exp_int2))
      done = 1;

    /**** wait 1 millisecond ****/
    current->state = TASK_INTERRUPTIBLE;
    schedule_timeout(HZ/1000); /* HZ = 1 second */
    temp_32++;
  }

  temp_32 = 0;


  iCom_adapter_ptr->tpr = DIAG_INST_CHECK_FOR_INTS;
  if(!done) /* failed */
  {
      /* fail the appropriate port/s */
      rc = diag_check_ints(iCom_adapter_ptr,
			   exp_int1,
			   exp_int2,
			   "Instruction Test. Int did not occur.\n",
			   DIAG_INST_WRONG_INT_FAIL);
      return -1;
  }
  
  /* stop all procs */
  iCom_adapter_ptr->tpr = DIAG_INST_STOP_PROCS;
  for(i=0;i<iCom_adapter_ptr->numb_ports;i++)
    if(iCom_adapter_ptr->port_info[i].status == ICOM_PORT_ACTIVE)
      stop_processor(&iCom_adapter_ptr->port_info[i]);

  /**** check the results data ****/
   iCom_adapter_ptr->tpr = DIAG_INST_CHECK_RESULTS;
   for(i=0;i<iCom_adapter_ptr->numb_ports;i++)
       if(iCom_adapter_ptr->port_info[i].status == ICOM_PORT_ACTIVE)

       {
      /* read the results word */
	   temp_32 = iCom_readl((void*)(iCom_adapter_ptr->base_addr+(i*0x2000)+4));

      /* check results word */
	   if(temp_32 != 0)
	   {
	       error_data[0] = temp_32;

	       diag_error(iCom_adapter_ptr,	/* port to fail */
			  i,  /* IOA_FAILURE indicates an Adapter failure */
			  "iCom: Instruction Test. Bad results data.\n", /* error message */
			  DIAG_INST_BAD_RESULTS_DATA_FAIL,		/* tpr */
			  error_ptr,		/* error data pointer */
			  1);	/* number of error words. */
	       return 0x0000000f;
	   }
       }


  /**** return int handler ****/
  iCom_adapter_ptr->tpr = DIAG_INST_FREE_INT_HANDLE;
  free_irq(iCom_adapter_ptr->pci_dev->irq, (void *)iCom_adapter_ptr);
  iCom_adapter_ptr->resources &= ~HAVE_INT_HANDLE;

  /**** return storage from malloc ****/
  kfree(iCom_adapter_ptr->malloc_addr_1);
  iCom_adapter_ptr->resources &= ~HAVE_MALLOC_1;

  iCom_adapter_ptr->tpr = DIAG_INST_END;
return 0; 
} 

/****************************/
/* Diagnostic Register Test */
/****************************/
static u32 __init diag_reg(struct iCom_adapter *iCom_adapter_ptr)
{

  u32           temp_32;
  u32           rc;
  u8            i;
  u32           exp_int1, exp_int2;
  u8            done;
  u32		error_data[NUM_ERROR_ENTRIES];
  u32		*error_ptr = &error_data[0];
  u32		itteration;
  u8		port;


  /* make sure the all ints are cleared before we start */
      iCom_adapter_ptr->tpr = DIAG_REG_CLEAR_INTS+itteration;
      if ((iCom_adapter_ptr->version | ADAPTER_V2) == ADAPTER_V2) {
	  temp_32 = iCom_readl((void*)(iCom_adapter_ptr->base_addr+0x8004));
	  iCom_writel(temp_32,(void*)(iCom_adapter_ptr->base_addr+0x8004)); 
	  temp_32 = iCom_readl((void*)(iCom_adapter_ptr->base_addr+0x8024));
	  iCom_writel(temp_32,(void*)(iCom_adapter_ptr->base_addr+0x8024)); 
      }
      else{
	  temp_32 = iCom_readl((void*)(iCom_adapter_ptr->base_addr+0x4004));
	  iCom_writel(temp_32,(void*)(iCom_adapter_ptr->base_addr+0x4004)); 
      }

  /* save off irq and request irq line */
      iCom_adapter_ptr->tpr = DIAG_REG_GET_INT_HANDLE+itteration;
      if ((rc = request_irq(iCom_adapter_ptr->pci_dev->irq, diag_int_handler, SA_INTERRUPT |
			   SA_SHIRQ, ICOM_DRIVER_NAME, (void *)iCom_adapter_ptr))) {
	  error_data[0] = rc;

	  diag_error(iCom_adapter_ptr,
		     IOA_FAILURE,  /* IOA_FAILURE indicates an Adapter failure */
		     "iCom: Register Test. request_irq call failed.\n", /* error message */
		     DIAG_REG_GET_IRQ_LINE_FAIL+itteration,		/* tpr */
		     error_ptr,		/* error data pointer */
		     1);	/* number of error words. */
	  return 0x0000000f;
      }
      else
	  iCom_adapter_ptr->resources |= HAVE_INT_HANDLE;

  /* Do this test twice, once for each pico code load */
  for(itteration=0;itteration<200;itteration+=100)
  {

  /* load all active procs */
      iCom_adapter_ptr->tpr = DIAG_REG_LOAD_PROCS+itteration;
      for(i=0;i<iCom_adapter_ptr->numb_ports;i++)
      {
	  if(iCom_adapter_ptr->port_info[i].status == ICOM_PORT_ACTIVE)
	  {
	      if(itteration == 0) /* if first time through this code */
	      {
		  diag_port_load(i,
				 &iCom_adapter_ptr->port_info[i],
				 iCom_adapter_ptr->base_addr,
				 toregdiag, 
				 sizeof(toregdiag));
	      }
	      else
	      {
		  diag_port_load(i,
				 &iCom_adapter_ptr->port_info[i],
				 iCom_adapter_ptr->base_addr,
				 toreg2diag, 
				 sizeof(toreg2diag));
	      }
	  }
      }


  /* clear int regs */
      iCom_adapter_ptr->tpr = DIAG_REG_CLEAR_INT_VARS+itteration;
      iCom_adapter_ptr->diag_int_reset1 = 0;
      iCom_adapter_ptr->diag_int_reset2 = 0;

  /* if the port number is 1 or 3, put 1 in the port field(offset 8) */
  /* The pico code needs to know if it is the second port of two.    */
      for(i=0;i<iCom_adapter_ptr->numb_ports;i++)
      {
	  port = iCom_adapter_ptr->port_info[i].port;
	  if((port == 1) || (port == 3))
	      iCom_writel(1,(void*)(iCom_adapter_ptr->base_addr+(port*0x2000)+8));
      }

  /* start all procs */
      iCom_adapter_ptr->tpr = DIAG_REG_START_PROCS+itteration;
      for(i=0;i<iCom_adapter_ptr->numb_ports;i++)
	  if(iCom_adapter_ptr->port_info[i].status == ICOM_PORT_ACTIVE)
	      start_processor(&iCom_adapter_ptr->port_info[i]);

  /**** send the command to start the pico code ****/
      for(i=0;i<iCom_adapter_ptr->numb_ports;i++)
	  if(iCom_adapter_ptr->port_info[i].status == ICOM_PORT_ACTIVE)
	      iCom_writeb(1,(void*)(iCom_adapter_ptr->base_addr+(i*0x2000)));


  /**** determine what ints to look for ****/
   iCom_adapter_ptr->tpr = DIAG_REG_DETERMINE_INTS_TO_CHECK_FOR+itteration;
   if ((iCom_adapter_ptr->version | ADAPTER_V2) != ADAPTER_V2) {
       exp_int1 = (u32)0x01000100;
       exp_int2 = 0;
   }
  else {
      if(iCom_adapter_ptr->subsystem_id == FOUR_PORT_MODEL)
      {
	  exp_int1 = 0x01000100;
	  exp_int2 = 0x01000100;
      }
      else /* must be a two port model */
      {
	  exp_int1 = 0x00000100;
	  exp_int2 = 0x00000100;
      }
   }

  /**** wait for all ints to set or for a second to pass ****/
   done = 0;
   temp_32 = 0;
   iCom_adapter_ptr->tpr = DIAG_REG_WAIT_FOR_INTS+itteration;
   while((!done) && (temp_32 < 1000))
   {
       if((iCom_adapter_ptr->diag_int_reset1 == exp_int1) &&
	  (iCom_adapter_ptr->diag_int_reset2 == exp_int2))
	   done = 1;

    /**** wait 1 millisecond ****/
       current->state = TASK_INTERRUPTIBLE;
       schedule_timeout(HZ/1000); /* HZ = 1 second */
       temp_32++;
   }

   temp_32 = 0;

   iCom_adapter_ptr->tpr = DIAG_REG_CHECK_FOR_INTS+itteration;

   if(!done) /* failed */ 
   {
      /* fail the appropriate port/s */
       rc = diag_check_ints(iCom_adapter_ptr,
			    exp_int1,
			    exp_int2,
			    "Register Test. Int did not occur.\n",
			    DIAG_REG_WRONG_INT_FAIL);
       return -1;
   }

  /**** stop all procs ****/
   iCom_adapter_ptr->tpr = DIAG_REG_STOP_PROCS+itteration;
   for(i=0;i<iCom_adapter_ptr->numb_ports;i++)
       if(iCom_adapter_ptr->port_info[i].status == ICOM_PORT_ACTIVE)
	   stop_processor(&iCom_adapter_ptr->port_info[i]);

  /**** check the results data ****/
   iCom_adapter_ptr->tpr = DIAG_REG_CHECK_RESULTS+itteration;
   for(i=0;i<iCom_adapter_ptr->numb_ports;i++)
       if(iCom_adapter_ptr->port_info[i].status == ICOM_PORT_ACTIVE)

       {
      /* read the results word */
	   temp_32 = iCom_readl((void*)(iCom_adapter_ptr->base_addr+(i*0x2000)+4));

      /* check results word */
	   if(temp_32 != 0)
	   {
	       error_data[0] = temp_32;

	       diag_error(iCom_adapter_ptr,	/* port to fail */
			  i,  /* IOA_FAILURE indicates an Adapter failure */
			  "iCom: Register Test. Bad results data.\n", /* error message */
			  DIAG_REG_BAD_RESULTS_DATA_FAIL+itteration,		/* tpr */
			  error_ptr,		/* error data pointer */
			  1);	/* number of error words. */
	       return 0x0000000f;
	   }
       }
  } /* end of "for each pico code load" */

  /**** return int handler ****/
  iCom_adapter_ptr->tpr = DIAG_REG_FREE_INT_HANDLE+itteration;
  free_irq(iCom_adapter_ptr->pci_dev->irq, (void *)iCom_adapter_ptr);
  iCom_adapter_ptr->resources &= ~HAVE_INT_HANDLE;

  iCom_adapter_ptr->tpr = DIAG_REG_END+itteration;

return 0; 
} 
 
/***************************************/
/* Diagnostic RVX Status Register Test */
/***************************************/
static u32 __init diag_rvx_sts(struct iCom_adapter *iCom_adapter_ptr)
{

  u32           temp_32;
  u32           rc;
  u8            i;
  u32           exp_int1, exp_int2;
  u8            done;
  u32		error_data[NUM_ERROR_ENTRIES];
  u32		*error_ptr = &error_data[0];
  
  /* make sure the all ints are cleared before we start */
  iCom_adapter_ptr->tpr = DIAG_RVXSTS_CLEAR_INTS;
  if ((iCom_adapter_ptr->version | ADAPTER_V2) == ADAPTER_V2) {
    temp_32 = iCom_readl((void*)(iCom_adapter_ptr->base_addr+0x8004));
    iCom_writel(temp_32,(void*)(iCom_adapter_ptr->base_addr+0x8004)); 
    temp_32 = iCom_readl((void*)(iCom_adapter_ptr->base_addr+0x8024));
    iCom_writel(temp_32,(void*)(iCom_adapter_ptr->base_addr+0x8024)); 
  }
  else{
    temp_32 = iCom_readl((void*)(iCom_adapter_ptr->base_addr+0x4004));
    iCom_writel(temp_32,(void*)(iCom_adapter_ptr->base_addr+0x4004)); 
  }
  
  /* save off irq and request irq line */
  iCom_adapter_ptr->tpr = DIAG_RVXSTS_GET_INT_HANDLE;
  if ((rc = request_irq(iCom_adapter_ptr->pci_dev->irq, diag_int_handler, SA_INTERRUPT |
		       SA_SHIRQ, ICOM_DRIVER_NAME, (void *)iCom_adapter_ptr))) {
      error_data[0] = rc;

      diag_error(iCom_adapter_ptr,
		 IOA_FAILURE,  /* IOA_FAILURE indicates an Adapter failure */
		 "iCom: RVX Status Register Test. request_irq call failed.\n", /* error message */
		 DIAG_RVXSTS_GET_IRQ_LINE_FAIL,		/* tpr */
		 error_ptr,		/* error data pointer */
		 1);	/* number of error words. */
      return 0x0000000f;
  }
  else
      iCom_adapter_ptr->resources |= HAVE_INT_HANDLE;

  /* load all active procs */
  iCom_adapter_ptr->tpr = DIAG_RVXSTS_LOAD_PROCS;
  for(i=0;i<iCom_adapter_ptr->numb_ports;i++)
    if(iCom_adapter_ptr->port_info[i].status == ICOM_PORT_ACTIVE)
	diag_port_load(i,
		       &iCom_adapter_ptr->port_info[i],
                       iCom_adapter_ptr->base_addr,
		       tostsdiag, 
		       sizeof(tostsdiag));

  /* clear int regs */
  iCom_adapter_ptr->tpr = DIAG_RVXSTS_CLEAR_INT_VARS;
  iCom_adapter_ptr->diag_int_reset1 = 0;
  iCom_adapter_ptr->diag_int_reset2 = 0;

  /* start all procs */
  iCom_adapter_ptr->tpr = DIAG_RVXSTS_START_PROCS;
  for(i=0;i<iCom_adapter_ptr->numb_ports;i++)
    if(iCom_adapter_ptr->port_info[i].status == ICOM_PORT_ACTIVE)
      start_processor(&iCom_adapter_ptr->port_info[i]);
  
  /* determine what ints to look for */
  iCom_adapter_ptr->tpr = DIAG_RVXSTS_DETERMINE_INTS_TO_CHECK_FOR;
  if ((iCom_adapter_ptr->version | ADAPTER_V2) != ADAPTER_V2) {
    exp_int1 = 0x01000100;
    exp_int2 = 0;
  }
  else
  {
      if(iCom_adapter_ptr->subsystem_id == FOUR_PORT_MODEL)
      {
	  exp_int1 = 0x01000100;
	  exp_int2 = 0x01000100;
      }
      else /* must be a two port model */
      {
	  exp_int1 = 0x00000100;
	  exp_int2 = 0x00000100;
      }
  }
  
  /* wait for all ints to set or for a second to pass */
  done = 0;
  temp_32 = 0;
  iCom_adapter_ptr->tpr = DIAG_RVXSTS_WAIT_FOR_INTS;
  while((!done) && (temp_32 < 1000))
  {
    if((iCom_adapter_ptr->diag_int_reset1 == exp_int1) &&
       (iCom_adapter_ptr->diag_int_reset2 == exp_int2))
      done = 1;

    /**** wait 1 millisecond ****/
    current->state = TASK_INTERRUPTIBLE;
    schedule_timeout(HZ/1000); /* HZ = 1 second */
    temp_32++;
  }

  temp_32 = 0;

  iCom_adapter_ptr->tpr = DIAG_RVXSTS_CHECK_FOR_INTS;
  if(!done) /* failed */
  {
      /* fail the appropriate port/s */
      rc = diag_check_ints(iCom_adapter_ptr,
			   exp_int1,
			   exp_int2,
			   "Instruction Test. Int did not occur.\n",
			   DIAG_RVXSTS_WRONG_INT_FAIL);
      return -1;
  }
  
  /* stop all procs */
  iCom_adapter_ptr->tpr = DIAG_RVXSTS_STOP_PROCS;
  for(i=0;i<iCom_adapter_ptr->numb_ports;i++)
    if(iCom_adapter_ptr->port_info[i].status == ICOM_PORT_ACTIVE)
      stop_processor(&iCom_adapter_ptr->port_info[i]);

  /**** check the results data ****/
   iCom_adapter_ptr->tpr = DIAG_RVXSTS_CHECK_RESULTS;
   for(i=0;i<iCom_adapter_ptr->numb_ports;i++)
       if(iCom_adapter_ptr->port_info[i].status == ICOM_PORT_ACTIVE)
       {
      /* read the results word */
	   temp_32 = iCom_readl((void*)(iCom_adapter_ptr->base_addr+(i*0x2000)+4));

      /* check results word */
	   if(temp_32 != 0xAAAAAAAA) /* AAAAAAAA is passing */
	   {
	       error_data[0] = temp_32;

	       diag_error(iCom_adapter_ptr,	/* port to fail */
			  i,  /* IOA_FAILURE indicates an Adapter failure */
			  "iCom: Instruction Test. Bad results data.\n", /* error message */
			  DIAG_RVXSTS_BAD_RESULTS_DATA_FAIL,		/* tpr */
			  error_ptr,		/* error data pointer */
			  1);	/* number of error words. */
	       return 0x0000000f;
	   }
       }


  /**** return int handler ****/
  iCom_adapter_ptr->tpr = DIAG_RVXSTS_FREE_INT_HANDLE;
  free_irq(iCom_adapter_ptr->pci_dev->irq, (void *)iCom_adapter_ptr);
  iCom_adapter_ptr->resources &= ~HAVE_INT_HANDLE;

  iCom_adapter_ptr->tpr = DIAG_RVXSTS_END;
return 0; 
}

/*************************************/
/* Diagnostic RVX Internal Wrap Test */
/*************************************/
static u32 __init diag_rvx_internal_wrap(struct iCom_adapter *iCom_adapter_ptr)
{

  u32           temp_32;
  u32           rc;
  u8            i;
  u32           exp_int1, exp_int2;
  u8            done;
  u32		error_data[NUM_ERROR_ENTRIES];
  u32		*error_ptr = &error_data[0];
  
  /* make sure the all ints are cleared before we start */
  iCom_adapter_ptr->tpr = DIAG_RVXINTWRP_CLEAR_INTS;
  if ((iCom_adapter_ptr->version | ADAPTER_V2) == ADAPTER_V2) {
    temp_32 = iCom_readl((void*)(iCom_adapter_ptr->base_addr+0x8004));
    iCom_writel(temp_32,(void*)(iCom_adapter_ptr->base_addr+0x8004)); 
    temp_32 = iCom_readl((void*)(iCom_adapter_ptr->base_addr+0x8024));
    iCom_writel(temp_32,(void*)(iCom_adapter_ptr->base_addr+0x8024)); 
  }
  else{
    temp_32 = iCom_readl((void*)(iCom_adapter_ptr->base_addr+0x4004));
    iCom_writel(temp_32,(void*)(iCom_adapter_ptr->base_addr+0x4004)); 
  }
  
  /* save off irq and request irq line */
  iCom_adapter_ptr->tpr = DIAG_RVXINTWRP_GET_INT_HANDLE;
  if ((rc = request_irq(iCom_adapter_ptr->pci_dev->irq, diag_int_handler, SA_INTERRUPT |
		       SA_SHIRQ, ICOM_DRIVER_NAME, (void *)iCom_adapter_ptr))) {
      error_data[0] = rc;

      diag_error(iCom_adapter_ptr,
		 IOA_FAILURE,  /* IOA_FAILURE indicates an Adapter failure */
		 "iCom: RVX Internal Wrap Test. request_irq call failed.\n", /* error message */
		 DIAG_RVXINTWRP_GET_IRQ_LINE_FAIL,		/* tpr */
		 error_ptr,		/* error data pointer */
		 1);	/* number of error words. */
      return 0x0000000f;
  }
  else
      iCom_adapter_ptr->resources |= HAVE_INT_HANDLE;

  /* load all active procs */
  iCom_adapter_ptr->tpr = DIAG_RVXINTWRP_LOAD_PROCS;
  for(i=0;i<iCom_adapter_ptr->numb_ports;i++)
    if(iCom_adapter_ptr->port_info[i].status == ICOM_PORT_ACTIVE)
	diag_port_load(i,
		       &iCom_adapter_ptr->port_info[i],
                       iCom_adapter_ptr->base_addr,
		       towrapdiag, 
		       sizeof(towrapdiag));

  /* clear int regs */
  iCom_adapter_ptr->tpr = DIAG_RVXINTWRP_CLEAR_INT_VARS;
  iCom_adapter_ptr->diag_int_reset1 = 0;
  iCom_adapter_ptr->diag_int_reset2 = 0;

  /* Load the Transmit data */
  for(i=0;i<iCom_adapter_ptr->numb_ports;i++)
    if(iCom_adapter_ptr->port_info[i].status == ICOM_PORT_ACTIVE)
    {
	iCom_writel(0x11000000,(void*)(iCom_adapter_ptr->base_addr+ 0x8  +(0x2000*i)));
	iCom_writel(0x22000000,(void*)(iCom_adapter_ptr->base_addr+ 0xc  +(0x2000*i)));
	iCom_writel(0x33000000,(void*)(iCom_adapter_ptr->base_addr+ 0x10 +(0x2000*i)));
	iCom_writel(0x44000000,(void*)(iCom_adapter_ptr->base_addr+ 0x14 +(0x2000*i)));
    }

  /* start all procs */
  iCom_adapter_ptr->tpr = DIAG_RVXINTWRP_START_PROCS;
  for(i=0;i<iCom_adapter_ptr->numb_ports;i++)
    if(iCom_adapter_ptr->port_info[i].status == ICOM_PORT_ACTIVE)
      start_processor(&iCom_adapter_ptr->port_info[i]);
  
  /* determine what ints to look for */
  iCom_adapter_ptr->tpr = DIAG_RVXINTWRP_DETERMINE_INTS_TO_CHECK_FOR;
  if ((iCom_adapter_ptr->version | ADAPTER_V2) != ADAPTER_V2) {
    exp_int1 = 0x01000100;
    exp_int2 = 0;
  }
  else
  {
      if(iCom_adapter_ptr->subsystem_id == FOUR_PORT_MODEL)
      {
	  exp_int1 = 0x01000100;
	  exp_int2 = 0x01000100;
      }
      else /* must be a two port model */
      {
	  exp_int1 = 0x00000100;
	  exp_int2 = 0x00000100;
      }
  }
  
  /* wait for all ints to set or for a second to pass */
  done = 0;
  temp_32 = 0;
  iCom_adapter_ptr->tpr = DIAG_RVXINTWRP_WAIT_FOR_INTS;
  while((!done) && (temp_32 < 1000))
  {
    if((iCom_adapter_ptr->diag_int_reset1 == exp_int1) &&
       (iCom_adapter_ptr->diag_int_reset2 == exp_int2))
      done = 1;

    /**** wait 1 millisecond ****/
    current->state = TASK_INTERRUPTIBLE;
    schedule_timeout(HZ/1000); /* HZ = 1 second */
    temp_32++;
  }

  temp_32 = 0;


  iCom_adapter_ptr->tpr = DIAG_RVXINTWRP_CHECK_FOR_INTS;
  if(!done) /* failed */
  {
      /* fail the appropriate port/s */
      rc = diag_check_ints(iCom_adapter_ptr,
			   exp_int1,
			   exp_int2,
			   "Instruction Test. Int did not occur.\n",
			   DIAG_RVXINTWRP_WRONG_INT_FAIL);
      return -1;
  }
  
  /* stop all procs */
  iCom_adapter_ptr->tpr = DIAG_RVXINTWRP_STOP_PROCS;
  for(i=0;i<iCom_adapter_ptr->numb_ports;i++)
    if(iCom_adapter_ptr->port_info[i].status == ICOM_PORT_ACTIVE)
      stop_processor(&iCom_adapter_ptr->port_info[i]);

  /**** check the results data ****/
   iCom_adapter_ptr->tpr = DIAG_RVXINTWRP_CHECK_RESULTS;
   for(i=0;i<iCom_adapter_ptr->numb_ports;i++)
       if(iCom_adapter_ptr->port_info[i].status == ICOM_PORT_ACTIVE)

       {
      /* read the results word */
	   temp_32 = iCom_readl((void*)(iCom_adapter_ptr->base_addr+(i*0x2000)+4));

      /* check results word */
	   if(temp_32 != 0xAAAAAAAA) /* AAAAAAAA is passing */
	   {
	       error_data[0] = temp_32;

	       diag_error(iCom_adapter_ptr,	/* port to fail */
			  i,  /* IOA_FAILURE indicates an Adapter failure */
			  "iCom: Instruction Test. Bad results data.\n", /* error message */
			  DIAG_RVXINTWRP_BAD_RESULTS_DATA_FAIL,		/* tpr */
			  error_ptr,		/* error data pointer */
			  1);	/* number of error words. */
	       return 0x0000000f;
	   }
       }


  /**** return int handler ****/
  iCom_adapter_ptr->tpr = DIAG_RVXINTWRP_FREE_INT_HANDLE;
  free_irq(iCom_adapter_ptr->pci_dev->irq, (void *)iCom_adapter_ptr);
  iCom_adapter_ptr->resources &= ~HAVE_INT_HANDLE;

  iCom_adapter_ptr->tpr = DIAG_RVXINTWRP_END;
return 0; 
} 


/********************************/
/* Imbedded Modem Internal Wrap */
/********************************/
int diag_sm_int_wrap(void* data)
{
    struct iCom_port *iCom_port_ptr;
    u32		error_data[NUM_ERROR_ENTRIES];
    u32        *error_ptr = &error_data[0];
    struct iCom_adapter *iCom_adapter_ptr;
    u8		port;
    char *att1 = "AT&T1\r\n";
    char *atc0 = "AT%C0\r\n";
    char *ok = "OK";
    char *ath = "+++ATH\r\n";
    char *connect = "CONNECT";
    /* 50 U's */
    char *Us = "UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU"; 

    /* get the port pointer */
    iCom_port_ptr = (struct iCom_port *)data;

    /* get the adapter pointer */
    iCom_adapter_ptr = &iCom_adapter_info[iCom_port_ptr->adapter];

    iCom_adapter_ptr->tpr = DIAG_SMINTWRP_START_TPR;                 

    /* get the port number */
    port = iCom_port_ptr->port;

    /* mask all interrupts */
    if ((iCom_adapter_ptr->version | ADAPTER_V2) != ADAPTER_V2)
	iCom_writel(0xFFFFFFFF,(void*)(iCom_adapter_ptr->base_addr+0x4008));
    else
    {
	iCom_writel(0xFFFF7FFF,(void*)(iCom_adapter_ptr->base_addr+0x8008));
	iCom_writel(0xFFFFFFFF,(void*)(iCom_adapter_ptr->base_addr+0x8028));
    }

    /* turn off compression */  /* send at%c0 */

    /* this is done to init the pico code */
    mdm_send(iCom_port_ptr,"",0);

    iCom_adapter_ptr->tpr = DIAG_SMINTWRP_DISABLE_COMP_TPR;                 
    mdm_send(iCom_port_ptr, atc0, strlen(atc0));

    /* Wait for the receive data */
    iCom_adapter_ptr->tpr = DIAG_SMINTWRP_WAIT_FOR_DATA_TPR;                 
    if(mdm_rcv(iCom_port_ptr, ok))
    {
	/* Put the modem into Analog Loopback mode */  /* send at&t1 */
	iCom_adapter_ptr->tpr = DIAG_SMINTWRP_SET_LOOPBACK_TPR;                 
	mdm_send(iCom_port_ptr, att1, strlen(att1));

	/* Wait for the receive data */
	iCom_adapter_ptr->tpr = DIAG_SMINTWRP_WAIT_FOR_DATA2_TPR;                 
	if(mdm_rcv(iCom_port_ptr, connect))
	{
	    /* send the test data */
	    mdm_send(iCom_port_ptr, Us, strlen(Us));

	    /* Wait for the receive data */
	    iCom_adapter_ptr->tpr = DIAG_SMINTWRP_WAIT_FOR_DATA4_TPR;                 
	    /*if(mdm_rcv(iCom_port_ptr, Us))*/
	    if(mdm_rcv(iCom_port_ptr, Us))
	    {
		/* If there was data received */
		iCom_adapter_ptr->tpr = DIAG_SMINTWRP_SEND_ATH_TPR;                 
		mdm_send(iCom_port_ptr, ath, strlen(ath));

		/* Wait for the receive data */
		iCom_adapter_ptr->tpr = DIAG_SMINTWRP_WAIT_FOR_ATH_OK_TPR;
		if(mdm_rcv(iCom_port_ptr, ok))
		{
                    /* turn off DTR and RTS */
		    iCom_writeb(0,&iCom_port_ptr->dram->osr);
		    iCom_writeb(0,&iCom_port_ptr->dram->CmdReg);

		    /* let the line settle */
		    current->state = TASK_INTERRUPTIBLE;
		    schedule_timeout(HZ/100); /* HZ = 1 second */
		}
		else 
		{
		    /* fail because we did not get an OK after ath */
		    diag_error(iCom_adapter_ptr,
			       port,  /* IOA_FAILURE indicates an Adapter failure */
			       "iCom: SM Int Wrap - Imbedded Modem did not receive OK after ath.\n",
			       DIAG_SMINTWRP_NOT_OK_FAIL_TPR,		/* tpr */
			       error_ptr,		/* error data pointer */
			       0);	/* number of error words. */

		    iCom_port_ptr->thread_status = DIAG_SMINTWRP_NOT_OK_FAIL_TPR;
		    return 0x0000000f;
		}
	    }
	    else
	    {
	        /* fail here because we did not receive all U's */
		diag_error(iCom_adapter_ptr,
			   port,  /* IOA_FAILURE indicates an Adapter failure */
			   "iCom: SM Int Wrap - Imbedded Modem did not receive all U's.\n",
			   DIAG_SMINTWRP_NOT_U_FAIL_TPR,		/* tpr */
			   error_ptr,		/* error data pointer */
			   0);	/* number of error words. */

		iCom_port_ptr->thread_status = DIAG_SMINTWRP_NOT_U_FAIL_TPR;
		return 0x0000000f;
	    }
	}
	else
	{
            /* fail here because we did not receive CONNECT */
	    /* after sending the at&t1 command.             */
	    diag_error(iCom_adapter_ptr,
		       port,  /* IOA_FAILURE indicates an Adapter failure */
		       "iCom: SM Int Wrap - Imbedded Modem did not receive connect after at&t1.\n",
		       DIAG_SMINTWRP_NOT_CONNECT_FAIL_TPR,		/* tpr */
		       error_ptr,		/* error data pointer */
		       0);	/* number of error words. */

	    iCom_port_ptr->thread_status = DIAG_SMINTWRP_NOT_CONNECT_FAIL_TPR;
	    return 0x0000000f;
	}
    }
    else
    {
	/* fail here because we did not receive OK */
	/* after the at%c0 call.                   */
	diag_error(iCom_adapter_ptr,
		   port,  /* IOA_FAILURE indicates an Adapter failure */
		   "iCom: SM Int Wrap - Imbedded Modem did not receive OK after at%%c0.\n",
		   DIAG_SMINTWRP_NOT_OK_FAIL1_TPR,		/* tpr */
		   error_ptr,		/* error data pointer */
		   0);	/* number of error words. */

        iCom_port_ptr->thread_status = DIAG_SMINTWRP_NOT_OK_FAIL1_TPR;
	return 0x0000000f;
    }

    iCom_adapter_ptr->tpr = DIAG_SMINTWRP_END_TPR;

    iCom_port_ptr->thread_status = STATUS_PASS;
    return 0;

}



/*********************************/
/* Four port internal modem test */
/*********************************/
static u32 __init diag_concurrent_mdm_int_wrap(struct iCom_adapter *iCom_adapter_ptr)
{
    struct iCom_port *iCom_port_ptr;
    int               port_num;
    int               task;
    int               loop;
    int               still_waiting;
    int               ports_to_test=0; /* false */


    /* Determine what ports need to run the imbedded modem wrap */
    for (port_num = 0; port_num < iCom_adapter_ptr->numb_ports; port_num++)
    {
	iCom_port_ptr = &iCom_adapter_ptr->port_info[port_num];

	/* init status to pass(for non modem ports) */
	iCom_port_ptr->thread_status = STATUS_PASS; 

        /* If this is a imbedded modem, */
        /* run the internal wrap test.  */
        if ((iCom_port_ptr->status == ICOM_PORT_ACTIVE) &&
            (iCom_port_ptr->passed_diags) &&
	    (iCom_port_ptr->imbed_modem == ICOM_IMBED_MODEM))
	{
            /* change the status to INIT */
            iCom_port_ptr->thread_status = STATUS_INIT;

            /* start a new thread to test the port */
            task = kernel_thread(diag_sm_int_wrap, iCom_port_ptr, 0);

            ports_to_test = 1;
	}
    }

    /* Wait for the ports to finish */
    loop = 0;
    still_waiting = 1;
    while((loop<20) && (still_waiting) && (ports_to_test))
    {
	/* see if they have all completed */
	for(port_num = still_waiting = 0; port_num < iCom_adapter_ptr->numb_ports; port_num++)
	{
	    iCom_port_ptr = &iCom_adapter_ptr->port_info[port_num];
	    if(iCom_port_ptr->thread_status == STATUS_INIT)
	    {
		still_waiting |= 1;
		continue;
	    }
	}
	loop++;

	current->state = TASK_INTERRUPTIBLE;
	schedule_timeout(HZ); /* HZ = 1 second */
    }

    /* see if all the ports passed */
    for(port_num = 0; port_num < iCom_adapter_ptr->numb_ports; port_num++)
    {
	iCom_port_ptr = &iCom_adapter_ptr->port_info[port_num];

	if(iCom_port_ptr->imbed_modem == ICOM_IMBED_MODEM) /* true */
	{
	    /* turn off DTR and RTS */
	    iCom_writeb(0,&iCom_port_ptr->dram->osr);
	    iCom_writeb(0,&iCom_port_ptr->dram->CmdReg);

	    /* check if passed */
	    if(iCom_port_ptr->thread_status != STATUS_PASS)
	    {
		iCom_port_ptr->passed_diags = 0; /* FAIL */
	    }
	}
    }

    return 0;
}

	      

  

/**************************/
/* Main Diagnostic Driver */
/**************************/
static u32 __init diag_main(struct iCom_adapter *iCom_adapter_ptr)
{
  u32           rc = 0;
  int		i;
#ifdef icom_debug
  struct LocationDataStruct *p_location_data = NULL;
#endif

  iCom_adapter_ptr->tpr = DIAG_MAIN_START;
#ifdef icom_debug
  printk("iCom: Diagnostic started. ********\n");
#ifdef CONFIG_PPC_ISERIES
  p_location_data = iSeries_GetLocationData(iCom_adapter_ptr->pci_dev);

  if (p_location_data != NULL)
  {
      printk("iCom: Frame ID: %d Card Position: %s.\n",p_location_data->FrameId, p_location_data->CardLocation);
      kfree(p_location_data);
  }
#endif
#endif

  iCom_adapter_ptr->tpr = DIAG_MAIN_INIT;

  diag_init(iCom_adapter_ptr);


  iCom_adapter_ptr->tpr = DIAG_MAIN_BIST;
  /* run BIST test */
  if(rc==0)
  {
#ifdef icom_debug
      printk("iCom: Diagnostic BIST test.\n");
#endif
      rc = diag_bist(iCom_adapter_ptr);
      if(rc!=0)
      {
	  printk("iCom: Diagnostic BIST test FAILED.\n");
      }
  }

  iCom_adapter_ptr->tpr = DIAG_MAIN_MEMORY;
  /* run memory test */
  if(rc==0)
  {
#ifdef icom_debug
      printk("iCom: Diagnostic memory.\n");
#endif
      rc = diag_memory(iCom_adapter_ptr);
      if(rc!=0)
      {
	  printk("iCom: Diagnostic memory test FAILED.\n");
      }
  }


  iCom_adapter_ptr->tpr = DIAG_MAIN_INTERRUPT;
  /* run interrupt test */
  if(rc==0)
  {
#ifdef icom_debug
      printk("iCom: Diagnostic interrupt test.\n");
#endif
      rc = diag_int(iCom_adapter_ptr);
      if(rc!=0)
      {
	  printk("iCom: Diagnostic interrupt test FAILED.\n");
      }
  }

  iCom_adapter_ptr->tpr = DIAG_MAIN_INSTRUCTION;
  /* run instruction test */
  if(rc==0)
  {
#ifdef icom_debug
      printk("iCom: Diagnostic instruction test.\n");
#endif
      rc = diag_inst(iCom_adapter_ptr);
      if(rc!=0)
      {
	  printk("iCom: Diagnostic instruction test FAILED.\n");
      }
  }

  iCom_adapter_ptr->tpr = DIAG_MAIN_REGISTER;
  /* run register test */
  if(rc==0)
  {
#ifdef icom_debug
      printk("iCom: Diagnostic register test.\n");
#endif
      rc = diag_reg(iCom_adapter_ptr);
      if(rc!=0)
      {
	  printk("iCom: Diagnostic register test FAILED.\n");
      }
  }


  iCom_adapter_ptr->tpr = DIAG_MAIN_RVX_STATUS;
  /* run RVX status register test */
  if(rc==0)
  {
#ifdef icom_debug
      printk("iCom: Diagnostic RVX Status Register test.\n");
#endif
      rc = diag_rvx_sts(iCom_adapter_ptr);
      if(rc!=0)
      {
	  printk("iCom: Diagnostic RVX Status register test FAILED.\n");
      }
  }

  iCom_adapter_ptr->tpr = DIAG_MAIN_RVX_INTERNAL_WRAP;
  /* run RVX internal wrap test */
  if(rc==0)
  {
#ifdef icom_debug
      printk("iCom: Diagnostic RVX Internal Wrap test.\n");
#endif
      rc = diag_rvx_internal_wrap(iCom_adapter_ptr);
      if(rc!=0)
      {
	  printk("iCom: Diagnostic RVX Internal Wrap test FAILED.\n");
      }
  }

  /* make sure all procs are stoped */
  for(i=0;i<iCom_adapter_ptr->numb_ports;i++)
      if(iCom_adapter_ptr->port_info[i].status == ICOM_PORT_ACTIVE)
	  stop_processor(&iCom_adapter_ptr->port_info[i]);

  /* return the resources */
  diag_return_resources(iCom_adapter_ptr);

  /* mask all interrupts */
  if ((iCom_adapter_ptr->version | ADAPTER_V2) != ADAPTER_V2)
      iCom_writel(0xFFFFFFFF,(void*)(iCom_adapter_ptr->base_addr+0x4008));
  else
  {
      iCom_writel(0xFFFF7FFF,(void*)(iCom_adapter_ptr->base_addr+0x8008));
      iCom_writel(0xFFFFFFFF,(void*)(iCom_adapter_ptr->base_addr+0x8028));
  }

  iCom_adapter_ptr->tpr = DIAG_MAIN_END;

  if(rc==0)
  {
#ifdef icom_debug
     printk("iCom: Diagnostic Completed Successfully ********\n");
#endif

      /* set the TPR and error data to zero for all ports */
      for(i=0;i<iCom_adapter_ptr->numb_ports;i++)
      {
	  iCom_adapter_ptr->port_info[i].passed_diags = 1; /* true */
      }
  }

  else
  {
#ifdef icom_debug
      printk("iCom: Diagnostic FAILED ********\n");
#endif
  }


  return rc;
}

static void __init diag_return_resources(struct iCom_adapter *iCom_adapter_ptr)
{
  /* make sure all resources are returned before leaving */
  if(iCom_adapter_ptr->resources & HAVE_INT_HANDLE)
  {
      free_irq(iCom_adapter_ptr->pci_dev->irq, (void *)iCom_adapter_ptr);
      iCom_adapter_ptr->resources &= ~HAVE_INT_HANDLE; /* clear resource bit */
  }

  if(iCom_adapter_ptr->resources & HAVE_MALLOC_1)
  {
      kfree(iCom_adapter_ptr->malloc_addr_1);
      iCom_adapter_ptr->resources &= ~HAVE_MALLOC_1; /* clear resource bit */
  }

}


/*
   Module operations
*/
int iCom_init(void)
{
    int               index, 
                      index_v2,
                      index2,
                      scan_index;
    struct pci_dev   *dev[MAX_ADAPTERS];
    unsigned int      irq_number[MAX_ADAPTERS];
    unsigned long int base_addr[MAX_ADAPTERS];
    unsigned char     valid_indices[MAX_ADAPTERS];
#define VALID 1
#define INVALID 0
    unsigned int      command_reg;
    struct iCom_port *iCom_port_ptr;
    int               retval;
    int               status;
    int               port_num;
    int               adapter_count = 0;
    int               duplicate;
    unsigned int      subsystem_id;
    int               minor_number = 0;


    /*
     * Find base addresses and IRQs for any/all installed cards
     */
    for (index=0; index < MAX_ADAPTERS; index++) {
	valid_indices[index] = INVALID;
	dev[index] = NULL;
    }

    /* check for Version 1 Adapters */
    for (index = 0; index < MAX_ADAPTERS; index++){
	if (index == 0) {
	    if (!(dev[index] = pci_find_device(VENDOR_ID, DEVICE_ID, dev[index])))
		break;
	}
	else {
	    if (!(dev[index] = pci_find_device(VENDOR_ID, DEVICE_ID, dev[index-1])))
		break;
	}

	adapter_count++;

	if (pci_enable_device(dev[index])) {
	    printk("iCom:  Device enable FAILED\n");
	    continue;
	}

	if (pci_read_config_dword(dev[index], PCI_COMMAND, &command_reg)) {
	    printk("iCom:  PCI Config read FAILED\n");
	    continue;
	}	

	pci_write_config_dword(dev[index],PCI_COMMAND, command_reg | 0x00000146);
	pci_write_config_dword(dev[index],0x44, 0x8300830A);

	base_addr[index] = pci_resource_start(dev[index],0);
	base_addr[index] &= PCI_BASE_ADDRESS_MEM_MASK;

	duplicate = 0;
	for (index2 = 0; index2 < index; index2++) {
	    if (base_addr[index] == base_addr[index2])
		duplicate = 1;
	}
	if (duplicate) continue;

	irq_number[index] = dev[index]->irq;

	valid_indices[index] = ADAPTER_V1;
    }

    /* check for version 2 Adapters */
    for (index_v2=0; index < MAX_ADAPTERS; index_v2++){
	if (index_v2 == 0) {
	    if (!(dev[index] = pci_find_device(VENDOR_ID, DEVICE_ID2, NULL))) {
		break;
	    }
	}
	else {
	    if (!(dev[index] = pci_find_device(VENDOR_ID, DEVICE_ID2, dev[index-1]))) {
		break;
	    }
	}

	adapter_count++;

	if (pci_enable_device(dev[index])) {
	    printk("iCom:  Device enable FAILED\n");
	    continue;
	}

	if (pci_read_config_dword(dev[index], PCI_COMMAND, &command_reg)) {
	    printk("iCom:  PCI Config read FAILED\n");
	    continue;
	}	

	pci_write_config_dword(dev[index],PCI_COMMAND, command_reg | 0x00000146);
	pci_write_config_dword(dev[index],0x44, 0x42004200);
	pci_write_config_dword(dev[index],0x48, 0x42004200);

	base_addr[index] = pci_resource_start(dev[index],0);	
	base_addr[index] &= PCI_BASE_ADDRESS_MEM_MASK;

	duplicate = 0;
	for (index2 = 0; index2 < index; index2++) {
	    if (base_addr[index] == base_addr[index2])
		duplicate = 1;
	}
	if (duplicate) continue;

	irq_number[index] = dev[index]->irq;

	valid_indices[index++] = ADAPTER_V2;
    }

    /* allocate memory for control blocks representing each adapter */
    iCom_adapter_info = (struct iCom_adapter *)
      kmalloc(adapter_count*sizeof(struct iCom_adapter),GFP_KERNEL);

    if (!iCom_adapter_info) {
	return -ENOMEM;
    }

    memset(iCom_adapter_info, 0,adapter_count*sizeof(struct iCom_adapter));

    /* store information just obtained on base_addr and irq */
    for (index = scan_index = 0; (scan_index < MAX_ADAPTERS) &
        (index < adapter_count);       scan_index++) {

	if (valid_indices[scan_index]) {
	    iCom_adapter_info[index].base_addr = base_addr[scan_index];
	    iCom_adapter_info[index].irq_number = irq_number[scan_index];
	    iCom_adapter_info[index].pci_dev = dev[scan_index];
	    iCom_adapter_info[index].version = valid_indices[scan_index];
	    pci_read_config_dword(dev[scan_index], PCI_SUBSYSTEM_VENDOR_ID, &subsystem_id);
	    iCom_adapter_info[index].subsystem_id = subsystem_id;

	    if (iCom_adapter_info[index].version == ADAPTER_V1) {
		iCom_adapter_info[index].numb_ports = 2;
		iCom_adapter_info[index].port_info[0].port = 0;
                iCom_adapter_info[index].port_info[0].minor_number = minor_number;
		iCom_adapter_info[index].port_info[0].status = ICOM_PORT_ACTIVE;
		iCom_adapter_info[index].port_info[0].imbed_modem = ICOM_UNKNOWN;
		iCom_adapter_info[index].port_info[1].port = 1;
                iCom_adapter_info[index].port_info[1].minor_number = minor_number + 1;
		iCom_adapter_info[index].port_info[1].status = ICOM_PORT_ACTIVE;
		iCom_adapter_info[index].port_info[1].imbed_modem = ICOM_UNKNOWN;
	    }
	    else {
		if (subsystem_id == FOUR_PORT_MODEL) {
		    iCom_adapter_info[index].numb_ports = 4;

		    iCom_adapter_info[index].port_info[0].port = 0;
                    iCom_adapter_info[index].port_info[0].minor_number = minor_number;
		    iCom_adapter_info[index].port_info[0].status = ICOM_PORT_ACTIVE;
		    iCom_adapter_info[index].port_info[0].imbed_modem = ICOM_IMBED_MODEM;

		    iCom_adapter_info[index].port_info[1].port = 1;
                    iCom_adapter_info[index].port_info[1].minor_number = minor_number + 1;
		    iCom_adapter_info[index].port_info[1].status = ICOM_PORT_ACTIVE;
		    iCom_adapter_info[index].port_info[1].imbed_modem = ICOM_IMBED_MODEM;

		    iCom_adapter_info[index].port_info[2].port = 2;
                    iCom_adapter_info[index].port_info[2].minor_number = minor_number + 2;
		    iCom_adapter_info[index].port_info[2].status = ICOM_PORT_ACTIVE;
		    iCom_adapter_info[index].port_info[2].imbed_modem = ICOM_IMBED_MODEM;

		    iCom_adapter_info[index].port_info[3].port = 3;
                    iCom_adapter_info[index].port_info[3].minor_number = minor_number + 3;
		    iCom_adapter_info[index].port_info[3].status = ICOM_PORT_ACTIVE;
		    iCom_adapter_info[index].port_info[3].imbed_modem = ICOM_IMBED_MODEM;
		}
		else {
		    iCom_adapter_info[index].numb_ports = 4;

		    iCom_adapter_info[index].port_info[0].port = 0;
                    iCom_adapter_info[index].port_info[0].minor_number = minor_number;
		    iCom_adapter_info[index].port_info[0].status = ICOM_PORT_ACTIVE;
		    if (subsystem_id == V2_ONE_PORT_RVX_ONE_PORT_IMBED_MDM)
			iCom_adapter_info[index].port_info[0].imbed_modem = ICOM_IMBED_MODEM;
		    else
			iCom_adapter_info[index].port_info[0].imbed_modem = ICOM_RVX;

		    iCom_adapter_info[index].port_info[1].status = ICOM_PORT_OFF;

		    iCom_adapter_info[index].port_info[2].port = 2;
                    iCom_adapter_info[index].port_info[2].minor_number = minor_number + 1;
		    iCom_adapter_info[index].port_info[2].status = ICOM_PORT_ACTIVE;
		    iCom_adapter_info[index].port_info[2].imbed_modem = ICOM_RVX;

		    iCom_adapter_info[index].port_info[3].status = ICOM_PORT_OFF;
		}
	    }
            minor_number += 4;

	    if (!request_mem_region(iCom_adapter_info[index].base_addr,
				    pci_resource_len(iCom_adapter_info[index].pci_dev,0),
				    "iCom")) {
		printk("iCom:  request_mem_region FAILED\n");
	    }


	    retval = diag_main(&iCom_adapter_info[index]);

	    if(retval)
	    {
		release_mem_region(iCom_adapter_info[index].base_addr,
				   pci_resource_len(iCom_adapter_info[index].pci_dev,0));
		continue;	
	    }

            /* save off irq and request irq line */
	    if (request_irq(irq_number[scan_index], iCom_interrupt, SA_INTERRUPT |
			    SA_SHIRQ, ICOM_DRIVER_NAME, (void *)&iCom_adapter_info[index])) {
		release_mem_region(iCom_adapter_info[index].base_addr,
				   pci_resource_len(iCom_adapter_info[index].pci_dev,0));
		printk("iCom:  request_irq FAILED\n");
		continue;
	    }

	    for (port_num = 0; port_num < iCom_adapter_info[index].numb_ports; port_num++) {
		iCom_port_ptr = &iCom_adapter_info[index].port_info[port_num];

		if (iCom_port_ptr->status == ICOM_PORT_ACTIVE) {
		    /* initialize wait queues */
		    init_waitqueue_head(&iCom_port_ptr->open_wait);
		    init_waitqueue_head(&iCom_port_ptr->close_wait);
		    init_waitqueue_head(&iCom_port_ptr->delta_msr_wait);

		    /* initialize port specific variables */
		    iCom_port_ptr->tqueue.routine = do_softint;
		    iCom_port_ptr->tqueue.data = iCom_port_ptr;
		    if (iCom_adapter_info[index].version == ADAPTER_V1) {
			iCom_port_ptr->global_reg = (struct iCom_regs *)((char *)iCom_adapter_info[index].base_addr + 0x4000);
			iCom_port_ptr->int_reg = (unsigned long)iCom_adapter_info[index].base_addr + 0x4004 + 2 - 2 * port_num;
		    }
		    else {
			iCom_port_ptr->global_reg = (struct iCom_regs *)((char *)iCom_adapter_info[index].base_addr + 0x8000);
			if (iCom_port_ptr->port < 2)
			    iCom_port_ptr->int_reg = (unsigned long)iCom_adapter_info[index].base_addr + 0x8004 + 2 - 2 * iCom_port_ptr->port;
			else
			    iCom_port_ptr->int_reg = (unsigned long)iCom_adapter_info[index].base_addr + 0x8024 + 2 - 2 * (iCom_port_ptr->port - 2);
		    }
		    iCom_port_ptr->dram = (struct func_dram*)((char*)iCom_adapter_info[index].base_addr + 0x2000 * iCom_port_ptr->port);
		    iCom_port_ptr->close_delay = 5*HZ/10;
		    iCom_port_ptr->closing_wait = 30*HZ;
		    iCom_port_ptr->adapter = index;

		    /*
		     * Load and start processor
		     */
		    loadCode(iCom_port_ptr);

	   	    /* get port memory */
		    if ((status = get_port_memory(iCom_port_ptr)) != 0) {
			printk("iCom:  memory allocation for minor number %d port FAILED\n",iCom_port_ptr->minor_number);

			/* Fail the port, though not technically correct to call the port bad
			   due to diagnostics the end result is this port should not be accessed
			 */
			iCom_port_ptr->passed_diags = 0;
			continue;
		    }

		    /* Set Country Code */
		    iCom_set_code(iCom_port_ptr);
		}
	    }
/*            retval = diag_concurrent_mdm_int_wrap(&iCom_adapter_info[index]);
	    if(retval) continue;
*/	    index++;
	}
    }
    active_adapters = index;

    printk("iCom:  Adapter detection complete, %d adapters found with %d valid\n",adapter_count,active_adapters);

    if (active_adapters > 0) {

      /* Initialize the tty_driver structure */
	memset(&serial_driver, 0, sizeof(struct tty_driver));
	serial_driver.magic = TTY_DRIVER_MAGIC;
	serial_driver.driver_name = ICOM_DRIVER_NAME;
#if defined(CONFIG_DEVFS_FS)
	serial_driver.name = "ttyA%d";
#else
	serial_driver.name = "ttyA";
#endif
	serial_driver.major = 243;
	serial_driver.minor_start = 0;
	serial_driver.num = NR_PORTS;
	serial_driver.type = TTY_DRIVER_TYPE_SERIAL;
	serial_driver.subtype = SERIAL_TYPE_NORMAL;
	serial_driver.init_termios = tty_std_termios;
	serial_driver.init_termios.c_cflag = B9600 | CS8 | CREAD | HUPCL | CLOCAL;
	serial_driver.flags = TTY_DRIVER_REAL_RAW | TTY_DRIVER_NO_DEVFS;
	serial_driver.refcount = &serial_refcount;
	serial_driver.table = serial_table;
	serial_driver.termios = serial_termios;
	serial_driver.termios_locked = serial_termios_locked;

	serial_driver.open = iCom_open;
	serial_driver.close = iCom_close;
	serial_driver.write = iCom_write;
	serial_driver.put_char = iCom_put_char;
	serial_driver.flush_chars = iCom_flush_chars;
	serial_driver.write_room = iCom_write_room;
	serial_driver.chars_in_buffer = iCom_chars_in_buffer;
	serial_driver.flush_buffer = iCom_flush_buffer;
	serial_driver.ioctl = iCom_ioctl;
	serial_driver.throttle = iCom_throttle;
	serial_driver.unthrottle = iCom_unthrottle;
	serial_driver.send_xchar = iCom_send_xchar;
	serial_driver.set_termios = iCom_set_termios;
	serial_driver.stop = iCom_stop;
	serial_driver.start = iCom_start;
	serial_driver.hangup = iCom_hangup;
	serial_driver.break_ctl = iCom_break;
	serial_driver.wait_until_sent = iCom_wait_until_sent;
	serial_driver.read_proc = iCom_read_proc;


	for (index=0; index < active_adapters; index++) {
	    iCom_adapter_info[index].port_info[0].callout_termios = serial_driver.init_termios;
	    iCom_adapter_info[index].port_info[0].normal_termios = serial_driver.init_termios;
	    iCom_adapter_info[index].port_info[1].callout_termios = serial_driver.init_termios;
	    iCom_adapter_info[index].port_info[1].normal_termios = serial_driver.init_termios;
	    iCom_adapter_info[index].port_info[2].callout_termios = serial_driver.init_termios;
	    iCom_adapter_info[index].port_info[2].normal_termios = serial_driver.init_termios;
	    iCom_adapter_info[index].port_info[3].callout_termios = serial_driver.init_termios;
	    iCom_adapter_info[index].port_info[3].normal_termios = serial_driver.init_termios;
	}

	if (tty_register_driver(&serial_driver)) {
	    for (index=0; index < active_adapters; index++) {
		free_irq(iCom_adapter_info[index].irq_number, (void *)&iCom_adapter_info[index]);
	    }
	    kfree(iCom_adapter_info);
	    panic("Couldn't register serial driver\n");
	}

#if defined(CONFIG_DEVFS_FS)
	for (index = 0; index < active_adapters; index++) {
	    tty_register_devfs(&serial_driver,
			       0, index*4 + serial_driver.minor_start);
	    tty_register_devfs(&serial_driver,
			       0, index*4 + serial_driver.minor_start + 1);

	    if ((iCom_adapter_info[index].version == ADAPTER_V2) &&
		(iCom_adapter_info[index].subsystem_id == FOUR_PORT_MODEL)) {
		tty_register_devfs(&serial_driver,
				   0, index*4 + serial_driver.minor_start + 2);
		tty_register_devfs(&serial_driver,
				   0, index*4 + serial_driver.minor_start + 3);
	    }
	}
#endif

	/* lastly, register unique ioctl */
	register_ioctl32_conversion(0x4300,NULL);

	return 0;
    }
    else {
	if (adapter_count > 0) {
	    kfree(iCom_adapter_info);
	}
    }

    return -ENODEV;
}

int init_module(void)
{
#ifdef MODULE
   printk("icom:  Start of module: 0x%lx\n",(unsigned long)THIS_MODULE + (u32)THIS_MODULE->size_of_struct);
#endif
   return iCom_init();
}

void cleanup_module(void) 
{
  unsigned long       flags;
  int                 e1;
  int                 index;
  int                 port_num;
  struct iCom_port   *iCom_port_ptr;

  /* remove registered ioctl */
  unregister_ioctl32_conversion(0x4300);

  spin_lock_irqsave(&iComlock,flags);
  if ((e1 = tty_unregister_driver(&serial_driver)))
    printk("iCom:  failed to unregister serial driver (%d)\n",e1);

#if defined(CONFIG_DEVFS_FS)
  for (index = 0; index < active_adapters; index++) {
      tty_unregister_devfs(&serial_driver,
			   index*4 + serial_driver.minor_start);
      tty_unregister_devfs(&serial_driver,
			   index*4 + serial_driver.minor_start + 1);

      if ((iCom_adapter_info[index].version == ADAPTER_V2) &&
	  (iCom_adapter_info[index].subsystem_id == FOUR_PORT_MODEL)) {
	  tty_unregister_devfs(&serial_driver,
			     index*4 + serial_driver.minor_start + 2);
	  tty_unregister_devfs(&serial_driver,
			     index*4 + serial_driver.minor_start + 3);
      }
  }
#endif

  for (index=0; index < active_adapters; index++) {

    for (port_num = 0; port_num < iCom_adapter_info[index].numb_ports; port_num++) {
      iCom_port_ptr = &iCom_adapter_info[index].port_info[port_num];

      if (iCom_port_ptr->status == ICOM_PORT_ACTIVE) {

          /* be sure that DTR and RTS are dropped */
	  iCom_writeb(0x00,&iCom_port_ptr->dram->osr);

          /* Wait 0.1 Sec for simple Init to complete */
	  current->state = TASK_INTERRUPTIBLE;
	  schedule_timeout(HZ/10);

          /* Stop proccessor */
	  stop_processor(iCom_port_ptr);

	  return_port_memory(iCom_port_ptr);
      }
    }

    free_irq(iCom_adapter_info[index].irq_number, (void *)&iCom_adapter_info[index]);
    release_mem_region(iCom_adapter_info[index].base_addr,
		       pci_resource_len(iCom_adapter_info[index].pci_dev,0));
  }
  spin_unlock_irqrestore(&iComlock,flags);

  kfree(iCom_adapter_info);
  printk("iCom:  Driver removed\n");
}

/* the interrupts should be disabled here so that no hardware
 interrupts should occur, polling will be done to check received
 data to avoid interrupt level processing */
static int mdm_rcv(struct iCom_port *iCom_port_ptr, char *exp_str) {
  int status = 0;
  int loop_count = 0;
  char string[256];


  /* init string to null */
  memset(string, 0, 256);

  /* search expected string in received data */
  while (!status && (loop_count++ < 50))
  {
      if(cpu_to_le16(iCom_port_ptr->statStg->rcv[0].flags) & SA_FL_RCV_DONE)
      {
	  /* copy received data to string */
	  strcat(string, (char *)iCom_port_ptr->recv_buf);

	  /* reset the buffer */
	  memset((char *)iCom_port_ptr->recv_buf, 0, RCV_BUFF_SZ);
	  iCom_port_ptr->statStg->rcv[0].flags = 0;
	  iCom_port_ptr->statStg->rcv[0].leLength = 0;
	  iCom_port_ptr->statStg->rcv[0].WorkingLength = (unsigned short int)cpu_to_le16(RCV_BUFF_SZ);
      }
      else if(cpu_to_le16(iCom_port_ptr->statStg->rcv[1].flags) & SA_FL_RCV_DONE)
      {
	  /* copy received data to string */
	  strcat(string, (char *)iCom_port_ptr->recv_buf + 2048);

	  /* reset the buffer */
	  memset((char *)iCom_port_ptr->recv_buf+2048, 0, RCV_BUFF_SZ);
	  iCom_port_ptr->statStg->rcv[1].flags = 0;
	  iCom_port_ptr->statStg->rcv[1].leLength = 0;
	  iCom_port_ptr->statStg->rcv[1].WorkingLength = (unsigned short int)cpu_to_le16(RCV_BUFF_SZ);
      }

      /* see if we have a match */
      if(strstr(string, exp_str))
      {
	  status = 1;
      }
      else
      {

          /* wait .1 seconds */
	  current->state = TASK_INTERRUPTIBLE;
	  schedule_timeout(HZ/10);
      }
  }

  /* clear interrupts */
  iCom_writew(0x3FFF,(void *)iCom_port_ptr->int_reg);

  return status;
}

static void mdm_send(struct iCom_port *iCom_port_ptr, char *mdm_cmnd,
                     int cmnd_length) {

  unsigned char     cmdReg;
  unsigned long int offset;
  char	tmp_byte;
  int index;

  if (0 == cmnd_length) {

    /* turn on DTR and RTS */
    iCom_writeb(0xC0,&iCom_port_ptr->dram->osr);
    tmp_byte = iCom_readb(&(iCom_port_ptr->dram->HDLCConfigReg));
    tmp_byte |= HDLC_HDW_FLOW;
    iCom_writeb(tmp_byte, &(iCom_port_ptr->dram->HDLCConfigReg));

    /* initialize transmit and receive operations */
    offset = (unsigned long int)&iCom_port_ptr->statStg->rcv[0] - (unsigned long int)iCom_port_ptr->statStg;
    iCom_writel(iCom_port_ptr->statStg_pci + offset,&iCom_port_ptr->dram->RcvStatusAddr);
    iCom_port_ptr->next_rcv = 0;
    iCom_port_ptr->put_length = 0;
    *iCom_port_ptr->xmitRestart = 0;
    iCom_writel(iCom_port_ptr->xmitRestart_pci,&iCom_port_ptr->dram->XmitStatusAddr);
    iCom_writeb(CMD_XMIT_RCV_ENABLE,&iCom_port_ptr->dram->CmdReg);

    /* wait for CTS */
    for (index=0; index < 20; index++) {
	current->state = TASK_INTERRUPTIBLE;
	schedule_timeout(HZ/100); /* HZ = .01 second */
	if (iCom_readb(&iCom_port_ptr->dram->isr) & ICOM_CTS) break;
    }
    if (index == 20) printk("iCom:  WARNING CTS not up in 200ms for BATs\n");
    current->state = TASK_INTERRUPTIBLE;
    schedule_timeout(HZ); /* HZ = 1 second */

    return;
  }
 
  /* clear target receive buffers 1 and 2 */
  memset(iCom_port_ptr->recv_buf,0,4096);

  memcpy(iCom_port_ptr->xmit_buf, mdm_cmnd, cmnd_length);

  iCom_port_ptr->statStg->xmit[0].flags = (unsigned short int)cpu_to_le16(SA_FLAGS_READY_TO_XMIT);
  iCom_port_ptr->statStg->xmit[0].leLength = (unsigned short int)cpu_to_le16(cmnd_length);
    offset = (unsigned long int)&iCom_port_ptr->statStg->xmit[0] - (unsigned long int)iCom_port_ptr->statStg;
  *iCom_port_ptr->xmitRestart = cpu_to_le32(iCom_port_ptr->statStg_pci + offset);

  cmdReg = iCom_readb(&iCom_port_ptr->dram->CmdReg);
  iCom_writeb(cmdReg | CMD_XMIT_RCV_ENABLE,&iCom_port_ptr->dram->CmdReg);

  iCom_writeb(START_XMIT,&iCom_port_ptr->dram->StartXmitCmd);
  TRACE(iCom_port_ptr,TRACE_WRITE_START,cmnd_length);
}

static void iCom_set_code(struct iCom_port *iCom_port_ptr) {
  char mdm_cmnd[15];
  int index;

  /* check if country code should be set at all */
  if (strlen(iCom_country_code) == 0) return;

  /* check if cabling present */
  if (iCom_port_ptr->cable_id != RS232_CABLE) return;

  /* initialize xmit/rcv */
  mdm_send(iCom_port_ptr,"",0);

  printk("iCom:  Checking for country code on serial adapter %d port %d...",iCom_port_ptr->adapter, iCom_port_ptr->port);

  /* sync up modems (if present) */
  mdm_send(iCom_port_ptr,"AT\r\n",4);
  mdm_rcv(iCom_port_ptr,"OK");

  printk(".");

  /* send ATI0 to check for internal modem */
  mdm_send(iCom_port_ptr,"ATI0\r\n",6);

  /* check returned data for internal modem identification */
  if (!mdm_rcv(iCom_port_ptr,"SMI")) {
    /* must not be internal modem - return */
    printk("\n\tno internal modem on this port\n");
    return;
  }

  /* send ATE0S0=0 to turn off command Echo and turn off Auto Answer */
  mdm_send(iCom_port_ptr,"ATE0S0=0\r\n",10);

  /* wait for OK */
  if (!mdm_rcv(iCom_port_ptr,"OK")) {
    /* unable to send command to modem - error */
    printk("\niCom:  Error, unable to set modem Country Code\n");
    return;
  }

  printk(".");

  /* Send new country code AT%T19,0,<new value> */
  sprintf(mdm_cmnd,"AT%%T19,0,%s\r\n",iCom_country_code);
  mdm_send(iCom_port_ptr,mdm_cmnd,strlen(mdm_cmnd));

  /* wait for OK */
  if (!mdm_rcv(iCom_port_ptr,"OK")) {
    /* unable to set country code */
    printk("\niCom:  Error, unable to set modem Country Code\n");
    return;
  }

  printk(".\n");

  /* send ATE1S0=2 to enable command Echo and auto answer */
  mdm_send(iCom_port_ptr,"ATE1S0=2\r\n",10);

  /* wait for OK */
  if (!mdm_rcv(iCom_port_ptr,"OK")) {
    /* modem state unknown */
    printk("iCom:  Warning, modem state unknown\n");
  }

  /* print message that Country Code set appropriately */
  printk("iCom:  Modem country code for adapter %d port %d has been set to %s\n",iCom_port_ptr->adapter, iCom_port_ptr->port,iCom_country_code);

  /* disable xmitter/recvr */
  iCom_writeb(CMD_RCV_DISABLE,&iCom_port_ptr->dram->CmdReg);
  for (index = 0; index < 10; index++) {
      /* Wait 0.1 Sec for receive operations to complete*/
      current->state = TASK_INTERRUPTIBLE;
      schedule_timeout(HZ/10);

      if (iCom_readb(&iCom_port_ptr->dram->PrevCmdReg) == 0x00) {
	  break;
      }
  }

  /* clear interrupts */
  iCom_writew(0x3FFF,(void *)iCom_port_ptr->int_reg);
}

#ifdef ICOM_TRACE
void TRACE(struct iCom_port *iCom_port_ptr, u32 trace_pt,
	   u32 trace_data) {

  u32 *tp_start, *tp_end, **tp_next;

  if (trace_pt == TRACE_GET_MEM) {
    if (iCom_port_ptr->trace_blk != 0) return;
    iCom_port_ptr->trace_blk = kmalloc(TRACE_BLK_SZ,GFP_KERNEL);
    memset(iCom_port_ptr->trace_blk, 0,TRACE_BLK_SZ);
    iCom_port_ptr->trace_blk[0] = (unsigned long)iCom_port_ptr->trace_blk + 3*sizeof(unsigned long);
    iCom_port_ptr->trace_blk[1] = (unsigned long)iCom_port_ptr->trace_blk + TRACE_BLK_SZ;
    iCom_port_ptr->trace_blk[2] = iCom_port_ptr->trace_blk[0];
  }
  if (iCom_port_ptr->trace_blk == 0) return;

  if (trace_pt == TRACE_RET_MEM) {
    kfree(iCom_port_ptr->trace_blk);
    iCom_port_ptr->trace_blk = 0;
    return;
  }

  tp_start  = (u32 *)iCom_port_ptr->trace_blk[0];
  tp_end    = (u32 *)iCom_port_ptr->trace_blk[1];
  tp_next   = (u32 **)&iCom_port_ptr->trace_blk[2];

  if (trace_data != 0) {
    **tp_next = trace_data;
    *tp_next = *tp_next + 1;
    if (*tp_next == tp_end) *tp_next = tp_start;
    **tp_next = TRACE_WITH_DATA | trace_pt;
  }
  else
    **tp_next = trace_pt;
  
  *tp_next = *tp_next + 1;
  if (*tp_next == tp_end) *tp_next = tp_start;
}
#endif
